@@ -19,11 +19,6 @@ typedef struct _attrib_t {
 	char timestr[2][MAXTIMESTR + 1];
 } attrib_t;
 
-#ifndef	_NODOSDRIVE
-extern int preparedrv __P_((int));
-extern int shutdrv __P_((int));
-#endif
-
 extern reg_t *findregexp;
 extern int subwindow;
 extern int win_x;
@@ -33,7 +28,7 @@ extern int tradlayout;
 extern int sizeinfo;
 extern int mark;
 #ifdef	HAVEFLAGS
-extern u_long flaglist[];
+extern u_long fflaglist[];
 #endif
 #if	!defined (_USEDOSCOPY) && !defined (_NOEXTRACOPY)
 extern int inheritcopy;
@@ -76,7 +71,7 @@ int copypolicy = 0;
 int removepolicy = 0;
 char *destpath = NULL;
 
-static u_int attrmode = 0;
+static u_short attrmode = 0;
 #ifdef	HAVEFLAGS
 static u_long attrflags = 0;
 #endif
@@ -183,7 +178,7 @@ char *mes, *arg;
 
 	if (arg && *arg) dir = strdup2(arg);
 	else if (!(dir = inputstr(mes, 1, -1, NULL, 1))) return(NULL);
-	else if (!*(dir = evalpath(dir, 1))) {
+	else if (!*(dir = evalpath(dir, 0))) {
 		dir = realloc2(dir, 2);
 		dir[0] = '.';
 		dir[1] = '\0';
@@ -246,7 +241,7 @@ char *file, *dest;
 struct stat *stp1, *stp2;
 {
 	char *cp, *tmp, *str[MAXCOPYITEM];
-	int i, n, val[MAXCOPYITEM];
+	int i, n, ch, val[MAXCOPYITEM];
 #ifndef	_NOEXTRACOPY
 	char path[MAXPATHLEN];
 #endif
@@ -261,9 +256,8 @@ struct stat *stp1, *stp2;
 		warning(-1, dest);
 		return(-1);
 	}
-	if ((stp2 -> st_mode & S_IFMT) != S_IFDIR)
-		n = (copypolicy & ~FLAG_SAMEDIR);
-	else if ((stp1 -> st_mode & S_IFMT) != S_IFDIR) n = 2;
+	if (!s_isdir(stp2)) n = (copypolicy & ~FLAG_SAMEDIR);
+	else if (!s_isdir(stp1)) n = 2;
 #ifdef	_NOEXTRACOPY
 	else return(1);
 #else
@@ -298,15 +292,16 @@ struct stat *stp1, *stp2;
 			val[4] = 5;
 #endif
 			i = 1;
-			if (selectstr(&i, MAXCOPYITEM, 0, str, val) == K_ESC)
-				return(-1);
+			ch = selectstr(&i, MAXCOPYITEM, 0, str, val);
+			if (ch == K_ESC) return(-1);
+			else if (ch != K_CR) return(-2);
 #ifndef	_NOEXTRACOPY
 			if (i == 5) {
 				str[0] = UPDAT_K;
 				str[1] = OVERW_K;
 				val[0] = 5;
 				val[1] = 6;
-				if (selectstr(&i, 2, 0, str, val) == K_ESC)
+				if (selectstr(&i, 2, 0, str, val) != K_CR)
 					continue;
 # ifndef	_NODOSDRIVE
 				dupdestdrive = destdrive;
@@ -348,9 +343,7 @@ struct stat *stp1, *stp2;
 					warning(-1, dest);
 					return(-1);
 				}
-				if ((stp1 -> st_mode & S_IFMT) == S_IFDIR
-				&& (stp2 -> st_mode & S_IFMT) == S_IFDIR)
-					return(1);
+				if (s_isdir(stp1) && s_isdir(stp2)) return(1);
 				putterm(t_bell);
 				break;
 			case 3:
@@ -396,7 +389,7 @@ char *path;
 int mode;
 {
 	char *cp, *str[4];
-	int len, val[4];
+	int ch, len, val[4];
 #if	MSDOS
 
 	if (Xaccess(path, mode) >= 0) return(0);
@@ -416,7 +409,7 @@ int mode;
 			warning(-1, path);
 			return(-1);
 		}
-		if ((st.st_mode & S_IFMT) == S_IFLNK) return(0);
+		if (s_islnk(&st)) return(0);
 
 		if (!(tmp = strrdelim(path, 0))) strcpy(dir, ".");
 		else if (tmp == path) strcpy(dir, _SS_);
@@ -469,8 +462,9 @@ int mode;
 	val[1] = -1;
 	val[2] = 2;
 	val[3] = 1;
-	if (selectstr(&removepolicy, 4, 0, str, val) == K_ESC)
-		removepolicy = -1;
+	ch = selectstr(&removepolicy, 4, 0, str, val);
+	if (ch == K_ESC) removepolicy = -1;
+	else if (ch != K_CR) removepolicy = -2;
 	return((removepolicy > 0) ? removepolicy - 2 : removepolicy);
 }
 
@@ -479,8 +473,10 @@ char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
+	int n;
 
-	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
+	if ((n = checkdupl(path, dest, &st1, &st2)) < 0)
+		return((n < -1) ? -2 : 1);
 	if (safecpfile(path, dest, &st1, &st2) < 0) return(-1);
 	return(0);
 }
@@ -490,9 +486,11 @@ char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
+	int n;
 
-	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
-	if (checkrmv(path, W_OK) < 0) return(1);
+	if ((n = checkdupl(path, dest, &st1, &st2)) < 0
+	|| (n = checkrmv(path, W_OK)) < 0)
+		return((n < -1) ? -2 : 1);
 	if (safemvfile(path, dest, &st1, &st2) < 0) return(-1);
 	return(0);
 }
@@ -520,6 +518,7 @@ char *path;
 			destatime = st1.st_atime;
 			break;
 		case -1:
+		case -2:
 		/* Abandon copy */
 			return(-2);
 /*NOTREACHED*/
@@ -530,8 +529,12 @@ char *path;
 			destatime = st2.st_atime;
 			break;
 	}
+
 	if (destdir) free(destdir);
-	destdir = strdup2(&(dest[strlen(destpath) + 1]));
+	destdir = &(dest[strlen(destpath)]);
+	while (*destdir == _SC_) destdir++;
+	destdir = strdup2(destdir);
+
 	return(0);
 }
 
@@ -552,7 +555,7 @@ char *path;
 		st.st_atime = destatime;
 	}
 # ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 # endif
 	if (touchfile(dest, &st) < 0) return(-1);
 #endif	/* _USEDOSCOPY || !_NOEXTRACOPY */
@@ -565,6 +568,7 @@ char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
+	int n;
 
 	if (getdestpath(path, dest, &st) < 0) return(-2);
 	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
@@ -572,10 +576,11 @@ char *path;
 		st.st_atime = destatime;
 	}
 # ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 # endif
 	if (touchfile(dest, &st) < 0) return(-1);
-	if (checkrmv(path, R_OK | W_OK | X_OK) < 0) return(1);
+	if ((n = checkrmv(path, R_OK | W_OK | X_OK) < 0))
+		return((n < -1) ? -2 : 1);
 	return(Xrmdir(path));
 }
 #endif	/* !_NOEXTRACOPY */
@@ -583,14 +588,19 @@ char *path;
 int rmvfile(path)
 char *path;
 {
-	if (checkrmv(path, W_OK) < 0) return(1);
+	int n;
+
+	if ((n = checkrmv(path, W_OK)) < 0) return((n < -1) ? -2 : 1);
 	return(Xunlink(path));
 }
 
 int rmvdir(path)
 char *path;
 {
-	if (checkrmv(path, R_OK | W_OK | X_OK) < 0) return(1);
+	int n;
+
+	if ((n = checkrmv(path, R_OK | W_OK | X_OK)) < 0)
+		return((n < -1) ? -2 : 1);
 	return(Xrmdir(path));
 }
 
@@ -809,6 +819,9 @@ int flag;
 				}
 				else if (x < WMODE - 2) x++;
 				break;
+			case K_BS:
+				if (y < WMODELINE) break;
+/*FALLTHRU*/
 			case K_LEFT:
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
@@ -839,7 +852,7 @@ int flag;
 			case ' ':
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
-					attr.flags ^= flaglist[x];
+					attr.flags ^= fflaglist[x];
 					locate(xx, yy + y + 2);
 					putflags(buf, attr.flags);
 					cputs2(buf);
@@ -902,18 +915,26 @@ int flag;
 	|| tm -> tm_hour > 23 || tm -> tm_min > 59 || tm -> tm_sec > 59)
 		return(-1);
 
-	attrmode = (flag & 1) ? attr.mode : 0xffff;
+	attrmode = attr.mode;
 #ifdef	HAVEFLAGS
-	attrflags = (flag & 1) ? attr.flags : 0xffffffff;
+	attrflags = attr.flags;
 #endif
-	attrtime = (flag & 2) ? timelocal2(tm) : (time_t)-1;
+	attrtime = timelocal2(tm);
 	if (flag == 3) {
-		if (attrmode == namep -> st_mode) attrmode = 0xffff;
+		if (attrmode == namep -> st_mode) attrmode = (u_short)-1;
 #ifdef	HAVEFLAGS
-		if (attrflags == namep -> st_flags) attrflags = 0xffffffff;
+		if (attrflags == namep -> st_flags) attrflags = (u_long)-1;
 #endif
 		if (attrtime == namep -> st_mtim) attrtime = (time_t)-1;
 	}
+	else {
+		if (!(flag & 1)) attrmode = (u_short)-1;
+#ifdef	HAVEFLAGS
+		attrflags = (u_long)-1;
+#endif
+		if (!(flag & 2)) attrtime = (time_t)-1;
+	}
+
 	return(1);
 }
 
@@ -947,7 +968,7 @@ char *endmes;
 
 	if (mark <= 0) {
 		i = (*func)(fnodospath(path, filepos));
-		if (i < 0) warning(-1, filelist[filepos].name);
+		if (i == -1) warning(-1, filelist[filepos].name);
 		else if (!i) ret++;
 		return(ret);
 	}
@@ -964,6 +985,10 @@ char *endmes;
 		locate(win_x, win_y);
 		tflush();
 		i = (*func)(fnodospath(path, filepos));
+		if (i < -1) {
+			endmes = NULL;
+			break;
+		}
 		if (i < 0) warning(-1, filelist[filepos].name);
 		else if (!i && ret == filepos) ret++;
 
@@ -994,8 +1019,7 @@ int *maxp, depth;
 		if (isdotdir(dp -> d_name)) continue;
 		strcpy(cp, dp -> d_name);
 
-		if (Xlstat(dir, &st) < 0 || (st.st_mode & S_IFMT) != S_IFDIR)
-			continue;
+		if (Xlstat(dir, &st) < 0 || !s_isdir(&st)) continue;
 		dirlist = b_realloc(dirlist, *maxp, char *);
 		dirlist[(*maxp)++] = strdup2(dir);
 		if (!(d = depth)) /*EMPTY*/;
@@ -1093,7 +1117,7 @@ int verbose;
 			strcpy(fname, dp -> d_name);
 
 			if (Xlstat(path, &st) < 0) warning(-1, path);
-			else if ((st.st_mode & S_IFMT) == S_IFDIR) continue;
+			else if (s_isdir(&st)) continue;
 			else if ((ret = (*funcf)(path)) < 0) {
 				if (ret < -1) break;
 				warning(-1, path);
@@ -1105,7 +1129,10 @@ int verbose;
 
 	destmtime = dupmtime;
 	destatime = dupatime;
-	if (funcd2 && (ret = (*funcd2)(dir)) == -1) warning(-1, dir);
+	if (funcd2 && (ret = (*funcd2)(dir)) < 0) {
+		if (ret < -1) return(ret);
+		warning(-1, dir);
+	}
 	if (endmes) warning(0, endmes);
 	return(ret);
 }
@@ -1270,7 +1297,7 @@ char *path;
 	if (getdestpath(path, dest, &st) < 0) return(0);
 	if (safecpfile(path, dest, &st, NULL) < 0) return(-1);
 #ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 #endif
 	return(touchfile(dest, &st));
 }
@@ -1280,21 +1307,23 @@ char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
+	int mode;
 
 	if (isdotdir(path)) return(0);
 	if (getdestpath(path, dest, &st) < 0) return(0);
+	mode = (st.st_mode & 0777);
 #if	MSDOS
-	while (Xmkdir(dest, (st.st_mode & 0777) | S_IWRITE) < 0)
-#else
-	while (Xmkdir(dest, st.st_mode & 0777) < 0)
+	mode |= S_IWRITE;
 #endif
-	{
-		if (errno != EEXIST) return(-1);
-		if (Xstat(dest, &st) < 0) return(-1);
-		if ((st.st_mode & S_IFMT) == S_IFDIR) break;
-		if (Xunlink(dest) < 0) return(-1);
+	if (Xmkdir(dest, mode) >= 0) return(0);
+	if (errno != EEXIST) return(-1);
+	if (stat2(dest, &st) >= 0) {
+		if (s_isdir(&st)) return(0);
+		if (Xunlink(dest) >= 0 && Xmkdir(dest, mode) >= 0) return(0);
 	}
-	return(0);
+
+	errno = EEXIST;
+	return(-1);
 }
 
 static int forcetouchdir(path)
@@ -1306,7 +1335,7 @@ char *path;
 	if (isdotdir(path)) return(0);
 	if (getdestpath(path, dest, &st) < 0) return(-2);
 #ifdef	HAVEFLAGS
-	st.st_flags = 0xffffffff;
+	st.st_flags = (u_long)-1;
 #endif
 	return(touchfile(dest, &st));
 }
