@@ -14,9 +14,6 @@
 #include "system.h"
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int shutdrv __P_((int));
-#endif
 #if	MSDOS
 extern char *unixgetcurdir __P_((char *, int));
 #endif
@@ -28,8 +25,6 @@ extern int tradlayout;
 #endif
 extern int sizeinfo;
 extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern namelist filestack[];
 extern char fullpath[];
 extern char typesymlist[];
@@ -61,15 +56,15 @@ static char *ign_lha[] = {
 	"  Name          Original *",
 	"--------------*",
 # if	defined (FAKEMETA) || !defined (BSPATHDELIM)
-	"* files * ???.?% ??-??-?? ??:??:??",
+	"* files * ???.?% ?\077-?\077-?? ??:??:??",	/* avoid trigraph */
 # else
-	"* files * ???.?%% ??-??-?? ??:??:??",
+	"* files * ???.?%% ?\077-?\077-?? ??:??:??",	/* avoid trigraph */
 # endif
 	"",
 #else	/* !MSDOS */
 	" PERMSSN * UID*GID *",
 	"----------*",
-	" Total * file* ???.?%*",
+	" Total * file* ???.*%*",
 #endif	/* !MSDOS */
 	NULL
 };
@@ -148,7 +143,7 @@ static char *form_tar[] = {
 		{0, 0, 0, 0, 0, 0, 0, 0, 0}, \
 		{255, 255, 255}, 1
 #define	LINESEP	'\t'
-#define	isarchbr(l)	(((l) -> topskip) < 255)
+#define	isarchbr(l)	(((l) -> topskip) < (u_char)255)
 #endif	/* FD < 2 */
 
 #define	iswhitespace(c)	((c) == ' ' || (c) == '\t')
@@ -193,6 +188,8 @@ static int NEAR parsearchive __P_((FILE *, launchtable *, namelist *, int *));
 static VOID NEAR unpackerror __P_((VOID_A));
 static int NEAR readarchive __P_((char *, launchtable *, int));
 static char *NEAR searcharcdir __P_((char *, int));
+static char *NEAR archoutdir __P_((VOID_A));
+static int NEAR undertmp __P_((char *));
 #ifdef	_NODOSDRIVE
 static char *NEAR genfullpath __P_((char *, char *, char *));
 #else
@@ -373,12 +370,12 @@ VOID escapearch(VOID_A)
 static VOID NEAR poparchdupl(VOID_A)
 #endif	/* !_NOBROWSE */
 {
-	winvartable *old;
-	char *cp, *dir;
-	int i;
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
+	winvartable *old;
+	char *cp, *dir;
+	int i;
 
 	if (!(old = archduplp)) return;
 
@@ -461,9 +458,7 @@ char *buf;
 	int i, c, len;
 	u_int mode;
 
-	if (!*buf) return(0);
 	len = strlen(buf);
-
 	if (len < 9) {
 		mode = (tmp -> st_mode & S_IFMT) | S_IREAD_ALL | S_IWRITE_ALL;
 		while (--len >= 0) {
@@ -551,7 +546,7 @@ char *buf;
 	}
 
 	tmp -> st_mode = mode;
-	return(1);
+	return(len);
 }
 
 #if	FD >= 2
@@ -595,12 +590,14 @@ namelist *tmp;
 char *line, *form;
 int skip;
 {
-	struct tm tm;
-#ifndef	NOUID
+# ifndef	NOUID
 	uidtable *up;
 	gidtable *gp;
-#endif
-	long n;
+	uid_t uid;
+	gid_t gid;
+# endif
+	struct tm tm;
+	off_t n;
 	int i, ch, l, len, hit, err, err2, score;
 	char *cp, *s, *buf, *rawbuf;
 
@@ -609,12 +606,12 @@ int skip;
 	tmp -> name = NULL;
 	tmp -> st_mode = 0644;
 	tmp -> st_nlink = 1;
-#ifndef	NOUID
+# ifndef	NOUID
 	tmp -> st_uid = (uid_t)-1;
 	tmp -> st_gid = (gid_t)-1;
 	tmp -> linkname = NULL;
-#endif
-	tmp -> st_size = 0L;
+# endif
+	tmp -> st_size = (off_t)0;
 	tmp -> flags = 0;
 	tmp -> tmpflags = F_STAT;
 	tm.tm_year = tm.tm_mon = tm.tm_mday = -1;
@@ -646,10 +643,10 @@ int skip;
 			for (i = 0; line[i]; i++) if (line[i] == '\n') break;
 			len = i;
 		}
-		else if (!(cp = evalnumeric(form, &n, 1))) len = -1;
+		else if (!(cp = sscanf2(form, "%+d", &len))) len = -1;
 		else {
 			form = cp;
-			for (i = 0; i < n; i++) if (!line[i]) break;
+			for (i = 0; i < len; i++) if (!line[i]) break;
 			len = i;
 			ch = -1;
 		}
@@ -663,9 +660,9 @@ int skip;
 			form = strchr2(s, '}');
 			if (!form || form <= s) {
 				if (tmp -> name) free(tmp -> name);
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 				if (tmp -> linkname) free(tmp -> linkname);
-#endif
+# endif
 				return(-1);
 			}
 			l = form++ - s;
@@ -700,25 +697,27 @@ int skip;
 				hit++;
 				break;
 			case 'u':
-#ifndef	NOUID
-				if ((cp = evalnumeric(buf, &n, -1)) && !*cp)
-					tmp -> st_uid = n;
+# ifndef	NOUID
+				cp = sscanf2(buf, "%-*d%$",
+					sizeof(uid_t), &uid);
+				if (cp) tmp -> st_uid = uid;
 				else if ((up = finduid(0, buf)))
 					tmp -> st_uid = up -> uid;
-#endif
+# endif
 				hit++;
 				break;
 			case 'g':
-#ifndef	NOUID
-				if ((cp = evalnumeric(buf, &n, -1)) && !*cp)
-					tmp -> st_gid = n;
+# ifndef	NOUID
+				cp = sscanf2(buf, "%-*d%$",
+					sizeof(gid_t), &gid);
+				if (cp) tmp -> st_gid = gid;
 				else if ((gp = findgid(0, buf)))
 					tmp -> st_gid = gp -> gid;
-#endif
+# endif
 				hit++;
 				break;
 			case 's':
-				if (!(cp = evalnumeric(buf, &n, 0))) break;
+				if (!(cp = sscanf2(buf, "%qd", &n))) break;
 				if (*cp) err2++;
 				tmp -> st_size = n;
 				hit++;
@@ -763,43 +762,36 @@ int skip;
 				hit++;
 				break;
 			case 't':
-				if (!(cp = evalnumeric(buf, &n, 0))
-				|| n > 23) {
-					tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
+				if (!(cp = sscanf2(buf, "%d", &i)) || i > 23)
 					break;
-				}
-				tm.tm_hour = n;
+				tm.tm_hour = i;
 				hit++;
-				if (*cp != ':'
-				|| !(cp = evalnumeric(++cp, &n, 0))
-				|| n > 59) {
+				if (!(cp = sscanf2(cp, ":%d", &i)) || i > 59) {
 					tm.tm_min = tm.tm_sec = 0;
 					err2++;
 					break;
 				}
-				tm.tm_min = n;
-				if (*cp != ':'
-				|| !(cp = evalnumeric(++cp, &n, 0))
-				|| n > 59) {
+				tm.tm_min = i;
+				if (!sscanf2(cp, ":%d%$", &i) || i > 59) {
 					tm.tm_sec = 0;
 					break;
 				}
-				tm.tm_sec = n;
+				tm.tm_sec = i;
 				break;
 			case 'f':
 				for (i = 0; i < len; i++) {
-#if	MSDOS
-# ifdef	BSPATHDELIM
+# if	MSDOS
+#  ifdef	BSPATHDELIM
 					if (rawbuf[i] == '/') rawbuf[i] = _SC_;
-# else
+#  else
 					if (rawbuf[i] == '\\')
 						rawbuf[i] = _SC_;
 					if (iskanji1(rawbuf, i)) {
 						i++;
 						continue;
 					}
-# endif
-#endif	/* MSDOS */
+#  endif
+# endif	/* MSDOS */
 					if (!iswhitespace(rawbuf[i])) continue;
 					cp = &(rawbuf[i]);
 					for (cp++; iswhitespace(*cp); cp++);
@@ -807,10 +799,10 @@ int skip;
 						if (!ch) i = len;
 						break;
 					}
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 					if (cp[0] == '-' && cp[1] == '>')
 						break;
-#endif
+# endif
 				}
 				cp = &(line[i]);
 				if (isdelim(rawbuf, i - 1)) {
@@ -822,7 +814,7 @@ int skip;
 				tmp -> name = readfname(rawbuf, i);
 				hit++;
 				err = 0;
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 				while (iswhitespace(*cp)) cp++;
 				if (ch && cp >= &(line[len])) break;
 				if (cp[0] != '-' || cp[1] != '>') break;
@@ -836,43 +828,46 @@ int skip;
 					i = &(cp[i]) - line;
 					if (i > len) len = i;
 				}
-#endif
+# endif
 				break;
 			case 'x':
 				hit++;
 				err = 0;
 				break;
 			default:
-				free(buf);
-				free(rawbuf);
-				if (tmp -> name) free(tmp -> name);
-#ifndef	NOSYMLINK
-				if (tmp -> linkname) free(tmp -> linkname);
-#endif
-				return(-1);
-/*NOTREACHED*/
+				hit = -1;
 				break;
 		}
+
+		free(buf);
+		free(rawbuf);
+
+		if (hit < 0) {
+			if (tmp -> name) free(tmp -> name);
+# ifndef	NOSYMLINK
+			if (tmp -> linkname) free(tmp -> linkname);
+# endif
+			return(hit);
+		}
+
 		if (!hit) score += 5;
 		else if (hit <= err2) score += err2;
 		score += err;
-		free(buf);
-		free(rawbuf);
 		line += len;
 		if (!ch) while (iswhitespace(*line)) line++;
 	}
 
 	if (score >= MAXSCORE || !(tmp -> name) || !*(tmp -> name)) {
 		if (tmp -> name) free(tmp -> name);
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 		if (tmp -> linkname) free(tmp -> linkname);
-#endif
+# endif
 		return(MAXSCORE);
 	}
 
 	if (!(tmp -> st_mode & S_IFMT)) tmp -> st_mode |= S_IFREG;
-	if ((tmp -> st_mode & S_IFMT) == S_IFDIR) tmp -> flags |= F_ISDIR;
-	else if ((tmp -> st_mode & S_IFMT) == S_IFLNK) tmp -> flags |= F_ISLNK;
+	if (s_isdir(tmp)) tmp -> flags |= F_ISDIR;
+	else if (s_islnk(tmp)) tmp -> flags |= F_ISLNK;
 	if (tm.tm_year < 0) {
 		tm.tm_year = today[0];
 		if (tm.tm_mon < 0 || tm.tm_mday < 0) tm.tm_year = 1970 - 1900;
@@ -890,11 +885,11 @@ int skip;
 
 	tmp -> st_mtim = timelocal2(&tm);
 	tmp -> flags |=
-#ifdef	NOUID
+# ifdef	NOUID
 		logical_access(tmp -> st_mode);
-#else
+# else
 		logical_access(tmp -> st_mode, tmp -> st_uid, tmp -> st_gid);
-#endif
+# endif
 
 	return(score);
 }
@@ -917,12 +912,12 @@ int field, *eolp;
 			j++;
 			sp = (j < MAXLAUNCHSEP) ? (int)sep[j] : 255;
 			for (; sp == 255 || i < sp; i++)
-				if (!strchr(" \t", line[i])) break;
+				if (!iswhitespace(line[i])) break;
 			if (f < 0) f = 1;
 			else f++;
 			s = 0;
 		}
-		else if (strchr(" \t", line[i])) s = 1;
+		else if (iswhitespace(line[i])) s = 1;
 		else if (s) {
 			f++;
 			s = 0;
@@ -934,7 +929,7 @@ int field, *eolp;
 			if (eolp) {
 				for (j = i; line[j]; j++) {
 					if ((sp < 255 && j >= sp)
-					|| strchr(" \t", line[j])) break;
+					|| iswhitespace(line[j])) break;
 				}
 				*eolp = j;
 			}
@@ -987,12 +982,14 @@ char *line;
 launchtable *list;
 int max;
 {
-	struct tm tm;
-#ifndef	NOUID
+# ifndef	NOUID
 	uidtable *up;
 	gidtable *gp;
-#endif
-	long n;
+	uid_t uid;
+	gid_t gid;
+# endif
+	struct tm tm;
+	off_t n;
 	int i, skip;
 	char *cp, *buf;
 
@@ -1004,24 +1001,24 @@ int max;
 	tmp -> tmpflags = F_STAT;
 	tmp -> st_mode = 0;
 	tmp -> st_nlink = 1;
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 	tmp -> linkname = NULL;
-#endif
+# endif
 	getfield(buf, line, skip, list, F_NAME);
 	if (!*buf) {
 		free(buf);
 		return(-1);
 	}
-#if	MSDOS
-# ifdef	BSPATHDELIM
+# if	MSDOS
+#  ifdef	BSPATHDELIM
 	for (i = 0; buf[i]; i++) if (buf[i] == '/') buf[i] = _SC_;
-# else
+#  else
 	for (i = 0; buf[i]; i++) {
 		if (buf[i] == '\\') buf[i] = _SC_;
 		if (iskanji1(buf, i)) i++;
 	}
-# endif
-#endif	/* MSDOS */
+#  endif
+# endif	/* MSDOS */
 	if (isdelim(buf, i = (int)strlen(buf) - 1)) {
 		if (i > 0) buf[i] = '\0';
 		tmp -> st_mode |= S_IFDIR;
@@ -1029,21 +1026,21 @@ int max;
 	tmp -> name = strdup2(buf);
 
 	getfield(buf, line, skip, list, F_MODE);
-	if (!readattr(tmp, buf)) tmp -> st_mode = 0644;
+	readattr(tmp, buf);
 	if (!(tmp -> st_mode & S_IFMT)) tmp -> st_mode |= S_IFREG;
-	if ((tmp -> st_mode & S_IFMT) == S_IFDIR) tmp -> flags |= F_ISDIR;
-	else if ((tmp -> st_mode & S_IFMT) == S_IFLNK) tmp -> flags |= F_ISLNK;
+	if (s_isdir(tmp)) tmp -> flags |= F_ISDIR;
+	else if (s_islnk(tmp)) tmp -> flags |= F_ISLNK;
 
-#ifndef	NOUID
+# ifndef	NOUID
 	getfield(buf, line, skip, list, F_UID);
-	if ((cp = evalnumeric(buf, &n, -1)) && !*cp) tmp -> st_uid = n;
+	if (sscanf2(buf, "%-*d%$", sizeof(uid_t), &uid)) tmp -> st_uid = uid;
 	else tmp -> st_uid = ((up = finduid(0, buf))) ? up -> uid : (uid_t)-1;
 	getfield(buf, line, skip, list, F_GID);
-	if ((cp = evalnumeric(buf, &n, -1)) && !*cp) tmp -> st_gid = n;
+	if (sscanf2(buf, "%-*d%$", sizeof(gid_t), &gid)) tmp -> st_gid = gid;
 	else tmp -> st_gid = ((gp = findgid(0, buf))) ? gp -> gid : (gid_t)-1;
-#endif
+# endif
 	getfield(buf, line, skip, list, F_SIZE);
-	tmp -> st_size = ((cp = evalnumeric(buf, &n, 0)) && !*cp) ? n : 0L;
+	tmp -> st_size = (sscanf2(buf, "%qd%$", &n)) ? n : (off_t)0;
 
 	getfield(buf, line, skip, list, F_MON);
 	if (!strncmp(buf, "Jan", 3)) tm.tm_mon = 0;
@@ -1064,7 +1061,7 @@ int max;
 	tm.tm_mday = ((i = atoi2(buf)) >= 1 && i < 31) ? i : 1;
 	getfield(buf, line, skip, list, F_YEAR);
 	if (!*buf) tm.tm_year = 1970;
-	else if ((cp = evalnumeric(buf, &n, 0)) && !*cp) tm.tm_year = n;
+	else if ((i = atoi2(buf)) >= 0) tm.tm_year = i;
 	else if (list -> field[F_YEAR] == list -> field[F_TIME]
 	&& strchr(buf, ':')) {
 		tm.tm_year = today[0];
@@ -1078,27 +1075,26 @@ int max;
 	tm.tm_year -= 1900;
 
 	getfield(buf, line, skip, list, F_TIME);
-	if (!(cp = evalnumeric(buf, &n, 0)) || n > 23)
+	if (!(cp = sscanf2(buf, "%d", &i)) || i > 23)
 		tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
 	else {
-		tm.tm_hour = n;
-		if (*cp != ':' || !(cp = evalnumeric(++cp, &n, 0)) || n > 59)
+		tm.tm_hour = i;
+		if (!(cp = sscanf2(cp, ":%d", &i)) || i > 59)
 			tm.tm_min = tm.tm_sec = 0;
 		else {
-			tm.tm_min = n;
-			if (*cp != ':' || (n = atoi2(++cp)) < 0 || n > 59)
-				tm.tm_sec = 0;
-			else tm.tm_sec = n;
+			tm.tm_min = i;
+			if (!sscanf2(cp, ":%d%$", &i) || i > 59) tm.tm_sec = 0;
+			else tm.tm_sec = i;
 		}
 	}
 
 	tmp -> st_mtim = timelocal2(&tm);
 	tmp -> flags |=
-#ifdef	NOUID
+# ifdef	NOUID
 		logical_access(tmp -> st_mode);
-#else
+# else
 		logical_access(tmp -> st_mode, tmp -> st_uid, tmp -> st_gid);
-#endif
+# endif
 
 	free(buf);
 	return(0);
@@ -1145,13 +1141,13 @@ char *file, *dir;
 
 		locate(TC_PATH, TL_PATH);
 		putterm(t_standout);
-#ifndef	_NOBROWSE
+# ifndef	_NOBROWSE
 		if (browselist) {
 			cputs2(TS_BROWSE);
 			len = TD_BROWSE;
 		}
 		else
-#endif
+# endif
 		{
 			cputs2(TS_ARCH);
 			len = TD_ARCH;
@@ -1247,15 +1243,14 @@ char *s1, *s2;
 static char *NEAR pseudodir(namep)
 namelist *namep;
 {
-	char *cp, *tmp;
+	char *cp, *next;
 	int i, len;
 	u_short ent;
 
-	cp = namep -> name;
-	while ((tmp = strdelim(cp, 0))) {
-		while (*(tmp + 1) == _SC_) tmp++;
-		if (!*(tmp + 1)) break;
-		len = tmp - (namep -> name);
+	for (cp = namep -> name; (next = strdelim(cp, 0)); cp = ++next) {
+		while (*(next + 1) == _SC_) next++;
+		if (!*(next + 1)) break;
+		len = next - namep -> name;
 		if (!len) len++;
 		for (i = 0; i < maxfile; i++) {
 			if (isdir(&(filelist[i]))
@@ -1267,21 +1262,22 @@ namelist *namep;
 			filelist[i].name = realloc2(filelist[i].name, len + 1);
 			strncpy2(filelist[i].name, namep -> name, len);
 		}
-		cp = tmp + 1;
 	}
+
 	if (isdir(namep) && !isdotdir(namep -> name))
 	for (i = 0; i < maxfile; i++) {
 		if (isdir(&(filelist[i]))
 		&& !dircmp(filelist[i].name, namep -> name)) {
-			tmp = filelist[i].name;
+			cp = filelist[i].name;
 			ent = filelist[i].ent;
 			memcpy((char *)&(filelist[i]), (char *)namep,
 				sizeof(namelist));
-			filelist[i].name = tmp;
+			filelist[i].name = cp;
 			filelist[i].ent = ent;
 			return(NULL);
 		}
 	}
+
 	return(namep -> name);
 }
 
@@ -1392,7 +1388,7 @@ int *linenop;
 	}
 
 #if	FD >= 2
-	if (!formlist) return(-3);
+	if (!formlist || !*formlist) return(-3);
 #endif
 	if (!(nline = countvar(lvar))) {
 		lvar = (char **)realloc2(lvar, (nline + 2) * sizeof(char *));
@@ -1615,11 +1611,11 @@ int argset;
 #ifndef	NOSYMLINK
 	filelist[0].linkname = NULL;
 #endif
-	filelist[0].flags = F_ISDIR;
+	filelist[0].flags = (F_ISEXE | F_ISRED | F_ISWRI | F_ISDIR);
 	filelist[0].tmpflags = F_STAT;
 	maxfile++;
 #ifdef	HAVEFLAGS
-	tmp.st_flags = 0;
+	tmp.st_flags = (u_long)0;
 #endif
 
 	no = 0;
@@ -1719,25 +1715,30 @@ reg_t *re;
 char *arcre;
 {
 	char *cp, *tmp;
-	int i, j, n, len;
+	int i, j, n, len, parent;
 
+	parent = (*archivedir) ? 0 : -1;
+	/* omit filelist[0] as pseudo ".." */
 	for (i = 1; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) {
-			memcpy((char *)&(filelist[0]),
-				(char *)&(arcflist[i]), sizeof(namelist));
-			filelist[0].name = arcflist[0].name;
-#ifndef	NOSYMLINK
-			filelist[0].linkname = NULL;
-#endif
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) {
+			parent = i;
 			continue;
 		}
-		if ((tmp = strdelim(cp, 0))) while (*(++tmp) == _SC_);
+
+		if (!(tmp = strdelim(cp, 0))) len = strlen(cp);
+		else {
+			len = (tmp == cp) ? 1 : tmp - cp;
+			while (*(++tmp) == _SC_);
+		}
+
+		if (parent <= 0 && len == 2 && cp[0] == '.' && cp[1] == '.')
+			parent = i;
 		if ((tmp && *tmp) || (re && !regexp_exec(re, cp, 1))) continue;
 		if (arcre) {
 			if (!(n = searcharc(arcre, arcflist, maxarcf, i)))
@@ -1755,10 +1756,20 @@ char *arcre;
 #ifndef	NOSYMLINK
 		filelist[maxfile].linkname = arcflist[i].linkname;
 #endif
-		if (isfile(&(filelist[maxfile])))
-			totalsize += getblock(filelist[maxfile].st_size);
 		maxfile++;
 	}
+
+	if (parent < 0) return;
+	for (i = 0; i < maxfile; i++)
+		if (!strcmp(filelist[i].name, "..")) return;
+	memmove((char *)&(filelist[1]), (char *)&(filelist[0]),
+		maxfile++ * sizeof(namelist));
+	memcpy((char *)&(filelist[0]),
+		(char *)&(arcflist[parent]), sizeof(namelist));
+	filelist[0].name = arcflist[0].name;
+#ifndef	NOSYMLINK
+	filelist[0].linkname = NULL;
+#endif
 }
 
 static char *NEAR searcharcdir(file, flen)
@@ -1769,14 +1780,14 @@ int flen;
 	int i, len;
 
 	errno = ENOENT;
-	for (i = 1; i < maxarcf; i++) {
+	for (i = 0; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) {
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) {
 			if (file) continue;
 		}
 		else {
@@ -1786,6 +1797,8 @@ int flen;
 				len = (tmp == cp) ? 1 : tmp - cp;
 				while (*(++tmp) == _SC_);
 			}
+			if (!*file && len == 2 && cp[0] == '.' && cp[1] == '.')
+				return(arcflist[0].name);
 			if ((tmp && *tmp) || len != flen
 			|| strnpathcmp(file, cp, flen))
 				continue;
@@ -1799,6 +1812,29 @@ int flen;
 	return(NULL);
 }
 
+static char *NEAR archoutdir(VOID_A)
+{
+	char *cp, *file;
+
+	if (!*archivedir) return((char *)-1);
+	if (!(file = searcharcdir(NULL, 0))) return(NULL);
+
+	cp = &(file[(int)strlen(file) - 1]);
+	while (cp > file && *cp == _SC_) cp--;
+#ifdef	BSPATHDELIM
+	if (onkanji1(file, cp - file)) cp++;
+#endif
+	cp = strrdelim2(file, cp);
+	if (!cp) *archivedir = '\0';
+	else {
+		if (cp == file) strcpy(archivedir, _SS_);
+		else strncpy2(archivedir, file, cp - file);
+		file = ++cp;
+	}
+
+	return(file);
+}
+
 char *archchdir(path)
 char *path;
 {
@@ -1807,11 +1843,11 @@ char *path;
 
 	if (findpattern) free(findpattern);
 	findpattern = NULL;
-	if (!path || !*path) path = "..";
 #ifndef	_NOBROWSE
 	if (browselist) {
 		int i, n, dupfilepos;
 
+		if (!path || !*path) path = "..";
 		if ((cp = strdelim(path, 0))) {
 			for (i = 1; cp[i]; i++) if (cp[i] != _SC_) break;
 			if (cp[i]) {
@@ -1854,41 +1890,28 @@ char *path;
 		return("..");
 	}
 #endif	/* !_NOBROWSE */
+
+	if (!path) return(archoutdir());
 	strcpy(duparcdir, archivedir);
 	do {
 		if (*path == _SC_) len = 1;
 		else if ((cp = strdelim(path, 0))) len = cp - path;
 		else len = strlen(path);
-		if (len != 2 || strncmp(path, "..", len)) {
-			if (!searcharcdir(path, len)) {
-				strcpy(archivedir, duparcdir);
-				errno = ENOENT;
-				return(NULL);
-			}
+
+		cp = path;
+		if (len == 2 && path[0] == '.' && path[1] == '.') cp = "";
+		if (searcharcdir(cp, len)) {
 			if (*(cp = archivedir)) cp = strcatdelim(archivedir);
 			strncpy2(cp, path, len);
 			file = "..";
 		}
-		else if (!*archivedir) return((char *)-1);
-		else {
-			if (!(file = searcharcdir(NULL, 0))) {
-				strcpy(archivedir, duparcdir);
-				errno = ENOENT;
-				return(NULL);
-			}
-			cp = file + (int)strlen(file) - 1;
-			while (cp > file && *cp == _SC_) cp--;
-#ifdef	BSPATHDELIM
-			if (onkanji1(file, cp - file)) cp++;
-#endif
-			cp = strrdelim2(file, cp);
-			if (!cp) *archivedir = '\0';
-			else {
-				if (cp == file) strcpy(archivedir, _SS_);
-				else strncpy2(archivedir, file, cp - file);
-				file = cp + 1;
-			}
+		else if (*cp || !(file = archoutdir())) {
+			strcpy(archivedir, duparcdir);
+			errno = ENOENT;
+			return(NULL);
 		}
+		else if (file == (char *)-1) break;
+
 		path += len;
 		while (*path == _SC_) path++;
 	} while (*path);
@@ -1903,17 +1926,17 @@ int flen, argc;
 char ***argvp;
 {
 	char *cp, *tmp, *new, *file, dir[MAXPATHLEN], duparcdir[MAXPATHLEN];
-	int i, len;
+	int i, len, parent;
 
-#ifdef	_USEDOSPATH
+# ifdef	_USEDOSPATH
 	if (_dospath(path)) return(argc);
-#endif
+# endif
 
 	strcpy(duparcdir, archivedir);
 	if (!(file = strrdelim(path, 0))) file = path;
-#ifndef	_NOBROWSE
+# ifndef	_NOBROWSE
 	else if (browselist) return(argc);
-#endif
+# endif
 	else {
 		strncpy2(dir, path, (file == path) ? 1 : file - path);
 		if (!(cp = archchdir(dir)) || cp == (char *)-1) {
@@ -1923,25 +1946,30 @@ char ***argvp;
 		flen -= ++file - path;
 	}
 
+	/* omit filelist[0] as pseudo ".." */
 	for (i = 1; i < maxarcf; i++) {
 		if (!*archivedir) len = 0;
 		else if (!(len = dirmatchlen(archivedir, arcflist[i].name)))
 			continue;
 
-		cp = arcflist[i].name + len;
-		if (len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
-		if (!arcflist[i].name[len]) continue;
+		cp = &(arcflist[i].name[len]);
+		if (*cp && len > 0 && (len > 1 || *archivedir != _SC_)) cp++;
+		if (!*cp) continue;
 
 		if (!(tmp = strdelim(cp, 0))) len = strlen(cp);
 		else {
 			len = (tmp == cp) ? 1 : tmp - cp;
 			while (*(++tmp) == _SC_);
 		}
-		if ((tmp && *tmp) || strnpathcmp(file, cp, flen)) continue;
+
+		parent = 0;
+		if (len == 2 && cp[0] == '.' && cp[1] == '.') parent++;
+		else if (tmp && *tmp) continue;
+		if (strnpathcmp(file, cp, flen)) continue;
 
 		new = malloc2(len + 1 + 1);
 		strncpy(new, cp, len);
-		if (isdir(&(arcflist[i]))) new[len++] = _SC_;
+		if (parent || isdir(&(arcflist[i]))) new[len++] = _SC_;
 		new[len] = '\0';
 		if (finddupl(new, argc, *argvp)) {
 			free(new);
@@ -1961,11 +1989,11 @@ int dolaunch(list, argset)
 launchtable *list;
 int argset;
 {
-	char *tmpdir;
-	int i;
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
+	char *tmpdir;
+	int i;
 
 	tmpdir = NULL;
 #ifndef	_NODOSDRIVE
@@ -2094,6 +2122,24 @@ int launcher(VOID_A)
 	return(dolaunch(&(launchlist[i]), 0));
 }
 
+static int NEAR undertmp(path)
+char *path;
+{
+	winvartable *wvp;
+	char *full, rpath[MAXPATHLEN], dir[MAXPATHLEN];
+
+	full = fullpath;
+	realpath2(path, rpath, 1);
+	for (wvp = archduplp; wvp; wvp = wvp -> v_archduplp) {
+		if (!(wvp -> v_fullpath)) continue;
+		realpath2(full, dir, 1);
+		if (underpath(rpath, dir, -1)) return(1);
+		full = wvp -> v_fullpath;
+	}
+
+	return(0);
+}
+
 #ifdef	_NODOSDRIVE
 static char *NEAR genfullpath(path, file, full)
 char *path, *file, *full;
@@ -2176,14 +2222,14 @@ char *path, **dirp, *full;
 int pack(arc)
 char *arc;
 {
-	winvartable *wvp;
-	reg_t *re;
-	char *tmpdir, *full, *duparchivefile, path[MAXPATHLEN];
-	int i, n, ret;
 #ifndef	_NODOSDRIVE
 	char *dest, *tmpdest;
 	int drive;
 #endif
+	winvartable *wvp;
+	reg_t *re;
+	char *tmpdir, *full, *duparchivefile, path[MAXPATHLEN];
+	int i, n, ret;
 
 	for (i = 0; i < maxarchive; i++) {
 		if (!archivelist[i].p_comm) continue;
@@ -2277,15 +2323,15 @@ char *arc, *dir;
 char *arg;
 int tr, ignorelist;
 {
-	reg_t *re;
-	char *cp, *tmpdir, path[MAXPATHLEN];
-	int i, n, ret;
 #ifndef	_NODOSDRIVE
 	winvartable *wvp;
 	namelist alist[1], *dupfilelist;
 	char *full, *dest, *tmpdest;
 	int dd, drive, dupmaxfile, dupfilepos;
 #endif
+	reg_t *re;
+	char *cp, *tmpdir, path[MAXPATHLEN];
+	int i, n, ret;
 
 #ifndef	_NOBROWSE
 	if (browselist) {
@@ -2309,7 +2355,7 @@ int tr, ignorelist;
 	if (i >= maxarchive) return(-1);
 
 	if (dir) strcpy(path, dir);
-	else {
+	else for (;;) {
 #ifndef	_NOTREE
 		if (tr) {
 # ifdef	_NODOSDRIVE
@@ -2324,7 +2370,7 @@ int tr, ignorelist;
 		{
 			if (arg && *arg) dir = strdup2(arg);
 			else dir = inputstr(UNPAC_K, 1, -1, NULL, 1);
-			dir = evalpath(dir, 1);
+			dir = evalpath(dir, 0);
 		}
 		if (!dir) return(0);
 		if (!*dir) strcpy(path, ".");
@@ -2336,6 +2382,12 @@ int tr, ignorelist;
 		}
 		free(dir);
 		dir = NULL;
+
+		if (!undertmp(path)) break;
+		locate(0, L_CMDLINE);
+		putterm(l_clear);
+		cprintf2("[%.*s]", n_column - 2, path);
+		if (yesno(UNPTM_K)) break;
 	}
 
 	ret = 1;
@@ -2356,17 +2408,17 @@ int tr, ignorelist;
 		dest = strdup2(path);
 		cp = tmpdest;
 	}
-#ifdef	FAKEUNINIT
+# ifdef	FAKEUNINIT
 	else dest = NULL;	/* fake for -Wuninitialized */
-#endif
+# endif
 
 	dupfilelist = filelist;
 	dupmaxfile = maxfile;
 	dupfilepos = filepos;
 	alist[0].name = arc;
-#ifndef	NOSYMLINK
+# ifndef	NOSYMLINK
 	alist[0].linkname = NULL;
-#endif
+# endif
 	alist[0].st_mode = 0700;
 	filelist = alist;
 	maxfile = 1;
@@ -2418,7 +2470,6 @@ int single;
 #ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 #endif
-
 	winvartable *wvp;
 	char *subdir, *tmpdir, path[MAXPATHLEN];
 	int i, ret, dupmark;
