@@ -74,6 +74,7 @@ extern int physical_path;
 
 #ifndef	_NOCOLOR
 static int NEAR getcolorid __P_((namelist *));
+static int NEAR biascolor __P_((int));
 static int NEAR getcolor __P_((int));
 #endif
 static VOID NEAR pathbar __P_((VOID_A));
@@ -90,6 +91,7 @@ static int NEAR putsize2 __P_((char *, namelist *, int));
 static int NEAR putfilename __P_((char *, namelist *, int));
 static VOID NEAR infobar __P_((VOID_A));
 static int NEAR calclocate __P_((int));
+static int NEAR calcfilepos __P_((namelist *, int, char *));
 #ifndef	_NOSPLITWIN
 static int NEAR listupwin __P_((char *));
 #endif
@@ -97,6 +99,8 @@ static int NEAR searchmove __P_((int, char *));
 #ifndef	_NOPRECEDE
 static VOID readstatus __P_((VOID_A));
 #endif
+static int NEAR readfilelist __P_((reg_t *, char *));
+static VOID NEAR getfilelist __P_((VOID_A));
 static int NEAR browsedir __P_((char *, char *));
 static char *NEAR initcwd __P_((char *, char *));
 
@@ -133,8 +137,8 @@ u_short typelist[] = {
 	S_IFDIR, S_IFBLK, S_IFCHR, S_IFLNK, S_IFSOCK, S_IFIFO
 };
 #ifdef	HAVEFLAGS
-char flagsymlist[] = "ANacuacu";
-u_long flaglist[] = {
+char fflagsymlist[] = "ANacuacu";
+u_long fflaglist[] = {
 	SF_ARCHIVED, UF_NODUMP,
 	SF_APPEND, SF_IMMUTABLE, SF_NOUNLINK,
 	UF_APPEND, UF_IMMUTABLE, UF_NOUNLINK
@@ -155,19 +159,19 @@ int win = 0;
 int calc_x = -1;
 int calc_y = -1;
 
-static u_short modelist[] = {
+static CONST u_short modelist[] = {
 	S_IFDIR, S_IFLNK, S_IFSOCK, S_IFIFO, S_IFBLK, S_IFCHR
 };
 #define	MAXMODELIST	(sizeof(modelist) / sizeof(u_short))
-static char suffixlist[] = {
+static CONST char suffixlist[] = {
 	'/', '@', '=', '|'
 };
 #define	MAXSUFFIXLIST	(sizeof(suffixlist) / sizeof(char))
 #ifndef	_NOCOLOR
-static u_char colorlist[] = {
+static CONST u_char colorlist[] = {
 	CL_DIR, CL_LINK, CL_SOCK, CL_FIFO, CL_BLOCK, CL_CHAR
 };
-static char defpalette[] = {
+static CONST char defpalette[] = {
 	ANSI_FG,	/* CL_NORM */
 	ANSI_BG,	/* CL_BACK */
 	ANSI_CYAN,	/* CL_DIR */
@@ -202,6 +206,16 @@ namelist *namep;
 	return(CL_NORM);
 }
 
+static int NEAR biascolor(color)
+int color;
+{
+	if (color == ANSI_FG)
+		color = (ansicolor == 3) ? ANSI_BLACK : ANSI_WHITE;
+	else if (color == ANSI_BG) color = (ansicolor == 2) ? ANSI_BLACK : -1;
+
+	return(color);
+}
+
 static int NEAR getcolor(id)
 int id;
 {
@@ -210,15 +224,11 @@ int id;
 	if (!ansipalette || id >= strlen(ansipalette)) color = defpalette[id];
 	else {
 		color = ansipalette[id];
-		if (color >= '0' && color <= '9') color -= '0';
+		if (isdigit2(color)) color -= '0';
 		else color = defpalette[id];
 	}
 
-	if (color == ANSI_FG)
-		color = (ansicolor == 3) ? ANSI_BLACK : ANSI_WHITE;
-	else if (color == ANSI_BG) color = (ansicolor == 2) ? ANSI_BLACK : -1;
-
-	return(color);
+	return(biascolor(color));
 }
 #endif	/* !_NOCOLOR */
 
@@ -361,7 +371,8 @@ static VOID NEAR statusbar(VOID_A)
 		if (haste) kanjiputs(OMIT_K);
 		else
 #endif
-		if (sorton & 7) {
+		if (!(sorton & 7)) kanjiputs(ORAW_K + 3);
+		else {
 			str[0] = ONAME_K;
 			str[1] = OEXT_K;
 			str[2] = OSIZE_K;
@@ -373,7 +384,6 @@ static VOID NEAR statusbar(VOID_A)
 			str[1] = ODEC_K;
 			cprintf2("(%k)", &(str[sorton / 8][3]));
 		}
-		else kanjiputs(ORAW_K + 3);
 	}
 
 	locate(C_FIND, L_STATUS);
@@ -452,7 +462,7 @@ static VOID NEAR sizebar(VOID_A)
 	off_t total, fre, bsize;
 
 	if (!hassizeinfo() || !*fullpath) return;
-	if (getinfofs(".", &total, &fre, &bsize) < 0) total = fre = -1L;
+	if (getinfofs(".", &total, &fre, &bsize) < 0) total = fre = (off_t)-1;
 
 #ifndef	_NOTRADLAYOUT
 	if (istradlayout()) {
@@ -560,8 +570,8 @@ u_long flags;
 {
 	int i;
 
-	for (i = 0; i < sizeof(flaglist) / sizeof(u_long); i++)
-		buf[i] = (flags & flaglist[i]) ? flagsymlist[i] : '-';
+	for (i = 0; i < sizeof(fflaglist) / sizeof(u_long); i++)
+		buf[i] = (flags & fflaglist[i]) ? fflagsymlist[i] : '-';
 	buf[i] = '\0';
 
 	return(i);
@@ -579,7 +589,7 @@ uid_t uid;
 	i = len = (iswellomit()) ? WOWNERMIN : WOWNER;
 	if (uid == (uid_t)-1) while (--i >= 0) buf[i] = '?';
 	else if ((up = finduid(uid, NULL))) strncpy3(buf, up -> name, &len, 0);
-	else snprintf2(buf, len + 1, "%-*d", len, uid);
+	else snprintf2(buf, len + 1, "%-*d", len, (int)uid);
 
 	return(len);
 }
@@ -594,7 +604,7 @@ gid_t gid;
 	i = len = (iswellomit()) ? WGROUPMIN : WGROUP;
 	if (gid == (gid_t)-1) while (--i >= 0) buf[i] = '?';
 	else if ((gp = findgid(gid, NULL))) strncpy3(buf, gp -> name, &len, 0);
-	else snprintf2(buf, len + 1, "%-*d", len, gid);
+	else snprintf2(buf, len + 1, "%-*d", len, (int)gid);
 
 	return(len);
 }
@@ -624,19 +634,19 @@ int width;
 	if (isdir(namep))
 		snprintf2(buf, width + 1, "%*.*s", width, width, "<DIR>");
 #if	MSDOS
-	else if ((namep -> st_mode & S_IFMT) == S_IFIFO)
+	else if (s_isfifo(namep))
 		snprintf2(buf, width + 1, "%*.*s", width, width, "<VOL>");
 #else	/* !MSDOS */
 # ifndef	_NODOSDRIVE
-	else if (dospath2("")
-	&& (namep -> st_mode & S_IFMT) == S_IFIFO)
+	else if (dospath2("") && s_isfifo(namep))
 		snprintf2(buf, width + 1, "%*.*s", width, width, "<VOL>");
 # endif
 	else if (isdev(namep))
-		snprintf2(buf, width + 1, "%<*d,%<*d",
-			width / 2, major((u_long)(namep -> st_size)),
+		snprintf2(buf, width + 1, "%<*lu,%<*lu",
+			width / 2,
+			(u_long)major((u_long)(namep -> st_size)),
 			width - (width / 2) - 1,
-			minor((u_long)(namep -> st_size)));
+			(u_long)minor((u_long)(namep -> st_size)));
 #endif	/* !MSDOS */
 	else snprintf2(buf, width + 1, "%<*qd", width, namep -> st_size);
 
@@ -674,7 +684,7 @@ int width;
 # ifndef	_NOARCHIVE
 	else if (archivefile) {
 		if (namep -> linkname)
-			strncpy3(buf, &(namep -> linkname[len]), &w, 0);
+			strncpy3(buf, namep -> linkname, &w, len);
 	}
 # endif
 	else {
@@ -683,7 +693,7 @@ int width;
 			tmp, width * 2 + len);
 		if (i >= 0) {
 			tmp[i] = '\0';
-			strncpy3(buf, &(tmp[len]), &w, 0);
+			strncpy3(buf, tmp, &w, len);
 		}
 		free(tmp);
 	}
@@ -697,7 +707,7 @@ static VOID NEAR infobar(VOID_A)
 	struct tm *tm;
 	int len;
 
-	if (!filelist || maxfile < 0) return;
+	if (!filelist || filepos < 0 || maxfile < 0) return;
 
 #ifndef	_NOTRADLAYOUT
 	if (istradlayout()) {
@@ -731,7 +741,7 @@ static VOID NEAR infobar(VOID_A)
 		buf[len++] = ' ';
 
 		snprintf2(&(buf[len]), WDATE + 1 + WTIME + 1 + 1,
-			"%02d-%02d-%02d %02d:%02d ",
+			"%02d-%02d-%02d %2d:%02d ",
 			tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday,
 			tm -> tm_hour, tm -> tm_min);
 		len += WDATE + 1 + WTIME + 1;
@@ -800,7 +810,7 @@ static VOID NEAR infobar(VOID_A)
 
 	if (!ishardomit()) {
 		snprintf2(&(buf[len]), 1 + WNLINK + 1 + 1, " %<*d ",
-			WNLINK, filelist[filepos].st_nlink);
+			WNLINK, (int)(filelist[filepos].st_nlink));
 		len += 1 + WNLINK + 1;
 
 #ifndef	NOUID
@@ -926,6 +936,7 @@ int no, isstandout;
 
 	if (list != filelist) {
 		len = strlen3(list[no].name);
+		if (isdisptyp(dispmode) && s_isdir(&(list[no]))) len++;
 		if (col > len) col = len;
 		if (width > len) width = len;
 	}
@@ -958,12 +969,13 @@ int no, isstandout;
 		for (j = 0; j < MAXMODELIST; j++)
 			if ((list[no].st_mode & S_IFMT) == modelist[j]) break;
 		if (j < MAXSUFFIXLIST) buf[i] = suffixlist[j];
-		else if ((list[no].st_mode & S_IFMT) == S_IFREG
+		else if (s_isreg(&(list[no]))
 		&& (list[no].st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
 			buf[i] = '*';
 	}
 #ifndef	_NOCOLOR
-	color = getcolor(getcolorid(&(list[no])));
+	if (list == filelist) color = getcolor(getcolorid(&(list[no])));
+	else color = biascolor(ANSI_FG);
 #endif
 	len = width;
 	width += col - wid;
@@ -1025,77 +1037,79 @@ int no, isstandout;
 	return(col);
 }
 
-int listupfile(list, max, def)
+static int NEAR calcfilepos(list, max, def)
 namelist *list;
 int max;
 char *def;
+{
+	int i;
+
+	if (def) for (i = 0; i < max; i++)
+		if (!strpathcmp(def, list[i].name)) return(i);
+
+	if (list == filelist && max > 1 && !strcmp(list[1].name, ".."))
+		return(1);
+	return(0);
+}
+
+int listupfile(list, max, def, isstandout)
+namelist *list;
+int max;
+char *def;
+int isstandout;
 {
 	char *cp;
-	int i, count, start, ret;
+	int i, n, pos;
 
 	for (i = 0; i < FILEPERROW; i++) {
 		locate(0, i + LFILETOP);
 		putterm(l_clear);
 	}
 
-	if (max <= 0) {
+	if (!list || max <= 0) {
 		i = (n_column / FILEPERLINE) - 2 - 1;
 		locate(1, LFILETOP);
 		putch2(' ');
-		putterm(t_standout);
+		if (isstandout) putterm(t_standout);
 		cp = NOFIL_K;
 		if (i <= strlen2(cp)) cp = "NoFiles";
 		cprintf2("%-*.*k", i, i, cp);
-		putterm(end_standout);
-		win_x = i + 2;
-		win_y = LFILETOP;
+		if (isstandout) putterm(end_standout);
+		win_x = calc_x = i + 2;
+		win_y = calc_y = LFILETOP;
 		return(0);
 	}
 
-	ret = -1;
-	if (!def) def = "..";
-
-	start = 0;
-	for (i = count = 0; i < max; i++, count++) {
-		if (count >= FILEPERPAGE) {
-			count = 0;
-			start = i;
-		}
-		if (!strpathcmp(def, list[i].name)) {
-			ret = i;
-			break;
-		}
-	}
-	if (ret < 0) {
-		start = 0;
-		ret = (max <= 1 || strcmp(list[1].name, "..")) ? 0 : 1;
-	}
+	pos = calcfilepos(list, max, def);
 
 	if (list == filelist) {
 #ifndef	_NOTRADLAYOUT
 		if (istradlayout()) {
 			locate(TC_PAGE + TW_PAGE, TL_SIZE);
-			cprintf2("%<*d", TD_PAGE, start / FILEPERPAGE + 1);
+			cprintf2("%<*d", TD_PAGE, pos / FILEPERPAGE + 1);
 		}
 		else
 #endif
 		{
 			locate(C_PAGE + W_PAGE, L_STATUS);
-			cprintf2("%<*d", D_PAGE, start / FILEPERPAGE + 1);
+			cprintf2("%<*d", D_PAGE, pos / FILEPERPAGE + 1);
 		}
 	}
 
-	for (i = start, count = 0; i < max; i++, count++) {
-		if (count >= FILEPERPAGE) break;
-		if (i != ret) putname(list, i, 0);
+	n = (pos / FILEPERPAGE) * FILEPERPAGE;
+	for (i = 0; i < FILEPERPAGE; i++, n++) {
+		if (n >= max) break;
+		if (!isstandout || n != pos) putname(list, n, 0);
 		else {
-			win_x = putname(list, i, 1) + 1;
-			win_x += calc_x;
+			calc_x += putname(list, n, 1) + 1;
+			win_x = calc_x;
 			win_y = calc_y;
 		}
 	}
 
-	return(ret);
+	calc_x = win_x;
+	calc_y = win_y;
+	return(pos);
 }
 
 #ifndef	_NOSPLITWIN
@@ -1120,16 +1134,12 @@ char *def;
 #endif
 	n = -1;
 	for (win = 0; win < windows; win++) {
-		if (!filelist) continue;
 		if (win == duplwin) {
-			n = listupfile(filelist, maxfile, def);
+			n = listupfile(filelist, maxfile, def, 1);
 			x = win_x;
 			y = win_y;
 		}
-		else {
-			listupfile(filelist, maxfile, filelist[filepos].name);
-			putname(filelist, filepos, -1);
-		}
+		else listupfile(filelist, maxfile, filelist[filepos].name, 0);
 	}
 	win = duplwin;
 
@@ -1192,14 +1202,14 @@ int old, funcstat;
 #endif
 	if (((funcstat & REWRITEMODE) >= RELIST)
 	|| old / FILEPERPAGE != filepos / FILEPERPAGE) {
-		filepos =
-			listupfile(filelist, maxfile, filelist[filepos].name);
+		filepos = listupfile(filelist, maxfile,
+			filelist[filepos].name, 1);
 		keyflush();
 	}
 	else if (((funcstat & REWRITEMODE) >= REWRITE) || old != filepos) {
 		if (old != filepos) putname(filelist, old, -1);
-		win_x = putname(filelist, filepos, 1) + 1;
-		win_x += calc_x;
+		calc_x += putname(filelist, filepos, 1) + 1;
+		win_x = calc_x;
 		win_y = calc_y;
 	}
 	infobar();
@@ -1208,7 +1218,7 @@ int old, funcstat;
 VOID rewritefile(all)
 int all;
 {
-	if (!filelist || maxfile < 0) return;
+	if (!filelist || filepos < 0 || maxfile < 0) return;
 	if (all > 0) {
 		title();
 		helpbar();
@@ -1233,7 +1243,8 @@ int all;
 #endif
 		if (filelist && filepos < maxfile) {
 #ifdef	_NOSPLITWIN
-			listupfile(filelist, maxfile, filelist[filepos].name);
+			listupfile(filelist, maxfile,
+				filelist[filepos].name, 1);
 #else
 			listupwin(filelist[filepos].name);
 #endif
@@ -1279,7 +1290,7 @@ char *buf;
 	}
 	else {
 		if (n == 1) buf[len = 0] = '\0';
-		if (isctl(ch) || ch >= K_MIN) isearch = 0;
+		if (iscntrl2(ch) || ch >= K_MIN) isearch = 0;
 		else if (len < MAXNAMLEN - 1) {
 			buf[len++] = ch;
 			buf[len] = '\0';
@@ -1345,8 +1356,8 @@ static VOID readstatus(VOID_A)
 	if (keywaitfunc != readstatus) return;
 	if (isfile(&(filelist[i])) && filelist[i].st_size) sizebar();
 	if (i == filepos) {
-		win_x = putname(filelist, i, 1) + 1;
-		win_x += calc_x;
+		calc_x += putname(filelist, i, 1) + 1;
+		win_x = calc_x;
 		win_y = calc_y;
 		infobar();
 	}
@@ -1355,104 +1366,99 @@ static VOID readstatus(VOID_A)
 	locate(win_x, win_y);
 	tflush();
 }
-#endif
+#endif	/* !_NOPRECEDE */
 
-static int NEAR browsedir(file, def)
-char *file, *def;
+static int NEAR readfilelist(re, arcre)
+reg_t *re;
+char *arcre;
 {
 #ifndef	_NOPRECEDE
 	char cwd[MAXPATHLEN];
-	int dupsorton;
 #endif
 	DIR *dirp;
 	struct dirent *dp;
-	reg_t *re;
-	char *cp, buf[MAXNAMLEN + 1];
-	int ch, i, no, old, funcstat;
+	char buf[MAXNAMLEN + 1];
+	int i, n;
 
+	n = 0;
 #ifndef	_NOPRECEDE
-	haste = 0;
-# ifdef	FAKEUNINIT
-	dupsorton = sorton;	/* fake for -Wuninitialized */
-# endif
-#endif
-#ifndef	_NOCOLOR
-	if (ansicolor) putterms(t_normal);
+	if (Xgetwd(cwd) && includepath(cwd, precedepath)) {
+		n = 1;
+		sorton = 0;
+	}
 #endif
 
-	mark = 0;
-	totalsize = marksize = (off_t)0;
-	chgorder = 0;
+	if (!(dirp = Xopendir("."))) {
+		lostcwd(NULL);
+		if (!(dirp = Xopendir("."))) error(".");
+	}
 
-	fnameofs = 0;
-	win_x = win_y = 0;
-	waitmes();
+	while ((dp = searchdir(dirp, re, arcre))) {
+		if (ishidedot(dispmode) && *(dp -> d_name) == '.'
+		&& !isdotdir(dp -> d_name)) continue;
+		strcpy(buf, dp -> d_name);
+		for (i = 0; i < stackdepth; i++)
+			if (!strcmp(buf, filestack[i].name)) break;
+		if (i < stackdepth) continue;
+
+		addlist();
+		filelist[maxfile].name = strdup2(buf);
+		filelist[maxfile].ent = maxfile;
+		filelist[maxfile].tmpflags = 0;
+#ifndef	_NOPRECEDE
+		if (n) {
+			if (isdotdir(dp -> d_name)) {
+				filelist[maxfile].flags = F_ISDIR;
+				filelist[maxfile].st_mode = S_IFDIR;
+			}
+		}
+		else
+#endif
+		if (getstatus(&(filelist[maxfile])) < 0) {
+			free(filelist[maxfile].name);
+			continue;
+		}
+
+		maxfile++;
+	}
+
+	Xclosedir(dirp);
+	return(n);
+}
+
+static VOID NEAR getfilelist(VOID_A)
+{
+	reg_t *re;
+	char *arcre;
 
 	re = NULL;
-	cp = NULL;
+	arcre = NULL;
 	if (findpattern) {
 #ifndef	_NOARCHIVE
-		if (*findpattern == '/') cp = findpattern + 1;
+		if (*findpattern == '/') arcre = findpattern + 1;
 		else
 #endif
 		re = regexp_init(findpattern, -1);
 	}
 
+	maxfile = 0;
 #ifndef	_NOARCHIVE
 	if (archivefile) {
-		maxfile = (*archivedir) ? 1 : 0;
 		blocksize = (off_t)1;
 		if (sorttype < 100) sorton = 0;
-		copyarcf(re, cp);
-		def = (*file) ? file : NULL;
+		copyarcf(re, arcre);
 	}
 	else
 #endif
 	{
-		maxfile = 0;
 		blocksize = getblocksize(".");
 		if (sorttype < 100) sorton = sorttype;
+		if (readfilelist(re, arcre)) {
 #ifndef	_NOPRECEDE
-		if (Xgetwd(cwd) && includepath(cwd, precedepath)) {
 			haste = 1;
-			dupsorton = sorton;
 			sorton = 0;
-		}
 #endif
-		if (!(dirp = Xopendir("."))) {
-			lostcwd(NULL);
-			if (!(dirp = Xopendir("."))) error(".");
 		}
-
-		while ((dp = searchdir(dirp, re, cp))) {
-			if (ishidedot(dispmode) && *(dp -> d_name) == '.'
-			&& !isdotdir(dp -> d_name)) continue;
-			strcpy(buf, dp -> d_name);
-			for (i = 0; i < stackdepth; i++)
-				if (!strcmp(buf, filestack[i].name)) break;
-			if (i < stackdepth) continue;
-			addlist();
-			filelist[maxfile].name = strdup2(buf);
-			filelist[maxfile].ent = maxfile;
-			filelist[maxfile].tmpflags = 0;
-#ifndef	_NOPRECEDE
-			if (haste) {
-				if (isdotdir(dp -> d_name)) {
-					filelist[maxfile].flags = F_ISDIR;
-					filelist[maxfile].st_mode = S_IFDIR;
-				}
-			}
-			else
-#endif
-			{
-				if (getstatus(&(filelist[maxfile])) < 0) {
-					free(filelist[maxfile].name);
-					continue;
-				}
-			}
-			maxfile++;
-		}
-		Xclosedir(dirp);
 	}
 	regexp_free(re);
 
@@ -1466,13 +1472,51 @@ char *file, *def;
 	maxstat = (haste) ? 0 : maxfile;
 #endif
 	if (sorton) qsort(filelist, maxfile, sizeof(namelist), cmplist);
+}
+
+static int NEAR browsedir(file, def)
+char *file, *def;
+{
+#ifndef	_NOPRECEDE
+	int dupsorton;
+#endif
+	char *cp, buf[MAXNAMLEN + 1];
+	int ch, i, no, old, funcstat;
+
+#ifndef	_NOPRECEDE
+	haste = 0;
+	dupsorton = sorton;
+#endif
+#ifndef	_NOCOLOR
+	if (ansicolor) putterms(t_normal);
+#endif
+
+	mark = 0;
+	totalsize = marksize = (off_t)0;
+	chgorder = 0;
+
+	fnameofs = 0;
+	win_x = win_y = 0;
+	waitmes();
+
+#ifndef	_NOARCHIVE
+	if (archivefile) def = (*file) ? file : "";
+#endif
+	getfilelist();
+
+	for (i = 0; i < maxfile; i++) {
+		if (!isfile(&(filelist[i])) || !filelist[i].st_size) continue;
+		totalsize += getblock(filelist[i].st_size);
+		if (ismark(&(filelist[i])))
+			marksize += getblock(filelist[i].st_size);
+	}
 
 	putterms(t_clear);
 	title();
 	helpbar();
 	rewritefile(-1);
 #ifdef	_NOSPLITWIN
-	old = filepos = listupfile(filelist, maxfile, def);
+	old = filepos = listupfile(filelist, maxfile, def, 1);
 #else
 	old = filepos = listupwin(def);
 #endif
@@ -1543,7 +1587,7 @@ char *file, *def;
 				maxfile += stackdepth;
 				stackdepth = 0;
 				filepos = listupfile(filelist, maxfile,
-					filelist[filepos].name);
+					filelist[filepos].name, 1);
 				stackbar();
 			}
 #ifndef	_NOARCHIVE
@@ -1614,8 +1658,9 @@ char *file, *def;
 		else if (no > 4) {
 			char *tmp;
 
-			tmp = (filepos < 0) ? ".." : filelist[filepos].name;
+			tmp = (filepos >= 0) ? filelist[filepos].name : NULL;
 			if (!(cp = archchdir(tmp))) {
+				if (!tmp) tmp = "..";
 				warning(-1, tmp);
 				strcpy(file, tmp);
 			}
@@ -1635,7 +1680,7 @@ char *file, *def;
 #endif	/* !_NOARCHIVE */
 	if (no >= 4) {
 		no -= 4;
-		strcpy(file, filelist[filepos].name);
+		strcpy(file, (maxfile) ? filelist[filepos].name : ".");
 	}
 #ifndef	_NOARCHIVE
 	if (!archivefile)
@@ -1645,6 +1690,7 @@ char *file, *def;
 		filelist[i].name = NULL;
 	}
 	maxfile = filepos = 0;
+
 #ifndef	_NOPRECEDE
 	if (haste && !sorton) sorton = dupsorton;
 #endif
@@ -1659,7 +1705,7 @@ char *path, *buf;
 
 	if (!path) return(NULL);
 
-	cp = evalpath(strdup2(path), 1);
+	cp = evalpath(strdup2(path), 0);
 #if	MSDOS
 	if (_dospath(cp)) {
 		if (setcurdrv(*cp, 0) >= 0 && !Xgetwd(fullpath))
@@ -1703,21 +1749,28 @@ char *path, *buf;
 	return(file);
 }
 
-VOID main_fd(path)
-char *path;
+VOID main_fd(pathlist)
+char **pathlist;
 {
 	char file[MAXNAMLEN + 1], prev[MAXNAMLEN + 1];
-	char *def;
-	int i, ischgdir;
+	char *def, *cwd;
+	int i, argc, ischgdir;
 
-	for (i = 0; i < MAXWINDOWS; i++) {
+	if (!pathlist) argc = 0;
+	else for (argc = 0; argc < MAXINVOKEARGS; argc++)
+		if (!pathlist[argc]) break;
+
+	cwd = getwd2();
+	def = NULL;
+	for (i = MAXWINDOWS - 1; i >= 0; i--) {
 #ifndef	_NOSPLITWIN
 		win = i;
 		winvar[i].v_fullpath = NULL;
 # ifndef	_NOARCHIVE
 		winvar[i].v_archivedir = NULL;
 # endif
-#endif
+#endif	/* !_NOSPLITWIN */
+
 #ifndef	_NOARCHIVE
 		archduplp = NULL;
 		archivefile = NULL;
@@ -1732,34 +1785,55 @@ char *path;
 		browselist = NULL;
 		browselevel = 0;
 # endif
-#endif
+#endif	/* !_NOARCHIVE */
+
 #ifndef	_NOTREE
 		treepath = NULL;
 #endif
 		findpattern = NULL;
 		filelist = NULL;
-		maxfile = maxent = filepos = sorton = dispmode = 0;
+		maxfile = maxent = filepos = 0;
+		sorton = sorttype % 100;
+		dispmode = displaymode;
+		curcolumns = defcolumns;
+
+		if (i >= argc) continue;
+
+		chdir2(cwd);
+		def = initcwd(pathlist[i], prev);
+
+#ifndef	_NOSPLITWIN
+		winvar[i].v_fullpath = strdup2(fullpath);
+		if (i) {
+			getfilelist();
+			sorton = sorttype % 100;
+			filepos = calcfilepos(filelist, maxfile, def);
+		}
+#endif
 	}
+
 #ifndef	_NOSPLITWIN
 	win = 0;
+	while (windows < argc) {
+		windows++;
+		if (FILEPERROW < WFILEMIN) {
+			windows--;
+			hideclock = 1;
+			warning(0, NOROW_K);
+			for (i = windows; i < argc; i++) shutwin(i);
+			break;
+		}
+	}
 #endif
 
 	strcpy(file, ".");
-	sorton = sorttype % 100;
-	dispmode = displaymode;
-	curcolumns = defcolumns;
-
-	def = initcwd(path, prev);
 	_chdir2(fullpath);
 
 	for (;;) {
 		if (!def && !strcmp(file, "..")) {
 			strcpy(prev, getbasename(fullpath));
 			if (*prev) def = prev;
-			else {
-				strcpy(file, ".");
-				def = NULL;
-			}
+			else strcpy(file, ".");
 		}
 
 		if (strcmp(file, ".")) {
@@ -1767,9 +1841,7 @@ char *path;
 			char buf[MAXPATHLEN];
 #endif
 
-			if (findpattern) free(findpattern);
-			findpattern = NULL;
-			if (chdir2(nodospath(buf, file)) < 0) {
+			if (chdir3(nodospath(buf, file), 1) < 0) {
 				hideclock = 1;
 				warning(-1, file);
 				strcpy(prev, file);
@@ -1784,7 +1856,10 @@ char *path;
 			ischgdir = browsedir(file, def);
 		} while (archivefile);
 #endif
-		if (ischgdir < 0) break;
+		if (ischgdir < 0) {
+			if (ischgdir > -2) chdir2(cwd);
+			break;
+		}
 		if (ischgdir) def = NULL;
 		else {
 			strcpy(prev, file);
@@ -1793,6 +1868,7 @@ char *path;
 		}
 	}
 
+	free(cwd);
 #ifdef	_NOSPLITWIN
 	if (filelist) free(filelist);
 	if (findpattern) free(findpattern);
