@@ -778,7 +778,7 @@ char *arg;
 
 	if (arg && *arg) path = strdup2(arg);
 	else if (!(path = inputstr(LOGD_K, 0, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
 		return(1);
 	}
@@ -1049,7 +1049,7 @@ char *arg;
 			i = 5;
 			tmp1 = val[0];
 		}
-		if (selectstr(&tmp1, i, 0, str, val) == K_ESC) return(1);
+		if (selectstr(&tmp1, i, 0, str, val) != K_CR) return(1);
 	}
 
 	if (!tmp1) {
@@ -1061,7 +1061,7 @@ char *arg;
 		str[1] = ODEC_K;
 		val[0] = 0;
 		val[1] = 8;
-		if (selectstr(&tmp2, 2, 56, str, val) == K_ESC)
+		if (selectstr(&tmp2, 2, 56, str, val) != K_CR)
 			return(1);
 		sorton = tmp1 + tmp2;
 		dupl = (int *)malloc2(maxfile * sizeof(int));
@@ -1145,12 +1145,36 @@ char *arg;
 static int quit_system(arg)
 char *arg;
 {
+	int n;
+
 #ifndef	_NOARCHIVE
 	if (archivefile) return(-1);
 #endif
+	n = 0;
+
+#ifndef	_NOORIGSHELL
+	if (!fdmode) {
+		char *str[3];
+		int val[3];
+
+		locate(0, L_MESLINE);
+		putterm(l_clear);
+		kanjiputs(QUIT_K);
+		val[0] = 0;
+		val[1] = 1;
+		val[2] = 2;
+		str[0] = QYES_K;
+		str[1] = QNO_K;
+		str[2] = QCHG_K;
+		if (selectstr(&n, 3, 20, str, val) != K_CR || n == 1)
+			return(1);
+	}
+	else
+#endif	/* _NOORIGSHELL */
 	if (!yesno(QUIT_K)) return(1);
+
 	if (savehist > 0) savehistory(0, histfile);
-	return(-1);
+	return((n) ? -2 : -1);
 }
 
 static int make_dir(arg)
@@ -1160,7 +1184,7 @@ char *arg;
 
 	if (arg && *arg) path = strdup2(arg);
 	else if (!(path = inputstr(MAKED_K, 1, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
 		return(1);
 	}
@@ -1470,7 +1494,7 @@ char *arg;
 
 	if (arg && *arg) file = strdup2(arg);
 	else if (!(file = inputstr(PACK_K, 1, -1, NULL, 1))) return(1);
-	else if (!*(file = evalpath(file, 1))) {
+	else if (!*(file = evalpath(file, 0))) {
 		free(file);
 		return(1);
 	}
@@ -1530,7 +1554,7 @@ char *arg;
 
 	if (arg && *arg) path = strdup2(arg);
 	else if (!(path = inputstr(FSDIR_K, 0, -1, NULL, 1))) return(1);
-	else if (!*(path = evalpath(path, 1))) {
+	else if (!*(path = evalpath(path, 0))) {
 		free(path);
 		path = strdup2(".");
 	}
@@ -1561,7 +1585,7 @@ char *arg;
 		locate(0, L_INFO);
 		putterm(l_clear);
 		kanjiputs(ATTRM_K);
-		if (selectstr(&flag, 2, 35, str, val) == K_ESC) return(1);
+		if (selectstr(&flag, 2, 35, str, val) != K_CR) return(1);
 	}
 	else {
 #ifndef	NOSYMLINK
@@ -1629,7 +1653,7 @@ char *arg;
 #else
 	else if (!(dev = inputstr(BKUP_K, 1, 5, "/dev/", 1))) return(1);
 #endif
-	else if (!*(dev = evalpath(dev, 1))) {
+	else if (!*(dev = evalpath(dev, 0))) {
 		free(dev);
 		return(1);
 	}
