@@ -8,7 +8,6 @@
 #include "fd.h"
 #include "func.h"
 #include "funcno.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifndef	_NOORIGSHELL
@@ -118,7 +117,6 @@ extern int maxarchive;
 extern int origmaxarchive;
 # endif
 # ifdef	_USEDOSEMU
-extern devinfo fdtype[];
 extern devinfo *origfdtype;
 # endif
 extern int inruncom;
@@ -127,14 +125,8 @@ extern int inruncom;
 #define	MAXCUSTOM	7
 #define	MAXCUSTNAM	14
 #define	MAXCUSTVAL	(n_column - MAXCUSTNAM - 3)
-#define	inputcuststr(p, d, s, h) \
-			(inputstr(p, d, ((s) ? strlen(s) : 0), s, h))
 #define	noselect(n, m, x, s, v) \
-			(selectstr(n, m, x, s, v) == K_ESC)
-#define	custputs(s)	cprintf2("%.*k", n_lastcolumn, s)
-#define	custputs2(s)	cprintf2("%.*k", n_column, s)
-#define	int2str(b, n, s) \
-			snprintf2(b, s, "%d", (int)(n))
+			(selectstr(n, m, x, s, v) != K_CR)
 #define	MAXSAVEMENU	5
 #define	MAXTNAMLEN	8
 #ifndef	O_BINARY
@@ -144,10 +136,14 @@ extern int inruncom;
 typedef struct _envtable {
 	char *env;
 	VOID_P var;
+#ifdef	FORCEDSTDC
 	union {
 		char *str;
 		int num;
 	} def;
+#else
+	char *def;
+#endif
 #ifndef	_NOJPNMES
 	char *hmes;
 #endif
@@ -157,6 +153,16 @@ typedef struct _envtable {
 	u_char type;
 } envtable;
 
+#ifdef	FORCEDSTDC
+#define	def_str(n)	(envlist[n].def.str)
+#define	def_num(n)	(envlist[n].def.num)
+#define	DEFVAL(d)	{(char *)(d)}
+#else
+#define	def_str(n)	(envlist[n].def)
+#define	def_num(n)	((int)(envlist[n].def))
+#define	DEFVAL(d)	(char *)(d)
+#endif
+
 #define	T_BOOL		0
 #define	T_SHORT		1
 #define	T_INT		2
@@ -185,9 +191,11 @@ static int NEAR atooctal __P_((char *));
 #endif
 static VOID NEAR _evalenv __P_((int));
 #ifndef	_NOCUSTOMIZE
+static int NEAR custputs __P_((char *));
 static char *NEAR strcatalloc __P_((char *, char *));
 static VOID NEAR putsep __P_((VOID_A));
 static VOID NEAR fillline __P_((int, int));
+static char *NEAR inputcuststr __P_((char *, int, char *, int));
 static VOID NEAR setnamelist __P_((int, namelist *, char *));
 static int NEAR browsenamelist __P_((namelist *, int,
 		int, char *, char *, char **));
@@ -202,6 +210,7 @@ static char *NEAR ascoctal __P_((int, char *));
 # ifndef	_NOORIGSHELL
 static VOID NEAR putargs __P_((char *, int, char *[], FILE *));
 # endif
+static char *NEAR int2str __P_((char *, int));
 static int NEAR dispenv __P_((int));
 static int NEAR editenv __P_((int));
 static int NEAR dumpenv __P_((char *, FILE *));
@@ -274,114 +283,114 @@ static int NEAR editcust __P_((VOID_A));
 int custno = -1;
 #endif
 
-static envtable envlist[] = {
-	{"FD_SORTTYPE", &sorttype, {(char *)SORTTYPE}, STTP_E, T_SORT},
-	{"FD_DISPLAYMODE", &displaymode, {(char *)DISPLAYMODE}, DPMD_E, T_DISP},
+static CONST envtable envlist[] = {
+	{"FD_SORTTYPE", &sorttype, DEFVAL(SORTTYPE), STTP_E, T_SORT},
+	{"FD_DISPLAYMODE", &displaymode, DEFVAL(DISPLAYMODE), DPMD_E, T_DISP},
 #ifndef	_NOTREE
-	{"FD_SORTTREE", &sorttree, {(char *)SORTTREE}, STTR_E, T_BOOL},
+	{"FD_SORTTREE", &sorttree, DEFVAL(SORTTREE), STTR_E, T_BOOL},
 #endif
 #ifndef	_NOWRITEFS
-	{"FD_WRITEFS", &writefs, {(char *)WRITEFS}, WRFS_E, T_WRFS},
+	{"FD_WRITEFS", &writefs, DEFVAL(WRITEFS), WRFS_E, T_WRFS},
 #endif
 #if	!defined (PATHNOCASE) && (FD >= 2)
-	{"FD_IGNORECASE", &pathignorecase,
-		{(char *)IGNORECASE}, IGNC_E, T_BOOL},
+	{"FD_IGNORECASE", &pathignorecase, DEFVAL(IGNORECASE), IGNC_E, T_BOOL},
 #endif
 #if	!defined (_USEDOSCOPY) && !defined (_NOEXTRACOPY)
-	{"FD_INHERITCOPY", &inheritcopy, {(char *)INHERITCOPY}, IHTM_E, T_BOOL},
+	{"FD_INHERITCOPY", &inheritcopy, DEFVAL(INHERITCOPY), IHTM_E, T_BOOL},
 #endif
 #if	!MSDOS
-	{"FD_ADJTTY", &adjtty, {(char *)ADJTTY}, AJTY_E, T_BOOL},
+	{"FD_ADJTTY", &adjtty, DEFVAL(ADJTTY), AJTY_E, T_BOOL},
 # if	FD >= 2
 	{"FD_USEGETCURSOR", &usegetcursor,
-		{(char *)USEGETCURSOR}, UGCS_E, T_BOOL},
+		DEFVAL(USEGETCURSOR), UGCS_E, T_BOOL},
 # endif
 #endif	/* !MSDOS */
-	{"FD_DEFCOLUMNS", &defcolumns, {(char *)DEFCOLUMNS}, CLMN_E, T_COLUMN},
+	{"FD_DEFCOLUMNS", &defcolumns, DEFVAL(DEFCOLUMNS), CLMN_E, T_COLUMN},
 	{"FD_MINFILENAME", &minfilename,
-		{(char *)MINFILENAME}, MINF_E, T_NATURAL},
-	{"FD_HISTFILE", &histfile, {HISTFILE}, HSFL_E, T_PATH},
-	{"FD_HISTSIZE", &(histsize[0]), {(char *)HISTSIZE}, HSSZ_E, T_SHORT},
-	{"FD_DIRHIST", &(histsize[1]), {(char *)DIRHIST}, DRHS_E, T_SHORT},
-	{"FD_SAVEHIST", &savehist, {(char *)SAVEHIST}, SVHS_E, T_INT},
+		DEFVAL(MINFILENAME), MINF_E, T_NATURAL},
+	{"FD_HISTFILE", &histfile, DEFVAL(HISTFILE), HSFL_E, T_PATH},
+	{"FD_HISTSIZE", &(histsize[0]), DEFVAL(HISTSIZE), HSSZ_E, T_SHORT},
+	{"FD_DIRHIST", &(histsize[1]), DEFVAL(DIRHIST), DRHS_E, T_SHORT},
+	{"FD_SAVEHIST", &savehist, DEFVAL(SAVEHIST), SVHS_E, T_INT},
 #ifndef	_NOTREE
 	{"FD_DIRCOUNTLIMIT", &dircountlimit,
-		{(char *)DIRCOUNTLIMIT}, DCLM_E, T_INT},
+		DEFVAL(DIRCOUNTLIMIT), DCLM_E, T_INT},
 #endif
 #ifndef	_NODOSDRIVE
-	{"FD_DOSDRIVE", &dosdrive, {(char *)DOSDRIVE}, DOSD_E, T_DDRV},
+	{"FD_DOSDRIVE", &dosdrive, DEFVAL(DOSDRIVE), DOSD_E, T_DDRV},
 #endif
-	{"FD_SECOND", &showsecond, {(char *)SECOND}, SCND_E, T_BOOL},
+	{"FD_SECOND", &showsecond, DEFVAL(SECOND), SCND_E, T_BOOL},
 #ifndef	_NOTRADLAYOUT
-	{"FD_TRADLAYOUT", &tradlayout, {(char *)TRADLAYOUT}, TRLO_E, T_BOOL},
+	{"FD_TRADLAYOUT", &tradlayout, DEFVAL(TRADLAYOUT), TRLO_E, T_BOOL},
 #endif
-	{"FD_SIZEINFO", &sizeinfo, {(char *)SIZEINFO}, SZIF_E, T_BOOL},
+	{"FD_SIZEINFO", &sizeinfo, DEFVAL(SIZEINFO), SZIF_E, T_BOOL},
 #ifndef	_NOCOLOR
-	{"FD_ANSICOLOR", &ansicolor, {(char *)ANSICOLOR}, ACOL_E, T_COLOR},
+	{"FD_ANSICOLOR", &ansicolor, DEFVAL(ANSICOLOR), ACOL_E, T_COLOR},
 # if	FD >= 2
-	{"FD_ANSIPALETTE", &ansipalette,
-		{(char *)ANSIPALETTE}, APAL_E, T_CHARP},
+	{"FD_ANSIPALETTE", &ansipalette, DEFVAL(ANSIPALETTE), APAL_E, T_CHARP},
 # endif
 #endif
 #ifndef	_NOEDITMODE
-	{"FD_EDITMODE", &editmode, {EDITMODE}, EDMD_E, T_EDIT},
+	{"FD_EDITMODE", &editmode, DEFVAL(EDITMODE), EDMD_E, T_EDIT},
 #endif
 #if	FD >= 2
-	{"FD_LOOPCURSOR", &loopcursor, {(char *)LOOPCURSOR}, LPCS_E, T_BOOL},
+	{"FD_LOOPCURSOR", &loopcursor, DEFVAL(LOOPCURSOR), LPCS_E, T_BOOL},
+#endif
+	{"FD_TMPDIR", &deftmpdir, DEFVAL(TMPDIR), TMPD_E, T_PATH},
+#if	FD >= 2
+	{"FD_TMPUMASK", &tmpumask, DEFVAL(TMPUMASK), TUMSK_E, T_OCTAL},
 #endif
-	{"FD_TMPDIR", &deftmpdir, {TMPDIR}, TMPD_E, T_PATH},
 #ifndef	_NOROCKRIDGE
-	{"FD_RRPATH", &rockridgepath, {RRPATH}, RRPT_E, T_PATHS},
+	{"FD_RRPATH", &rockridgepath, DEFVAL(RRPATH), RRPT_E, T_PATHS},
 #endif
 #ifndef	_NOPRECEDE
-	{"FD_PRECEDEPATH", &precedepath, {PRECEDEPATH}, PCPT_E, T_PATHS},
+	{"FD_PRECEDEPATH", &precedepath, DEFVAL(PRECEDEPATH), PCPT_E, T_PATHS},
 #endif
 #if	FD >= 2
-	{"FD_PS1", &promptstr, {PROMPT}, PRMP_E, T_CHARP},
+	{"FD_PS1", &promptstr, DEFVAL(PROMPT), PRMP_E, T_CHARP},
 #else
-	{"FD_PROMPT", &promptstr, {PROMPT}, PRMP_E, T_CHARP},
+	{"FD_PROMPT", &promptstr, DEFVAL(PROMPT), PRMP_E, T_CHARP},
 #endif
 #ifndef	_NOORIGSHELL
-	{"FD_PS2", &promptstr2, {PROMPT2}, PRM2_E, T_CHARP},
+	{"FD_PS2", &promptstr2, DEFVAL(PROMPT2), PRM2_E, T_CHARP},
+#endif
+#ifndef	_NOORIGSHELL
+	{"FD_DUMBSHELL", &dumbshell, DEFVAL(DUMBSHELL), DMSHL_E, T_BOOL},
 #endif
 #if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
-	{"FD_UNICODEBUFFER", &unicodebuffer, {UNICODEBUFFER}, UNBF_E, T_BOOL},
+	{"FD_UNICODEBUFFER", &unicodebuffer,
+		DEFVAL(UNICODEBUFFER), UNBF_E, T_BOOL},
 #endif
 #if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
-	{"FD_LANGUAGE", &outputkcode, {(char *)NOCNV}, LANG_E, T_KOUT},
+	{"FD_LANGUAGE", &outputkcode, DEFVAL(NOCNV), LANG_E, T_KOUT},
 #endif
 #ifndef	_NOKANJICONV
-	{"FD_INPUTKCODE", &inputkcode, {(char *)NOCNV}, IPKC_E, T_KIN},
+	{"FD_INPUTKCODE", &inputkcode, DEFVAL(NOCNV), IPKC_E, T_KIN},
 #endif
 #ifndef	_NOKANJIFCONV
-	{"FD_FNAMEKCODE", &fnamekcode, {(char *)NOCNV}, FNKC_E, T_KNAM},
-	{"FD_SJISPATH", &sjispath, {SJISPATH}, SJSP_E, T_PATHS},
-	{"FD_EUCPATH", &eucpath, {EUCPATH}, EUCP_E, T_PATHS},
-	{"FD_JISPATH", &jis7path, {JISPATH}, JISP_E, T_PATHS},
-	{"FD_JIS8PATH", &jis8path, {JIS8PATH}, JS8P_E, T_PATHS},
-	{"FD_JUNETPATH", &junetpath, {JUNETPATH}, JNTP_E, T_PATHS},
-	{"FD_OJISPATH", &ojis7path, {OJISPATH}, OJSP_E, T_PATHS},
-	{"FD_OJIS8PATH", &ojis8path, {OJIS8PATH}, OJ8P_E, T_PATHS},
-	{"FD_OJUNETPATH", &ojunetpath, {OJUNETPATH}, OJNP_E, T_PATHS},
-	{"FD_HEXPATH", &hexpath, {HEXPATH}, HEXP_E, T_PATHS},
-	{"FD_CAPPATH", &cappath, {CAPPATH}, CAPP_E, T_PATHS},
-	{"FD_UTF8PATH", &utf8path, {UTF8PATH}, UTF8P_E, T_PATHS},
-	{"FD_UTF8MACPATH", &utf8macpath, {UTF8MACPATH}, UTF8MACP_E, T_PATHS},
-	{"FD_NOCONVPATH", &noconvpath, {NOCONVPATH}, NCVP_E, T_PATHS},
+	{"FD_FNAMEKCODE", &fnamekcode, DEFVAL(NOCNV), FNKC_E, T_KNAM},
+	{"FD_SJISPATH", &sjispath, DEFVAL(SJISPATH), SJSP_E, T_PATHS},
+	{"FD_EUCPATH", &eucpath, DEFVAL(EUCPATH), EUCP_E, T_PATHS},
+	{"FD_JISPATH", &jis7path, DEFVAL(JISPATH), JISP_E, T_PATHS},
+	{"FD_JIS8PATH", &jis8path, DEFVAL(JIS8PATH), JS8P_E, T_PATHS},
+	{"FD_JUNETPATH", &junetpath, DEFVAL(JUNETPATH), JNTP_E, T_PATHS},
+	{"FD_OJISPATH", &ojis7path, DEFVAL(OJISPATH), OJSP_E, T_PATHS},
+	{"FD_OJIS8PATH", &ojis8path, DEFVAL(OJIS8PATH), OJ8P_E, T_PATHS},
+	{"FD_OJUNETPATH", &ojunetpath, DEFVAL(OJUNETPATH), OJNP_E, T_PATHS},
+	{"FD_HEXPATH", &hexpath, DEFVAL(HEXPATH), HEXP_E, T_PATHS},
+	{"FD_CAPPATH", &cappath, DEFVAL(CAPPATH), CAPP_E, T_PATHS},
+	{"FD_UTF8PATH", &utf8path, DEFVAL(UTF8PATH), UTF8P_E, T_PATHS},
+	{"FD_UTF8MACPATH", &utf8macpath,
+		DEFVAL(UTF8MACPATH), UTF8MACP_E, T_PATHS},
+	{"FD_NOCONVPATH", &noconvpath, DEFVAL(NOCONVPATH), NCVP_E, T_PATHS},
 #endif	/* !_NOKANJIFCONV */
-#if	FD >= 2
-	{"FD_TMPUMASK", &tmpumask, {(char *)TMPUMASK}, TUMSK_E, T_OCTAL},
-#endif
-#ifndef	_NOORIGSHELL
-	{"FD_DUMBSHELL", &dumbshell, {(char *)DUMBSHELL}, DMSHL_E, T_BOOL},
-#endif
 };
 #define	ENVLISTSIZ	((int)(sizeof(envlist) / sizeof(envtable)))
 
 #ifndef	_NOCUSTOMIZE
 # ifdef	_USEDOSEMU
-static devinfo mediadescr[] = {
+static CONST devinfo mediadescr[] = {
 	{0xf0, "2HD(PC/AT)", 2, 18, 80},
 	{0xf9, "2HD(PC98)", 2, 15, 80},
 	{0xf9, "2DD(PC/AT)", 2, 9, 80},
@@ -424,14 +433,7 @@ char *s;
 {
 	int n;
 
-	if (!s) return(-1);
-	n = 0;
-	while (*s) {
-		if (*s < '0' || *s > '7') break;
-		n = (n << 3) + *s - '0';
-		s++;
-	}
-	if (*s) return(-1);
+	if (!sscanf2(s, "%o%$", &n)) return(-1);
 	n &= 0777;
 	return(n);
 }
@@ -446,31 +448,31 @@ int no;
 	cp = getenv2(envlist[no].env);
 	switch (envlist[no].type) {
 		case T_BOOL:
-			if (!cp) n = envlist[no].def.num;
+			if (!cp) n = def_num(no);
 			else n = (*cp && atoi2(cp)) ? 1 : 0;
 			*((int *)(envlist[no].var)) = n;
 			break;
 		case T_SHORT:
-			if ((n = atoi2(cp)) < 0) n = envlist[no].def.num;
+			if ((n = atoi2(cp)) < 0) n = def_num(no);
 			*((short *)(envlist[no].var)) = n;
 			break;
 		case T_INT:
-			if ((n = atoi2(cp)) < 0) n = envlist[no].def.num;
+			if ((n = atoi2(cp)) < 0) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 		case T_NATURAL:
-			if ((n = atoi2(cp)) <= 0) n = envlist[no].def.num;
+			if ((n = atoi2(cp)) < 0) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 		case T_PATH:
-			if (!cp) cp = envlist[no].def.str;
-			cp = evalpath(strdup2(cp), 1);
+			if (!cp) cp = def_str(no);
+			cp = evalpath(strdup2(cp), 0);
 			if (*((char **)(envlist[no].var)))
 				free(*((char **)(envlist[no].var)));
 			*((char **)(envlist[no].var)) = cp;
 			break;
 		case T_PATHS:
-			if (!cp) cp = envlist[no].def.str;
+			if (!cp) cp = def_str(no);
 			cp = evalpaths(cp, ':');
 			if (*((char **)(envlist[no].var)))
 				free(*((char **)(envlist[no].var)));
@@ -479,7 +481,7 @@ int no;
 		case T_SORT:
 			if (((n = atoi2(cp)) < 0 || (n & 7) > 5)
 			&& (n < 100 || ((n - 100) & 7) > 5))
-				n = envlist[no].def.num;
+				n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 		case T_DISP:
@@ -488,24 +490,23 @@ int no;
 #else
 			if ((n = atoi2(cp)) < 0 || n > 7)
 #endif
-				n = envlist[no].def.num;
+				n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 #ifndef	_NOWRITEFS
 		case T_WRFS:
-			if ((n = atoi2(cp)) < 0 || n > 2)
-				n = envlist[no].def.num;
+			if ((n = atoi2(cp)) < 0 || n > 2) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 #endif
 		case T_COLUMN:
 			if ((n = atoi2(cp)) < 0 || n > 5 || n == 4)
-				n = envlist[no].def.num;
+				n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 #if	MSDOS && !defined (_NODOSDRIVE)
 		case T_DDRV:
-			if (!cp) n = envlist[no].def.num;
+			if (!cp) n = def_num(no);
 			else {
 				char *dupl;
 
@@ -523,14 +524,13 @@ int no;
 #endif	/* MSDOS && !_NODOSDRIVE */
 #ifndef	_NOCOLOR
 		case T_COLOR:
-			if ((n = atoi2(cp)) < 0 || n > 3)
-				n = envlist[no].def.num;
+			if ((n = atoi2(cp)) < 0 || n > 3) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 #endif
 #ifndef	_NOEDITMODE
 		case T_EDIT:
-			if (!cp) cp = envlist[no].def.str;
+			if (!cp) cp = def_str(no);
 			else if (strcmp(cp, "emacs")
 			&& strcmp(cp, "vi")
 			&& strcmp(cp, "wordstar")) cp = NULL;
@@ -548,12 +548,12 @@ int no;
 #endif	/* !_NOKANJICONV || (!_NOENGMES && !NOJPNMES) */
 #if	FD >= 2
 		case T_OCTAL:
-			if ((n = atooctal(cp)) < 0) n = envlist[no].def.num;
+			if ((n = atooctal(cp)) < 0) n = def_num(no);
 			*((int *)(envlist[no].var)) = n;
 			break;
 #endif
 		default:
-			if (!cp) cp = envlist[no].def.str;
+			if (!cp) cp = def_str(no);
 			*((char **)(envlist[no].var)) = cp;
 			break;
 	}
@@ -587,6 +587,12 @@ VOID freeenvpath(VOID_A)
 #endif
 
 #ifndef	_NOCUSTOMIZE
+static int NEAR custputs(s)
+char *s;
+{
+	return(cprintf2("%.*k", n_lastcolumn, s));
+}
+
 static char *NEAR strcatalloc(s1, s2)
 char *s1, *s2;
 {
@@ -615,6 +621,18 @@ int y, w;
 	putterm(end_standout);
 }
 
+static char *NEAR inputcuststr(prompt, delsp, s, h)
+char *prompt;
+int delsp;
+char *s;
+int h;
+{
+	int len;
+
+	len = (s) ? strlen(s) : 0;
+	return(inputstr(prompt, delsp, len, s, h));
+}
+
 static VOID NEAR setnamelist(n, list, s)
 int n;
 namelist *list;
@@ -633,17 +651,19 @@ int max, col;
 char *def, *prompt, **mes;
 {
 	int ch, pos, old;
-	int dupwin_x, dupwin_y, dupminfilename, dupcolumns;
+	int dupwin_x, dupwin_y, dupminfilename, dupcolumns, dupdispmode;
 
 	dupminfilename = minfilename;
 	dupcolumns = curcolumns;
+	dupdispmode = dispmode;
 	minfilename = n_column;
 	curcolumns = col;
+	dispmode = 0;
 
 	dupwin_x = win_x;
 	dupwin_y = win_y;
 	subwindow = 1;
-	pos = listupfile(list, max, def);
+	pos = listupfile(list, max, def, 1);
 
 	envcaption(prompt);
 	do {
@@ -708,11 +728,11 @@ char *def, *prompt, **mes;
 		}
 
 		if (old / FILEPERPAGE != pos / FILEPERPAGE)
-			pos = listupfile(list, max, list[pos].name);
+			pos = listupfile(list, max, list[pos].name, 1);
 		else if (old != pos) {
 			putname(list, old, -1);
-			win_x = putname(list, pos, 1) + 1;
-			win_x += calc_x;
+			calc_x += putname(list, pos, 1) + 1;
+			win_x = calc_x;
 			win_y = calc_y;
 		}
 	} while (ch != K_ESC && ch != K_CR);
@@ -722,6 +742,7 @@ char *def, *prompt, **mes;
 	subwindow = 0;
 	minfilename = dupminfilename;
 	curcolumns = dupcolumns;
+	dispmode = dupdispmode;
 
 	return((ch == K_CR) ? pos : -1);
 }
@@ -740,7 +761,7 @@ static VOID NEAR custtitle(VOID_A)
 	str[6] = TSAVE_K;
 
 	for (i = width = max = 0; i < MAXCUSTOM; i++) {
-		len = strlen3(str[i]);
+		len = strlen2(str[i]);
 		width += len + 2;
 		if (max < len) max = len;
 	}
@@ -804,7 +825,7 @@ char *s;
 	locate(0, L_HELP);
 	putterm(l_clear);
 	putterm(t_standout);
-	custputs2(s);
+	cprintf2("%.*k", n_column, s);
 	putterm(end_standout);
 }
 
@@ -862,7 +883,7 @@ static VOID NEAR cleanupenv(VOID_A)
 
 	for (i = 0; i < ENVLISTSIZ; i++) {
 		setenv2(envlist[i].env, NULL, 0);
-		cp = (envlist[i].type == T_CHARP) ? envlist[i].def.str : NULL;
+		cp = (envlist[i].type == T_CHARP) ? def_str(i) : NULL;
 		setenv2(&(envlist[i].env[FDESIZ]), cp, 0);
 		_evalenv(i);
 	}
@@ -873,18 +894,11 @@ static char *NEAR ascoctal(n, buf)
 int n;
 char *buf;
 {
-	int i;
-
 # ifdef	BASHSTYLE
-	i = 3;
+	snprintf2(buf, MAXLONGWIDTH + 1, "%03o", n & 0777);
 # else
-	i = 4;
+	snprintf2(buf, MAXLONGWIDTH + 1, "%04o", n & 0777);
 # endif
-	buf[i] = '\0';
-	while (--i >= 0) {
-		buf[i] = '0' + (n & 7);
-		n >>= 3;
-	}
 	return(buf);
 }
 # endif	/* FD >= 2 */
@@ -906,6 +920,14 @@ FILE *fp;
 }
 # endif	/* _NOORIGSHELL */
 
+static char *NEAR int2str(buf, n)
+char *buf;
+int n;
+{
+	snprintf2(buf, MAXLONGWIDTH + 1, "%d", n);
+	return(buf);
+}
+
 static int NEAR dispenv(no)
 int no;
 {
@@ -920,14 +942,12 @@ int no;
 			cp = str[(*((int *)(envlist[no].var))) ? 1 : 0];
 			break;
 		case T_SHORT:
-			cp = int2str(buf,
-				*((short *)(envlist[no].var)), sizeof(buf));
+			cp = int2str(buf, *((short *)(envlist[no].var)));
 			break;
 		case T_INT:
 		case T_NATURAL:
 		case T_COLUMN:
-			cp = int2str(buf,
-				*((int *)(envlist[no].var)), sizeof(buf));
+			cp = int2str(buf, *((int *)(envlist[no].var)));
 			break;
 		case T_SORT:
 			n = *((int *)(envlist[no].var));
@@ -1042,14 +1062,14 @@ int no;
 # endif
 		default:
 			if (!(cp = getenv2(envlist[no].env))) {
-				cp = strdup2(envlist[no].def.str);
+				cp = strdup2(def_str(no));
 				if (cp && *cp) cp = strcatalloc(cp, " ");
 				cp = new = strcatalloc(cp, VDEF_K);
 			}
 			break;
 	}
 	cprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
-	n = strlen3(cp);
+	n = strlen2(cp);
 	if (n > MAXCUSTVAL - 1) n = MAXCUSTVAL - 1;
 	if (new) free(new);
 	return(n);
@@ -1072,11 +1092,10 @@ int no;
 			str[1] = VBOL1_K;
 			envcaption(env);
 			if (noselect(&n, 2, 0, str, val)) return(0);
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 		case T_SHORT:
-			int2str(buf,
-				*((short *)(envlist[no].var)), sizeof(buf));
+			int2str(buf, *((short *)(envlist[no].var)));
 			cp = inputcuststr(env, 1, buf, -1);
 			if (!cp) return(0);
 			n = atoi2(cp);
@@ -1085,12 +1104,11 @@ int no;
 				warning(0, VALNG_K);
 				return(0);
 			}
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 		case T_INT:
 		case T_NATURAL:
-			int2str(buf,
-				*((int *)(envlist[no].var)), sizeof(buf));
+			int2str(buf, *((int *)(envlist[no].var)));
 			cp = inputcuststr(env, 1, buf, -1);
 			if (!cp) return(0);
 			n = atoi2(cp);
@@ -1099,11 +1117,10 @@ int no;
 				warning(0, VALNG_K);
 				return(0);
 			}
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 		case T_PATH:
-			if (!(cp = getenv2(envlist[no].env)))
-				cp = envlist[no].def.str;
+			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
 			new = inputcuststr(env, 1, cp, 1);
 			if (!new) return(0);
 			cp = new;
@@ -1143,7 +1160,7 @@ int no;
 			val[1] = 100;
 			if (noselect(&p, 2, 0, str, val)) return(0);
 			n += p;
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 		case T_DISP:
 			n = *((int *)(envlist[no].var));
@@ -1173,7 +1190,7 @@ int no;
 			if (noselect(&tmp, 2, 0, str, val)) return(0);
 			n = (n & ~8) | (tmp << 3);
 # endif
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 # ifndef	_NOWRITEFS
 		case T_WRFS:
@@ -1183,7 +1200,7 @@ int no;
 			str[2] = VWFS2_K;
 			envcaption(env);
 			if (noselect(&n, 3, 0, str, val)) return(0);
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 # endif	/* !_NOWRITEFS */
 # if	MSDOS && !defined (_NODOSDRIVE)
@@ -1193,7 +1210,7 @@ int no;
 			str[1] = VBOL1_K;
 			envcaption(env);
 			if (noselect(&n, 2, 0, str, val)) return(0);
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			if (*((int *)(envlist[no].var)) & 2) {
 				cp = strdup2(cp);
 				cp = new = strcatalloc(cp, ",BIOS");
@@ -1212,7 +1229,7 @@ int no;
 			val[3] = 5;
 			envcaption(env);
 			if (noselect(&n, 4, 0, str, val)) return(0);
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 # ifndef	_NOCOLOR
 		case T_COLOR:
@@ -1223,7 +1240,7 @@ int no;
 			str[3] = VCOL3_K;
 			envcaption(env);
 			if (noselect(&n, 4, 0, str, val)) return(0);
-			cp = int2str(buf, n, sizeof(buf));
+			cp = int2str(buf, n);
 			break;
 # endif	/* !_NOCOLOR */
 # ifndef	_NOEDITMODE
@@ -1240,53 +1257,69 @@ int no;
 			cp = str[n];
 			break;
 # endif	/* !_NOEDITMODE */
-# ifndef	_NOKANJICONV
-		case T_KIN:
-			n = *((int *)(envlist[no].var));
-			str[0] = "ShiftJIS";
-			str[1] = "EUC-JP";
-			val[0] = SJIS;
-			val[1] = EUC;
-			envcaption(env);
-			if (noselect(&n, 2, 0, str, val)) return(0);
-			str[SJIS] = "sjis";
-			str[EUC] = "euc";
-			cp = str[n];
-			break;
-# endif	/* !_NOKANJICONV */
 # if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
+		case T_KIN:
+		case T_KNAM:
 		case T_KOUT:
+			tmp = 0;
+			str[tmp] = VNCNV_K;
+			val[tmp++] = NOCNV;
+#  if	!defined (_NOENGMES) && !defined (_NOJPNMES)
+			str[tmp] = VENG_K;
+			val[tmp++] = ENG;
+#  endif
+#  ifndef	_NOKANJICONV
+			str[tmp] = "SJIS";
+			val[tmp++] = SJIS;
+			str[tmp] = "EUC";
+			val[tmp++] = EUC;
+			str[tmp] = "7bit-JIS";
+			val[tmp++] = JIS7;
+			str[tmp] = "8bit-JIS";
+			val[tmp++] = JIS8;
+			str[tmp] = "ISO-2022-JP";
+			val[tmp++] = JUNET;
+			str[tmp] = "UTF-8";
+			val[tmp++] = UTF8;
+#  endif	/* !_NOKANJICONV */
+			p = envlist[no].type - T_KIN;
+			for (n = 0; n < tmp; n++) {
+				switch (kanjiiomode[val[n]]) {
+					case L_INPUT:
+						if (p != L_INPUT) continue;
+						break;
+					case L_OUTPUT:
+						if (p == L_OUTPUT) continue;
+						break;
+					case L_FNAME:
+						if (p == L_FNAME) continue;
+						break;
+					default:
+						continue;
+/*NOTREACHED*/
+						break;
+				}
+				tmp--;
+				memmove(&(str[n]), &(str[n + 1]),
+					(tmp - n) * sizeof(char *));
+				memmove(&(val[n]), &(val[n + 1]),
+					(tmp - n) * sizeof(int));
+				n--;
+			}
+			if (!tmp) return(0);
 			n = *((int *)(envlist[no].var));
-			str[0] = VNCNV_K;
-			str[1] = VENG_K;
-			val[0] = NOCNV;
-			val[1] = ENG;
-#  ifdef	_NOKANJICONV
-			envcaption(env);
-			if (noselect(&n, 2, 0, str, val)) return(0);
-#  else	/* !_NOKANJICONV */
-
-			str[2] = "SJIS";
-			str[3] = "EUC";
-			str[4] = "7bit-JIS";
-			str[5] = "8bit-JIS";
-			str[6] = "ISO-2022-JP";
-			str[7] = "UTF-8";
-			val[2] = SJIS;
-			val[3] = EUC;
-			val[4] = JIS7;
-			val[5] = JIS8;
-			val[6] = JUNET;
-			val[7] = UTF8;
+#  ifndef	_NOKANJICONV
 			if (n != O_JIS7 && n != O_JIS8 && n != O_JUNET
 			&& n != M_UTF8) p = 0;
 			else {
 				p = 1;
 				n--;
 			}
+#  endif	/* !_NOKANJICONV */
 			envcaption(env);
-			if (noselect(&n, 8, 0, str, val)) return(0);
+			if (noselect(&n, tmp, 0, str, val)) return(0);
+#  ifndef	_NOKANJICONV
 			if (n >= JIS7 && n <= JUNET) {
 				str[0] = VNJIS_K;
 				str[1] = VOJIS_K;
@@ -1303,68 +1336,12 @@ int no;
 				if (noselect(&p, 2, 64, str, val)) return(0);
 				n += p;
 			}
-			str[SJIS] = "sjis";
-			str[EUC] = "euc";
-			str[JIS7] = "jis7";
-			str[O_JIS7] = "ojis7";
-			str[JIS8] = "jis8";
-			str[O_JIS8] = "ojis8";
-			str[JUNET] = "junet";
-			str[O_JUNET] = "ojunet";
-			str[UTF8] = "utf8";
-			str[M_UTF8] = "utf8-mac";
 #  endif	/* !_NOKANJICONV */
 			str[NOCNV] = "";
+#  if	!defined (_NOENGMES) && !defined (_NOJPNMES)
 			str[ENG] = "C";
-			cp = str[n];
-			break;
-# endif	/* !_NOKANJICONV || (!_NOENGMES && !NOJPNMES) */
-# ifndef	_NOKANJIFCONV
-		case T_KNAM:
-			n = *((int *)(envlist[no].var));
-			str[0] = VNCNV_K;
-			str[1] = "SJIS";
-			str[2] = "EUC";
-			str[3] = "7bit-JIS";
-			str[4] = "8bit-JIS";
-			str[5] = "ISO-2022-JP";
-			str[6] = "HEX";
-			str[7] = "CAP";
-			str[8] = "UTF-8";
-			val[0] = NOCNV;
-			val[1] = SJIS;
-			val[2] = EUC;
-			val[3] = JIS7;
-			val[4] = JIS8;
-			val[5] = JUNET;
-			val[6] = HEX;
-			val[7] = CAP;
-			val[8] = UTF8;
-			if (n != O_JIS7 && n != O_JIS8 && n != O_JUNET
-			&& n != M_UTF8) p = 0;
-			else {
-				p = 1;
-				n--;
-			}
-			envcaption(env);
-			if (noselect(&n, 9, 0, str, val)) return(0);
-			if (n >= JIS7 && n <= JUNET) {
-				str[0] = VNJIS_K;
-				str[1] = VOJIS_K;
-				val[0] = 0;
-				val[1] = 1;
-				if (noselect(&p, 2, 64, str, val)) return(0);
-				n += p;
-			}
-			else if (n == UTF8) {
-				str[0] = VUTF8_K;
-				str[1] = VUTFM_K;
-				val[0] = 0;
-				val[1] = 1;
-				if (noselect(&p, 2, 64, str, val)) return(0);
-				n += p;
-			}
-			str[NOCNV] = "";
+#  endif
+#  ifndef	_NOKANJICONV
 			str[SJIS] = "sjis";
 			str[EUC] = "euc";
 			str[JIS7] = "jis7";
@@ -1377,9 +1354,10 @@ int no;
 			str[CAP] = "cap";
 			str[UTF8] = "utf8";
 			str[M_UTF8] = "utf8-mac";
+#  endif	/* !_NOKANJICONV */
 			cp = str[n];
 			break;
-# endif	/* !_NOKANJIFCONV */
+# endif	/* !_NOKANJICONV || (!_NOENGMES && !NOJPNMES) */
 # if	FD >= 2
 		case T_OCTAL:
 			ascoctal(*((int *)(envlist[no].var)), buf);
@@ -1395,8 +1373,7 @@ int no;
 			break;
 # endif	/* FD >= 2 */
 		default:
-			if (!(cp = getenv2(envlist[no].env)))
-				cp = envlist[no].def.str;
+			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
 			new = inputcuststr(env, 0, cp, -1);
 			if (!new) return(0);
 			cp = new;
@@ -1420,8 +1397,8 @@ FILE *fp;
 	for (i = n = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
 		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
-		&& (envlist[i].type != T_CHARP
-		|| strenvcmp(cp, envlist[i].def.str))) n++;
+		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i))))
+			n++;
 		if ((!flaglist || !(flaglist[i] & 1))
 		&& getshellvar(envlist[i].env, -1)) n++;
 	}
@@ -1432,8 +1409,7 @@ FILE *fp;
 	for (i = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
 		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
-		&& (envlist[i].type != T_CHARP
-		|| strenvcmp(cp, envlist[i].def.str))) {
+		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i)))) {
 			cp = killmeta(cp);
 			fprintf2(fp, "%s=%s\n", &(envlist[i].env[FDESIZ]), cp);
 			free(cp);
@@ -1499,7 +1475,7 @@ FILE *fp;
 			if (flaglist) flaglist[i] |= f;
 			if ((cp = getshellvar(ident, -1))
 			&& (envlist[i].type != T_CHARP
-			|| strenvcmp(cp, envlist[i].def.str))) {
+			|| strenvcmp(cp, def_str(i)))) {
 				if (ns++) fputc(' ', fp);
 				cp = killmeta(cp);
 				fprintf2(fp, "%s=%s", ident, cp);
@@ -1581,7 +1557,7 @@ FILE *fp;
 			if (flaglist) flaglist[i] |= f;
 			if ((cp = getshellvar(ident, -1))
 			&& (envlist[i].type != T_CHARP
-			|| strenvcmp(cp, envlist[i].def.str))) {
+			|| strenvcmp(cp, def_str(i)))) {
 				if (ns++) fputc(' ', fp);
 				cp = killmeta(cp);
 				fprintf2(fp, "%s=%s", ident, cp);
@@ -1664,7 +1640,7 @@ int no;
 		cprintf2("%-*.*k",
 			MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width, cp2);
 	}
-	len = strlen3(cp1);
+	len = strlen2(cp1);
 	if (len > --width) len = width;
 	return(len);
 }
@@ -1980,7 +1956,7 @@ int no;
 	}
 	cp = encodestr(key.str, key.len);
 	cprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
-	len = strlen3(cp);
+	len = strlen2(cp);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 	free(cp);
 	return(len);
@@ -2001,8 +1977,7 @@ int no;
 	locate(0, L_INFO);
 	putterm(l_clear);
 	buf = asprintf3(KYMPK_K, cp);
-	kanjiputs(buf);
-	win_x = strlen3(buf);
+	win_x = kanjiputs(buf);
 	win_y = L_INFO;
 	locate(win_x, win_y);
 	tflush();
@@ -2014,7 +1989,8 @@ int no;
 	len = 0;
 	for (;;) {
 		buf = c_realloc(buf, len, &size);
-		buf[len++] = getch2();
+		if ((i = getch2()) == EOF) break;
+		buf[len++] = i;
 		if (!kbhit2(WAITKEYPAD * 1000L)) break;
 	}
 	buf = realloc2(buf, len + 1);
@@ -2239,11 +2215,11 @@ int no;
 		width = MAXCUSTVAL - 1 - 6 - width;
 		cprintf2("%-*.*k", width, width, launchlist[no].format[0]);
 		putsep();
-		cprintf2("%-2d", launchlist[no].topskip);
+		cprintf2("%-2d", (int)(launchlist[no].topskip));
 		putsep();
-		cprintf2("%-2d", launchlist[no].bottomskip);
+		cprintf2("%-2d", (int)(launchlist[no].bottomskip));
 	}
-	len = strlen3(launchlist[no].comm);
+	len = strlen2(launchlist[no].comm);
 	if (len > --width) len = width;
 	return(len);
 }
@@ -2256,7 +2232,7 @@ launchtable *list;
 	custtitle();
 	y = LFILETOP + 1;
 	fillline(y++, n_column);
-	len = strlen3(list -> ext + 1);
+	len = strlen2(list -> ext + 1);
 	if (list -> flags & LF_IGNORECASE) len++;
 	if (len >= MAXCUSTNAM) len = MAXCUSTNAM;
 	fillline(y++, MAXCUSTNAM + 1 - len);
@@ -2285,12 +2261,12 @@ launchtable *list;
 				putterm(t_standout);
 				cputs2("-t");
 				putterm(end_standout);
-				cprintf2("%3d", list -> topskip);
+				cprintf2("%3d", (int)(list -> topskip));
 				putsep();
 				putterm(t_standout);
 				cputs2("-b");
 				putterm(end_standout);
-				cprintf2("%3d", list -> bottomskip);
+				cprintf2("%3d", (int)(list -> bottomskip));
 				putsep();
 				putterm(t_standout);
 				cputs2("-f");
@@ -2458,7 +2434,8 @@ launchtable *list;
 				freevar(list -> lignore);
 				freevar(list -> lerror);
 				list -> lignore = list -> lerror = NULL;
-				list -> topskip = list -> bottomskip = 0;
+				list -> topskip =
+				list -> bottomskip = (u_char)0;
 			}
 			break;
 		}
@@ -2476,7 +2453,7 @@ launchtable *list;
 				skipp = &(list -> bottomskip);
 				cp = BTMSK_K;
 			}
-			cp = int2str(buf, *skipp, sizeof(buf));
+			int2str(buf, *skipp);
 			if (!(cp = inputcuststr(cp, 1, buf, -1))) continue;
 			i = atoi2(cp);
 			free(cp);
@@ -2524,7 +2501,8 @@ int no;
 			launchlist[i].comm = NULL;
 			launchlist[i].format =
 			launchlist[i].lignore = launchlist[i].lerror = NULL;
-			launchlist[i].topskip = launchlist[i].bottomskip = 0;
+			launchlist[i].topskip =
+			launchlist[i].bottomskip = (u_char)0;
 			launchlist[i].flags = list.flags;
 		}
 		no = i;
@@ -2569,7 +2547,7 @@ int no;
 			freevar(list.lignore);
 			freevar(list.lerror);
 			list.format = list.lignore = list.lerror = NULL;
-			list.topskip = list.bottomskip = 0;
+			list.topskip = list.bottomskip = (u_char)0;
 			break;
 		}
 
@@ -2788,7 +2766,7 @@ int no;
 	else cprintf2("%-*.*k",
 		MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width,
 		archivelist[no].u_comm);
-	len = (archivelist[no].p_comm) ? strlen3(archivelist[no].p_comm) : 0;
+	len = (archivelist[no].p_comm) ? strlen2(archivelist[no].p_comm) : 0;
 	if (len > --width) len = width;
 	return(len);
 }
@@ -3048,8 +3026,7 @@ int no;
 		char buf[sizeof("HDD98 #offset=") + MAXCOLSCOMMA(3)];
 
 		strcpy(buf, "HDD");
-		if (fdtype[no].head >= 'A' && fdtype[no].head <= 'Z')
-			strcat(buf, "98");
+		if (isupper2(fdtype[no].head)) strcat(buf, "98");
 		i = strlen(buf);
 		snprintf2(&(buf[i]), sizeof(buf) - i, " #offset=%'Ld",
 			fdtype[no].offset / fdtype[no].sect);
@@ -3060,17 +3037,17 @@ int no;
 	{
 		len = 0;
 		for (i = 0; i < MEDIADESCRSIZ; i++) {
-			w2 = strlen(mediadescr[i].name);
+			w2 = strlen2(mediadescr[i].name);
 			if (len < w2) len = w2;
 		}
 
 		w2 = (w1 - len) / 3 - 1;
 		if (w2 <= 0) w2 = 1;
-		cprintf2("%-*d", w2, fdtype[no].head);
+		cprintf2("%-*d", w2, (int)(fdtype[no].head));
 		putsep();
-		cprintf2("%-*d", w2, fdtype[no].sect);
+		cprintf2("%-*d", w2, (int)(fdtype[no].sect));
 		putsep();
-		cprintf2("%-*d", w2, fdtype[no].cyl);
+		cprintf2("%-*d", w2, (int)(fdtype[no].cyl));
 		putsep();
 
 		for (i = 0; i < MEDIADESCRSIZ; i++) {
@@ -3078,12 +3055,11 @@ int no;
 			&& fdtype[no].sect == mediadescr[i].sect
 			&& fdtype[no].cyl == mediadescr[i].cyl) break;
 		}
-		if (i >= MEDIADESCRSIZ) cprintf2("%*s", w1 - (w2 + 1) * 3, "");
-		else cprintf2("%-*.*k",
-			w1 - (w2 + 1) * 3, w1 - (w2 + 1) * 3,
-			mediadescr[i].name);
+		w1 -= (w2 + 1) * 3;
+		if (i >= MEDIADESCRSIZ) cprintf2("%*s", w1, "");
+		else cprintf2("%-*.*k", w1, w1, mediadescr[i].name);
 	}
-	len = strlen3(fdtype[no].name);
+	len = strlen2(fdtype[no].name);
 	if (len > --width) len = width;
 	return(len);
 }
@@ -3180,7 +3156,7 @@ int no;
 
 		for (;;) {
 			if (!(fdtype[no].name)) cp = NULL;
-			else cp = int2str(buf, fdtype[no].head, sizeof(buf));
+			else cp = int2str(buf, fdtype[no].head);
 
 			if (!(cp = inputcuststr(DRHED_K, 1, cp, -1))) break;
 			dev.head = n = atoi2(cp);
@@ -3192,7 +3168,7 @@ int no;
 
 		for (;;) {
 			if (!(fdtype[no].name)) cp = NULL;
-			else cp = int2str(buf, fdtype[no].sect, sizeof(buf));
+			else cp = int2str(buf, fdtype[no].sect);
 
 			if (!(cp = inputcuststr(DRSCT_K, 1, cp, -1))) break;
 			dev.sect = n = atoi2(cp);
@@ -3205,7 +3181,7 @@ int no;
 
 		for (;;) {
 			if (!(fdtype[no].name)) cp = NULL;
-			else cp = int2str(buf, fdtype[no].cyl, sizeof(buf));
+			else cp = int2str(buf, fdtype[no].cyl);
 
 			if (!(cp = inputcuststr(DRCYL_K, 1, cp, -1))) break;
 			dev.cyl = n = atoi2(cp);
@@ -3372,7 +3348,7 @@ int no;
 	mes[4] = COVWR_K;
 # endif
 	cprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, mes[no]);
-	len = strlen3(mes[no]);
+	len = strlen2(mes[no]);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 	return(len);
 }
@@ -3482,7 +3458,8 @@ char *file;
 			argc = (stree -> comm) -> argc;
 			argc = (stree -> comm) -> argc =
 				getsubst(argc, argv, &subst, &slen);
-			for (i = 0; i < argc; i++) stripquote(argv[i], 1);
+			for (i = 0; i < argc; i++)
+				stripquote(argv[i], EA_STRIPQ);
 
 			if (subst[0]) {
 				if (!argc && val[0])
@@ -3688,7 +3665,7 @@ int no;
 			if (!(file = inputcuststr(FSAVE_K, 1, RUNCOMFILE, 1)))
 				break;
 			done = 1;
-			file = evalpath(file, 1);
+			file = evalpath(file, 0);
 # ifdef	FAKEUNINIT
 			fp = NULL;	/* fake for -Wuninitialized */
 # endif
@@ -3727,7 +3704,7 @@ int no;
 			if (!(file = inputcuststr(FOVWR_K, 1, RUNCOMFILE, 1)))
 				break;
 			done = 1;
-			file = evalpath(file, 1);
+			file = evalpath(file, 0);
 			if (!*file
 			|| (Xaccess(file, F_OK) < 0 && errno == ENOENT
 			&& !yesno(FOVOK_K)))
@@ -3820,7 +3797,7 @@ int no, y, isstandout;
 			break;
 	}
 
-	len = strlen3(cp);
+	len = strlen2(cp);
 	if (len >= MAXCUSTNAM) len = MAXCUSTNAM;
 	fillline(y, MAXCUSTNAM + 1 - len);
 	if (isstandout) {
