@@ -211,8 +211,6 @@ extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
 extern int Xlstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
-extern int Xsymlink __P_((char *, char *));
-extern int Xreadlink __P_((char *, char *, int));
 extern int Xchmod __P_((char *, int));
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
@@ -230,7 +228,8 @@ extern off_t Xlseek __P_((int, off_t, int));
 # endif	/* !_NODOSDRIVE */
 extern int Xmkdir __P_((char *, int));
 extern int Xrmdir __P_((char *));
-extern int chdir3 __P_((char *));
+extern int stat2 __P_((char *, struct stat *));
+extern int chdir2 __P_((char *));
 #else	/* !FD */
 # if	MSDOS
 extern DIR *Xopendir __P_((char *));
@@ -259,8 +258,10 @@ extern int Xstat __P_((char *, struct stat *));
 # define	Xlstat		lstat
 # endif
 #define	Xaccess(p, m)	(access(p, m) ? -1 : 0)
-#define	Xsymlink(o, n)	(symlink(o, n) ? -1 : 0)
-#define	Xreadlink	readlink
+# ifndef	NOSYMLINK
+# define	Xsymlink(o, n)	(symlink(o, n) ? -1 : 0)
+# define	Xreadlink	readlink
+# endif
 #define	Xchmod		chmod
 #define	Xunlink(p)	(unlink(p) ? -1 : 0)
 #define	Xrename(o, n)	(rename(o, n) ? -1 : 0)
@@ -281,9 +282,14 @@ extern int Xmkdir __P_((char *, int));
 #define	Xrmdir(p)	(rmdir(p) ? -1 : 0)
 # if	MSDOS
 extern int intcall __P_((int, __dpmi_regs *, struct SREGS *));
-extern int chdir3 __P_((char *));
+extern int chdir2 __P_((char *));
 # else
-# define	chdir3(p)	(chdir(p) ? -1 : 0)
+# define	chdir2(p)	(chdir(p) ? -1 : 0)
+# endif
+# ifdef	NOSYMLINK
+# define	stat2		Xstat
+# else
+extern int stat2 __P_((char *, struct stat *));
 # endif
 #endif	/* !FD */
 
@@ -346,6 +352,12 @@ extern char *strdup2 __P_((char *));
 extern int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 extern char *realpath2 __P_((char *, char *, int));
 extern int touchfile __P_((char *, struct stat *));
+#ifndef	NODIRLOOP
+extern int issamebody __P_((char *, char *));
+#endif
+#ifndef	NOSYMLINK
+extern int cpsymlink __P_((char *, char *));
+#endif
 extern int safewrite __P_((int, char *, int));
 extern char *inputstr __P_((char *, int, int, char *, int));
 #else	/* !FD */
@@ -372,6 +384,12 @@ static VOID NEAR ttymode __P_((int));
 static int NEAR getkey2 __P_((int));
 # endif	/* !MSDOS */
 static int NEAR touchfile __P_((char *, struct stat *));
+#ifndef	NODIRLOOP
+static int NEAR issamebody __P_((char *, char *));
+#endif
+#ifndef	NOSYMLINK
+static int NEAR cpsymlink __P_((char *, char *));
+#endif
 static int NEAR safewrite __P_((int, char *, int));
 static char *NEAR inputstr __P_((char *, int, int, char *, int));
 #endif	/* !FD */
@@ -435,7 +453,7 @@ extern char *sys_errlist[];
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
 
-static char *doserrstr[] = {
+static CONST char *doserrstr[] = {
 	"",
 #define	ER_REQPARAM	1
 	"Required parameter missing",
@@ -692,6 +710,38 @@ struct stat *stp;
 # endif
 }
 
+# ifndef	NODIRLOOP
+static int NEAR issamebody(src, dest)
+char *src, *dest;
+{
+	struct stat st1, st2;
+
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
+}
+# endif	/* !NODIRLOOP */
+
+# ifndef	NOSYMLINK
+static int NEAR cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+#  ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+#  endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+
+	path[len] = '\0';
+	return(Xsymlink(path, dest));
+}
+# endif	/* !NOSYMLINK */
+
 static int NEAR safewrite(fd, buf, size)
 int fd;
 char *buf;
@@ -734,7 +784,7 @@ int h;
 		if (c != K_BS) {
 			cp = c_realloc(cp, i, &size);
 			cp[i] = c;
-			if (!isctl(c)) fputc(c, stdout);
+			if (!iscntrl2(c)) fputc(c, stdout);
 			else fprintf2(stdout, "^%c", (c + '@') & 0x7f);
 			fflush(stdout);
 		}
@@ -742,7 +792,7 @@ int h;
 		else {
 			i -= 2;
 			fputs("\b \b", stdout);
-			if (isctl(cp[i + 1])) fputs("\b \b", stdout);
+			if (iscntrl2(cp[i + 1])) fputs("\b \b", stdout);
 			fflush(stdout);
 		}
 	}
@@ -1038,14 +1088,14 @@ int len, lower;
 	if (len < 0) len = -len;
 
 	for (i = 0; s[i] && len > 0; i++, len--) {
-#ifdef	CODEEUC
-		if (isekana(s, i)) i++;
-#endif
 		if (iskanji1(s, i)) {
 			if (len <= 1) break;
 			i++;
 			len--;
 		}
+#ifdef	CODEEUC
+		else if (isekana(s, i)) i++;
+#endif
 	}
 	if (olen >= 0) while (len-- > 0) s[i++] = ' ';
 	else olen = -olen - len;
@@ -1249,8 +1299,7 @@ static VOID NEAR dosdirheader(VOID_A)
 			if (checkline(1) < 0) return;
 		}
 		else fputc(' ', stdout);
-		fprintf2(stdout, "Directory of %k", dirwd);
-		fputnl(stdout);
+		fprintf2(stdout, "Directory of %k\n", dirwd);
 	}
 #ifndef	MINIMUMSHELL
 	if (dirtype == 'V') {
@@ -1322,7 +1371,7 @@ off_t *sump, *bsump, *fp, *tp;
 		fputs("                  ", stdout);
 		fputsize(tp, bsizep);
 		fprintf2(stdout, " total disk space, %3d%% in use\n",
-			((*tp - *fp) * (off_t)100) / *tp);
+			(int)(((*tp - *fp) * (off_t)100) / *tp));
 	}
 #endif	/* !MINIMUMSHELL */
 }
@@ -1364,7 +1413,7 @@ off_t *sump, *bsump;
 	dirlist = NULL;
 	while ((dp = Xreaddir(dirp))) {
 		strcpy(file, dp -> d_name);
-		if (Xstat(dirwd, &st) < 0) continue;
+		if (stat2(dirwd, &st) < 0) continue;
 
 		dirlist = b_realloc(dirlist, n, struct filestat_t);
 		dirlist[n].nam = strdup2(dp -> d_name);
@@ -1568,7 +1617,7 @@ char *argv[];
 	}
 	if (dir != buf) strcpy(wd, cwd);
 	else {
-		if (chdir3(buf) < 0) {
+		if (chdir2(buf) < 0) {
 			dosperror(buf);
 			return(RET_FAIL);
 		}
@@ -1576,7 +1625,7 @@ char *argv[];
 			dosperror(NULL);
 			return(RET_FAIL);
 		}
-		if (chdir3(cwd) < 0) {
+		if (chdir2(cwd) < 0) {
 			dosperror(cwd);
 			return(RET_FAIL);
 		}
@@ -1702,7 +1751,7 @@ char *argv[];
 		} while (!strchr("ynYN", key));
 		if (key == 'n' || key == 'N') return(RET_SUCCESS);
 	}
-	if (!(wild = evalwild(argv[n]))) {
+	if (!(wild = evalwild(argv[n], 0))) {
 		doserror(argv[n], ER_FILENOTFOUND);
 		return(RET_FAIL);
 	}
@@ -1722,7 +1771,7 @@ char *argv[];
 			} while (!strchr("ynYN", key));
 			if (key == 'n' || key == 'N') continue;
 		}
-		if (Xunlink(wild[i]) != 0) {
+		if (Xunlink(wild[i]) < 0) {
 			dosperror(wild[i]);
 			ret = RET_FAIL;
 			ERRBREAK;
@@ -1780,7 +1829,7 @@ char *argv[];
 		return(RET_FAIL);
 	}
 
-	if (!(wild = evalwild(argv[1]))) {
+	if (!(wild = evalwild(argv[1], 0))) {
 		doserror(argv[1], ER_FILENOTFOUND);
 		return(RET_FAIL);
 	}
@@ -1865,8 +1914,8 @@ int bin;
 			return(-1);
 		}
 		*(cp++) = '\0';
-		if (*cp == 'B' || *cp == 'b') bin = CF_BINARY;
-		else if (*cp == 'A' || *cp == 'a') bin = CF_TEXT;
+		if (toupper2(*cp) == 'B') bin = CF_BINARY;
+		else if (toupper2(*cp) == 'A') bin = CF_TEXT;
 	}
 	return(bin);
 }
@@ -1916,6 +1965,14 @@ char *file, *src;
 
 	flags = (copyflag & CF_VERIFY) ? O_RDWR : O_WRONLY;
 	flags |= (O_BINARY | O_CREAT | O_TRUNC);
+
+#ifndef	NODIRLOOP
+	if (issamebody(src, file)) {
+		flags |= O_EXCL;
+		if (Xunlink(file) < 0) return(-1);
+	}
+#endif
+
 	return(Xopen(file, flags, 0666));
 }
 
@@ -1977,12 +2034,8 @@ int sbin, dbin, dfd;
 	int i, n, fd1, fd2, tty, retry, duperrno;
 
 #ifndef	NOSYMLINK
-	if (dfd < 0 && (stp -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, (char *)buf, BUFSIZ - 1)) < 0)
-			return(-1);
-		buf[i] = '\0';
-		return(Xsymlink((char *)buf, dest) >= 0);
-	}
+	if (dfd < 0 && (stp -> st_mode & S_IFMT) == S_IFLNK)
+		return(cpsymlink(src, dest));
 #endif
 
 	if ((fd1 = Xopen(src, O_TEXT | O_RDONLY, stp -> st_mode)) < 0)
@@ -2064,7 +2117,7 @@ int sbin, dbin, dfd;
 	}
 
 #if	!MSDOS && defined (UF_SETTABLE) && defined (SF_SETTABLE)
-	stp -> st_flags = 0xffffffff;
+	stp -> st_flags = (u_long)-1;
 #endif
 	if (!tty && touchfile(dest, stp) < 0) return(-1);
 	return(1);
@@ -2191,7 +2244,7 @@ char *argv[];
 		copyflag |= CF_VERBOSE;
 	}
 	for (n = 0; n < nf; n++) {
-		if (!(wild = evalwild(arg[n]))) {
+		if (!(wild = evalwild(arg[n], 0))) {
 			if (nf > 1) continue;
 			doserror(arg[n], ER_FILENOTFOUND);
 			ret = RET_FAIL;
