@@ -25,18 +25,13 @@ extern char *preparefile __P_((char *, char *));
 #endif
 
 #ifndef	_NODOSDRIVE
-extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
-extern int shutdrv __P_((int));
 extern int lastdrv;
-# if	MSDOS
+# if	MSDOS && !defined (_NOUSELFN)
 extern int checkdrive __P_((int));
 # endif
 #endif
 
 extern int mark;
-extern off_t marksize;
-extern off_t totalsize;
-extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
 extern int physical_path;
@@ -152,8 +147,8 @@ namelist *namep;
 	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
 	namep -> flags = 0;
-	if ((st.st_mode & S_IFMT) == S_IFDIR) namep -> flags |= F_ISDIR;
-	if ((lst.st_mode & S_IFMT) == S_IFLNK) namep -> flags |= F_ISLNK;
+	if (s_isdir(&st)) namep -> flags |= F_ISDIR;
+	if (s_islnk(&lst)) namep -> flags |= F_ISLNK;
 
 	if (isdisplnk(dispmode))
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
@@ -184,10 +179,6 @@ namelist *namep;
 		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
 #endif
 
-	if (isfile(namep) && namep -> st_size) {
-		totalsize += getblock(namep -> st_size);
-		if (ismark(namep)) marksize += getblock(namep -> st_size);
-	}
 	namep -> tmpflags |= F_STAT;
 	return(0);
 }
@@ -333,8 +324,7 @@ char *arcregstr;
 	}
 #ifndef	_NOARCHIVE
 	else if (arcregstr) while ((dp = Xreaddir(dirp))) {
-		if (stat2(dp -> d_name, &st) < 0
-		|| (st.st_mode & S_IFMT) == S_IFDIR) continue;
+		if (stat2(dp -> d_name, &st) < 0 || s_isdir(&st)) continue;
 
 		tmp.name = dp -> d_name;
 		tmp.flags = 0;
@@ -376,22 +366,18 @@ char *buf;
 	if (buf && !physical_path) strcpy(cwd, fullpath);
 	len = strlen(homedir);
 #if	MSDOS
-	if (len <= 3 || strnpathcmp(cwd, homedir, len)) {
+	if (len <= 3) cp = NULL;
 #else
-	if (len <= 1 || strnpathcmp(cwd, homedir, len)) {
+	if (len <= 1) cp = NULL;
 #endif
-		cp = NULL;
-		if (buf) *buf = '\0';
-	}
-	else {
-		cp = cwd + len;
-		if (buf) strcpy(buf, cp);
-	}
+	else cp = underpath(cwd, homedir, len);
+
+	if (buf) strcpy(buf, (cp) ? cp : "");
 #ifdef	DEBUG
 	free(homedir);
 	homedir = NULL;
 #endif
-	return(cp ? 1 : 0);
+	return((cp) ? 1 : 0);
 }
 
 int preparedir(dir)
@@ -422,7 +408,7 @@ char *dir;
 			}
 			else if (stat2(dir, &st) < 0) return(-1);
 		}
-		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+		if (!s_isdir(&st)) {
 			errno = ENOTDIR;
 			return(-1);
 		}
@@ -440,7 +426,7 @@ struct stat *stp;
 	int ret, duperrno;
 
 	if (Xlstat(path, &st) < 0) return(-1);
-	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
+	if (s_islnk(&st)) return(1);
 
 	ret = 0;
 #ifdef	FAKEUNINIT
@@ -473,7 +459,7 @@ struct stat *stp;
 #endif
 	}
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != 0xffffffff) {
+	if (stp -> st_flags != (u_long)-1) {
 # ifndef	_NODOSDRIVE
 		if (dospath(path, NULL)) {
 			duperrno = EACCES;
@@ -487,7 +473,7 @@ struct stat *stp;
 		}
 	}
 #endif
-	dummy.st_mode = 0xffff;
+	dummy.st_mode = (u_short)-1;
 	if (stp -> st_mode != dummy.st_mode) {
 		stp -> st_mode &= ~S_IFMT;
 		stp -> st_mode |= (st.st_mode & S_IFMT);
@@ -596,9 +582,44 @@ char *path;
 
 	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
+#ifndef	_NOUSEHASH
+	searchhash(NULL, "", "");
+#endif
 	errno = duperrno;
 }
 
+#ifndef	NODIRLOOP
+int issamebody(src, dest)
+char *src, *dest;
+{
+	struct stat st1, st2;
+
+	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
+}
+#endif	/* !NODIRLOOP */
+
+#ifndef	NOSYMLINK
+int cpsymlink(src, dest)
+char *src, *dest;
+{
+	struct stat st;
+	char path[MAXPATHLEN];
+	int len;
+
+	if ((len = Xreadlink(src, path, sizeof(path) - 1)) < 0) return(-1);
+	if (Xlstat(dest, &st) >= 0) {
+# ifndef	NODIRLOOP
+		if (issamebody(src, dest)) return(0);
+# endif
+		if (Xunlink(dest) < 0) return(-1);
+	}
+
+	path[len] = '\0';
+	return(Xsymlink(path, dest));
+}
+#endif	/* !NOSYMLINK */
+
 int safewrite(fd, buf, size)
 int fd;
 char *buf;
@@ -627,29 +648,35 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, duperrno;
+	int i, fd1, fd2, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
 	if (stp2 && !(stp2 -> st_mode & S_IWRITE))
 		Xchmod(src, stp2 -> st_mode | S_IWRITE);
 #endif
-	if ((stp1 -> st_mode & S_IFMT) == S_IFLNK) {
-		if ((i = Xreadlink(src, buf, BUFSIZ - 1)) < 0) return(-1);
-		buf[i] = '\0';
-		return(Xsymlink(buf, dest));
-	}
-	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
-		return(-1);
-#if	MSDOS
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
-		stp1 -> st_mode | S_IWRITE);
-#else
-	fd2 = Xopen(dest,
-		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
+#ifndef	NOSYMLINK
+	if (s_islnk(stp1)) return(cpsymlink(src, dest));
 #endif
-	if (fd2 < 0) {
+
+	flags = (O_BINARY | O_RDONLY);
+	mode = stp1 -> st_mode;
+	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
+
+	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
+#if	MSDOS
+	mode |= S_IWRITE;
+#endif
+#ifndef	NODIRLOOP
+	if (issamebody(src, dest)) {
+		flags |= O_EXCL;
+		if (Xunlink(dest) < 0) {
+			Xclose(fd1);
+			return(-1);
+		}
+	}
+#endif
+	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
 		Xclose(fd1);
 		return(-1);
 	}
@@ -674,7 +701,7 @@ struct stat *stp1, *stp2;
 		return(-1);
 	}
 #ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
+	stp1 -> st_flags = (u_long)-1;
 #endif
 #ifdef	_USEDOSCOPY
 	if (touchfile(dest, stp1) < 0) return(-1);
@@ -726,12 +753,11 @@ char *src, *dest;
 struct stat *stp1, *stp2;
 {
 	if (Xrename(src, dest) >= 0) return(0);
-	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
-		return(-1);
+	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
 #ifdef	HAVEFLAGS
-	stp1 -> st_flags = 0xffffffff;
+	stp1 -> st_flags = (u_long)-1;
 #endif
 	return (touchfile(dest, stp1));
 }
@@ -803,7 +829,7 @@ char *dir;
 	}
 	realpath2(deftmpdir, path, 1);
 	free(deftmpdir);
-#if	MSDOS && !defined (_NODOSDRIVE)
+#if	MSDOS && !defined (_NOUSELFN) && !defined (_NODOSDRIVE)
 	if (checkdrive(toupper2(path[0]) - 'A') && !nodosgetwd(path)) {
 		lostcwd(path);
 		deftmpdir = NULL;
@@ -813,7 +839,7 @@ char *dir;
 	deftmpdir = strdup2(path);
 #if	MSDOS
 	n = getcurdrv();
-	*path = (n >= 'A' && n <= 'Z') ? toupper2(*path) : tolower2(*path);
+	*path = (isupper2(n)) ? toupper2(*path) : tolower2(*path);
 #endif
 
 	mask = 0777 & ~tmpumask;
@@ -993,7 +1019,7 @@ char *dir, *file;
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
-	if ((pid = fork()) < 0L) return(-1);
+	if ((pid = fork()) < (p_id_t)0) return(-1);
 	else if (!pid) {
 		execle("/bin/rm", "rm", "-rf", buf, NULL, environ);
 		_exit(1);
@@ -1315,21 +1341,21 @@ int fs;
 
 	switch (fs) {
 #if	!MSDOS
-		case 2:	/* IRIX File System */
+		case FSID_EFS:		/* IRIX File System */
 			dos = 0;
 			headbyte = 4;
 			boundary = 2;
 			dirsize = sizeof(u_long);
 			namofs = 0;
 			break;
-		case 3:	/* SystemV R3 File System */
+		case FSID_SYSV:		/* SystemV R3 File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 8;
 			dirsize = sizeof(u_short);
 			namofs = 0;
 			break;
-		case 6:	/* Linux File System */
+		case FSID_LINUX:	/* Linux File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 4;
@@ -1337,7 +1363,7 @@ int fs;
 			namofs = 3;
 			break;
 # ifndef	_NODOSDRIVE
-		case 7:	/* Windows95 File System on DOSDRIVE */
+		case FSID_DOSDRIVE:	/* Windows95 File System on DOSDRIVE */
 			dos = 3;
 			headbyte = -1;
 			boundary = LFNENTSIZ;
@@ -1346,7 +1372,7 @@ int fs;
 			break;
 # endif
 #endif	/* !MSDOS */
-		case 4:	/* MS-DOS File System */
+		case FSID_FAT:		/* MS-DOS File System */
 			dos = 1;
 #if	!MSDOS
 			headbyte = -1;
@@ -1355,7 +1381,7 @@ int fs;
 			dirsize = DOSDIRENT;
 			namofs = 0;
 			break;
-		case 5:	/* Windows95 File System */
+		case FSID_LFN:		/* Windows95 File System */
 			dos = 2;
 #if	!MSDOS
 			headbyte = -1;
@@ -1459,7 +1485,7 @@ int fs;
 	}
 
 #if	!MSDOS
-	if (fs != 2) entnum = NULL;	/* except IRIX File System */
+	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
 	else if (!(entnum = getentnum(".", persec))) {
 		warning(-1, ".");
 		restorefile(tmpdir, path, fnamp);
@@ -1485,13 +1511,13 @@ int fs;
 		size = realdirsiz(fnamelist[i],
 			dos, boundary, dirsize, namofs);
 		switch (fs) {
-			case 2:	/* IRIX File System */
+			case FSID_EFS:	/* IRIX File System */
 				if (totalptr > ptr + 1) ent -= totalptr;
 				else ent -= ptr + 1;
 				break;
-			case 3:	/* SystemV R3 File System */
-			case 4:	/* MS-DOS File System */
-			case 5:	/* Windows95 File System */
+			case FSID_SYSV:	/* SystemV R3 File System */
+			case FSID_FAT:	/* MS-DOS File System */
+			case FSID_LFN:	/* Windows95 File System */
 				ent = size;
 				break;
 			default:
