@@ -7,7 +7,6 @@
 #include <errno.h>
 #include "fd.h"
 #include "funcno.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifdef	NOERRNO
@@ -201,12 +200,6 @@ extern VOID warning __P_((int, char *));
 extern int dospath __P_((char *, char *));
 #endif
 #if	MSDOS && !defined (_NOUSELFN)
-# ifdef	LSI_C
-# define	toupper2	toupper
-# else	/* !LSI_C */
-# define	toupper2(c)	(uppercase[(u_char)(c)])
-extern u_char uppercase[256];
-# endif	/* !LSI_C */
 extern int supportLFN __P_((char *));
 # ifndef	_NODOSDRIVE
 extern int checkdrive __P_((int));
@@ -217,6 +210,9 @@ extern int dosstatfs __P_((int, char *));
 #endif
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
+#if	!MSDOS
+extern int Xstat __P_((char *, struct stat *));
+#endif
 extern int Xaccess __P_((char *, int));
 
 extern bindtable bindlist[];
@@ -259,6 +255,9 @@ extern int needbavail;
 #ifndef	MNTTYPE_EXT2
 #define	MNTTYPE_EXT2	"ext2"	/* Linux */
 #endif
+#ifndef	MNTTYPE_EXT3
+#define	MNTTYPE_EXT3	"ext3"	/* Linux */
+#endif
 #ifndef	MNTTYPE_JFS
 #define	MNTTYPE_JFS	"jfs"	/* AIX */
 #endif
@@ -311,17 +310,54 @@ off_t calcKB __P_((off_t, off_t));
 int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 int infofs __P_((char *));
 
-static int keycodelist[] = {
+static CONST int keycodelist[] = {
 	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
 	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER, K_HELP,
 	K_BS, '\t', K_CR, K_ESC
 };
 #define	KEYCODESIZ	((int)(sizeof(keycodelist) / sizeof(int)))
-static char *keystrlist[] = {
+static CONST char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
 	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter", "Help",
 	"Bs", "Tab", "Ret", "Esc"
 };
+static CONST strtable mntlist[] = {
+	{FSID_FAT, MNTTYPE_PC},
+	{FSID_LFN, MNTTYPE_DOS7},
+	{FSID_LFN, MNTTYPE_FAT32},
+#if	MSDOS
+# ifndef	_NODOSDRIVE
+	{FSID_FAT, MNTTYPE_FAT12},
+	{FSID_FAT, MNTTYPE_FAT16},
+	{FSID_FAT, MNTTYPE_FAT16X},
+# endif
+#else	/* !MSDOS */
+	{FSID_UFS, MNTTYPE_43},
+	{FSID_UFS, MNTTYPE_42},
+	{FSID_UFS, MNTTYPE_UFS},
+	{FSID_UFS, MNTTYPE_FFS},
+	{FSID_UFS, MNTTYPE_JFS},
+	{FSID_EFS, MNTTYPE_EFS},
+	{FSID_SYSV, MNTTYPE_SYSV},
+	{FSID_SYSV, MNTTYPE_DGUX},
+	{FSID_LINUX, MNTTYPE_EXT2},
+	{FSID_LINUX, MNTTYPE_EXT3},
+# ifdef	LINUX
+	{FSID_FAT, MNTTYPE_MSDOS},
+# else
+	{FSID_LFN, MNTTYPE_MSDOS},
+# endif
+	{FSID_LFN, MNTTYPE_UMSDOS},
+	{FSID_LFN, MNTTYPE_VFAT},
+	{0, MNTTYPE_ADVFS},
+	{0, MNTTYPE_VXFS},
+#endif	/* !MSDOS */
+#ifdef	DARWIN
+	/* Macintosh HFS+ is pseudo file system covered with skin */
+	{0, MNTTYPE_HFS},
+#endif
+};
+#define	MNTLISTSIZ	((int)(sizeof(mntlist) / sizeof(strtable)))
 
 
 static int NEAR code2str(buf, code)
@@ -331,7 +367,7 @@ int code;
 	buf = buf + strlen(buf);
 	if (code >= K_F(1) && code <= K_F(20))
 		snprintf2(buf, KEYWID + 1, "F%-6d", code - K_F0);
-	else if ((code & ~0x7f) == 0x80 && isalpha(code & 0x7f))
+	else if ((code & ~0x7f) == 0x80 && isalpha2(code & 0x7f))
 		snprintf2(buf, KEYWID + 1, "Alt-%c  ", code & 0x7f);
 	else if (code == K_UP)
 		snprintf2(buf, KEYWID + 1, "%-*.*s", KEYWID, KEYWID, UPAR_K);
@@ -349,14 +385,20 @@ int code;
 		if (i < KEYCODESIZ)
 			snprintf2(buf, KEYWID + 1, "%-*.*s",
 				KEYWID, KEYWID, keystrlist[i]);
+#ifdef	CODEEUC
+		else if (isekana2(code))
+			snprintf2(buf, KEYWID + 1 + 1,
+				"'%c%c'    ", C_EKANA, code & 0xff);
+#endif
+		else if (code > MAXUTYPE(u_char)) return(0);
 #ifndef	CODEEUC
-		else if (iskna(code))
+		else if (iskana2(code))
 			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 #endif
-		else if (isctl(code))
+		else if (iscntrl2(code))
 			snprintf2(buf, KEYWID + 1,
 				"Ctrl-%c ", (code + '@') & 0x7f);
-		else if (code < K_MIN && !ismsb(code))
+		else if (!ismsb(code))
 			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 		else return(0);
 	}
@@ -376,7 +418,7 @@ int y;
 VOID help(mode)
 int mode;
 {
-	char buf[20];
+	char buf[(KEYWID + 1 + 1) * 2 + 1];
 	int i, j, c, x, y;
 
 	if (distributor && *distributor) {
@@ -653,11 +695,16 @@ char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
+	statfs_t fs;
+	mnt_t mnt;
 #if	MSDOS
 # ifndef	_NODOSDRIVE
 	int i;
 # endif
 
+	if (!fsbuf) fsbuf = &fs;
+	if (!mntbuf) mntbuf = &mnt;
+
 	mntbuf -> mnt_fsname = "MSDOS";
 	mntbuf -> mnt_dir[0] = dospath(path, NULL);
 	mntbuf -> mnt_dir[1] = ':';
@@ -703,9 +750,6 @@ mnt_t *mntbuf;
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) return(-1);
 #else	/* !MSDOS */
-# if	!defined (USEMNTENTH) && !defined (USEGETFSENT)
-	mnt_t mnt;
-# endif
 	mnt_t *mntp;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN];
@@ -714,8 +758,11 @@ mnt_t *mntbuf;
 	int drv;
 #endif
 
+	if (!fsbuf) fsbuf = &fs;
+	if (!mntbuf) mntbuf = &mnt;
+
 	dir = NULL;
-	if (!strncmp(path, "/dev/", 4)) {
+	if (!strncmp(path, "/dev/", sizeof("/dev/") - 1)) {
 		if (_chdir2(path) < 0) dir = path;
 	}
 #ifndef	_NODOSDRIVE
@@ -727,8 +774,8 @@ mnt_t *mntbuf;
 		mntbuf -> mnt_dir = dosmntdir;
 		dosmntdir[0] = drv;
 		strcpy(&(dosmntdir[1]), ":\\");
-		mntbuf -> mnt_type = (drv >= 'a' && drv <= 'z')
-			? MNTTYPE_DOS7 : MNTTYPE_PC;
+		mntbuf -> mnt_type =
+			(islower2(drv)) ? MNTTYPE_DOS7 : MNTTYPE_PC;
 		mntbuf -> mnt_opts = "";
 		if (dosstatfs(drv, buf) < 0) return(-1);
 		if (buf[sizeof(long) * 3] & 001)
@@ -749,6 +796,7 @@ mnt_t *mntbuf;
 		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
 # endif	/* !USEFSDATA */
 		fsbuf -> f_files = -1L;
+
 		return(0);
 	}
 #endif	/* !_NODOSDRIVE */
@@ -759,7 +807,7 @@ mnt_t *mntbuf;
 		if (_chdir2(fullpath) < 0) error(fullpath);
 		match = 0;
 
-		fp = setmntent(MOUNTED, "r");
+		fp = setmntent(MOUNTED, "rb");
 		for (;;) {
 #ifdef	DEBUG
 			_mtrace_file = "getmntent(start)";
@@ -789,7 +837,7 @@ mnt_t *mntbuf;
 		}
 		dir = fsname;
 	}
-	fp = setmntent(MOUNTED, "r");
+	fp = setmntent(MOUNTED, "rb");
 	while ((mntp = getmntent2(fp, &mnt)))
 		if (!strcmp(dir, mntp -> mnt_fsname)) break;
 	endmntent(fp);
@@ -802,6 +850,7 @@ mnt_t *mntbuf;
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0
 	&& (path == dir || statfs2(path, fsbuf) < 0)) return(-1);
 #endif	/* !MSDOS */
+
 	return(0);
 }
 
@@ -823,54 +872,25 @@ char *s1, *s2;
 int writablefs(path)
 char *path;
 {
-	statfs_t fsbuf;
-	mnt_t mntbuf;
 #ifdef	_USEDOSEMU
 	int drv;
 #endif
+	mnt_t mntbuf;
+	int i;
 
 	if (Xaccess(path, R_OK | W_OK | X_OK) < 0) return(-1);
 #ifdef	_USEDOSEMU
 	if ((drv = dospath(path, NULL)))
-		return((drv >= 'A' && drv <= 'Z') ? 4 : 7);
+		return((isupper2(drv)) ? FSID_FAT : FSID_DOSDRIVE);
 #endif
-	if (getfsinfo(path, &fsbuf, &mntbuf) < 0
-	|| hasmntopt2(&mntbuf, "ro")) return(0);
+	if (getfsinfo(path, NULL, &mntbuf) < 0
+	|| hasmntopt2(&mntbuf, "ro"))
+		return(0);
+
+	for (i = 0; i < MNTLISTSIZ; i++)
+		if (!strcmp(mntbuf.mnt_type, mntlist[i].str))
+			return(mntlist[i].no);
 
-	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT32)) return(5);
-#if	MSDOS
-# ifndef	_NODOSDRIVE
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_FAT12)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT16)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT16X)) return(4);
-# endif
-#else	/* !MSDOS */
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_42)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_UFS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FFS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_JFS)) return(1);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EFS)) return(2);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_SYSV)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_DGUX)) return(3);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EXT2)) return(6);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_MSDOS))
-# ifdef	LINUX
-		return(4);
-# else
-		return(5);
-# endif
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_UMSDOS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_VFAT)) return(5);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_ADVFS)) return(0);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_VXFS)) return(0);
-#endif	/* !MSDOS */
-#ifdef	DARWIN
-	/* Macintosh HFS+ is pseudo file system covered with skin */
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_HFS)) return(0);
-#endif
 	return(0);
 }
 
@@ -885,18 +905,26 @@ char *dir;
 	return((off_t)blocksize(fsbuf));
 #else	/* !MSDOS */
 	statfs_t fsbuf;
-	mnt_t mntbuf;
-# ifndef	DEV_BSIZE
 	struct stat st;
-# endif
 
-	if (!strcmp(dir, ".") && getfsinfo(dir, &fsbuf, &mntbuf) >= 0)
-		return((off_t)blocksize(fsbuf));
+	if (strcmp(dir, ".")) {
+		/* In case of the newborn directory */
+		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_size));
+	}
+	else {
+# ifdef	CYGWIN
+		if (statfs2(dir, &fsbuf) >= 0)
+# else
+		if (getfsinfo(dir, &fsbuf, NULL) >= 0)
+# endif
+			return((off_t)blocksize(fsbuf));
+		if (Xstat(dir, &st) >= 0) return((off_t)(st.st_blksize));
+	}
+
 # ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
 # else
-	if (Xstat(dir, &st) < 0) error(dir);
-	return((off_t)(st.st_size));
+	return((off_t)512);
 # endif
 #endif	/* !MSDOS */
 }
@@ -928,7 +956,7 @@ off_t block, byte;
 	if (block < (off_t)0 || byte <= (off_t)0) return((off_t)-1);
 	if (byte == (off_t)1024) return(block);
 	else if (byte > 1024) {
-		byte = (byte + (off_t)512L) / (off_t)1024;
+		byte = (byte + (off_t)512) / (off_t)1024;
 		return(block * byte);
 	}
 	else {
@@ -942,13 +970,12 @@ char *path;
 off_t *totalp, *freep, *bsizep;
 {
 	statfs_t fsbuf;
-	mnt_t mntbuf;
 	int n;
 
 #ifndef	_NODOSDRIVE
 	needbavail++;
 #endif
-	n = getfsinfo(path, &fsbuf, &mntbuf);
+	n = getfsinfo(path, &fsbuf, NULL);
 #ifndef	_NODOSDRIVE
 	needbavail--;
 #endif
