@@ -85,6 +85,7 @@ extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strdup2 __P_((char *));
 extern char *strndup2 __P_((char *, int));
+extern char *strchr2 __P_((char *, int));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
 # ifdef	_USEDOSPATH
@@ -95,6 +96,7 @@ extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
+extern int stat2 __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 # if	MSDOS && !defined (_NOUSELFN)
 extern char *shortname __P_((char *, char *));
@@ -109,49 +111,58 @@ char *realloc2 __P_((VOID_P, ALLOC_T));
 char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 char *strdup2 __P_((char *));
 char *strndup2 __P_((char *, int));
+char *strchr2 __P_((char *, int));
 char *strcpy2 __P_((char *, char *));
 char *strncpy2 __P_((char *, char *, int));
 # if	MSDOS
-# define	_dospath(s)	(isalpha(*(s)) && (s)[1] == ':')
+int _dospath __P_((char *));
 DIR *Xopendir __P_((char *));
 int Xclosedir __P_((DIR *));
 struct dirent *Xreaddir __P_((DIR *));
 static u_int NEAR getdosmode __P_((u_int));
 static time_t NEAR getdostime __P_((u_int, u_int));
 int Xstat __P_((char *, struct stat *));
-# else
+#  ifndef	NOSYMLINK
+#  define	Xlstat		Xstat
+#  endif
+# else	/* !MSDOS */
 # define	Xopendir	opendir
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
 # define	Xstat		stat
+#  ifndef	NOSYMLINK
+#  define	Xlstat		lstat
+#  endif
+# endif	/* !MSDOS */
+# ifdef	NOSYMLINK
+# define	stat2		Xstat
+# else
+int stat2 __P_((char *, struct stat *));
 # endif
 # ifdef	DJGPP
 char *Xgetwd __P_((char *));
-# else
+# else	/* !DJGPP */
 #  ifdef	USEGETWD
 #  define	Xgetwd		(char *)getwd
 #  else
 #  define	Xgetwd(p)	(char *)getcwd(p, MAXPATHLEN)
 #  endif
-# endif
+# endif	/* !DJGPP */
 #define	Xaccess		access
 #endif	/* !FD */
 
 static char *NEAR getenvvar __P_((char *, int));
 static int NEAR setvar __P_((char *, char *, int));
-static int NEAR ismeta __P_((char *s, int, int, int));
+static int NEAR ismeta __P_((char *s, int, int, int, int));
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp __P_((char *, int));
 #else
 static int NEAR _regexp_exec __P_((char **, char *));
 #endif
-static char *NEAR catpath __P_((char *, char *, int *, int, int));
-#ifdef	NODIRLOOP
-static int NEAR _evalwild __P_((int, char ***, char *, char *, int));
-#else
-static int NEAR _evalwild __P_((int, char ***, char *, char *, int,
-		int, devino_t *));
-#endif
+static void NEAR addstrbuf __P_((strbuf_t *, char *, int));
+static void NEAR duplwild __P_((wild_t *, wild_t *));
+static void NEAR freewild __P_((wild_t *));
+static int NEAR _evalwild __P_((int, char ***, wild_t *));
 #ifndef	_NOUSEHASH
 static int NEAR calchash __P_((char *));
 static VOID NEAR inithash __P_((VOID_A));
@@ -189,7 +200,7 @@ static int NEAR evalvar __P_((char **, int, char **, int));
 static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 
 #define	BUFUNIT		32
-#define	b_size(n, type) ((((n) / BUFUNIT) + 1) * BUFUNIT * sizeof(type))
+#define	b_size(n, type)	((((n) / BUFUNIT) + 1) * BUFUNIT * sizeof(type))
 #define	b_realloc(ptr, n, type) \
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
@@ -309,6 +320,22 @@ int n;
 	return(tmp);
 }
 
+char *strchr2(s, c)
+char *s;
+int c;
+{
+	int i;
+
+	for (i = 0; s[i]; i++) {
+		if (s[i] == c) return(&(s[i]));
+		if (iskanji1(s, i)) i++;
+#ifdef	CODEEUC
+		else if (isekana(s, i)) i++;
+#endif
+	}
+	return(NULL);
+}
+
 char *strcpy2(s1, s2)
 char *s1, *s2;
 {
@@ -331,6 +358,12 @@ int n;
 }
 
 # if	MSDOS
+int _dospath(path)
+char *path;
+{
+	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
+}
+
 DIR *Xopendir(dir)
 char *dir;
 {
@@ -448,6 +481,26 @@ struct stat *stp;
 }
 # endif	/* MSDOS */
 
+# ifndef	NOSYMLINK
+int stat2(path, stp)
+char *path;
+struct stat *stp;
+{
+	int duperrno;
+
+	if (Xstat(path, stp) < 0) {
+		duperrno = errno;
+		if (Xlstat(path, stp) < 0
+		|| (stp -> st_mode & S_IFMT) != S_IFLNK) {
+			errno = duperrno;
+			return(-1);
+		}
+		stp -> st_mode &= ~S_IFMT;
+	}
+	return(0);
+}
+# endif	/* !NOSYMLINK */
+
 # ifdef	DJGPP
 char *Xgetwd(path)
 char *path;
@@ -682,6 +735,20 @@ int n;
 }
 #endif	/* !PATHNOCASE */
 
+char *underpath(path, dir, len)
+char *path, *dir;
+int len;
+{
+	char *cp;
+
+	if (len < 0) len = strlen(dir);
+	if ((cp = strrdelim2(dir, &(dir[len]))) && !*(cp + 1)) len = cp - dir;
+	if (len <= 0 || strnpathcmp(path, dir, len)) return(NULL);
+	if (path[len] && path[len] != _SC_) return(NULL);
+
+	return(&(path[len]));
+}
+
 static char *NEAR getenvvar(ident, len)
 char *ident;
 int len;
@@ -727,7 +794,13 @@ int len;
 int isidentchar(c)
 int c;
 {
-	return(c == '_' || isalpha(c));
+	return(c == '_' || isalpha2(c));
+}
+
+int isidentchar2(c)
+int c;
+{
+	return(c == '_' || isalnum2(c));
 }
 
 int isdotdir(s)
@@ -737,7 +810,6 @@ char *s;
 	return(0);
 }
 
-/*ARGSUSED*/
 char *isrootdir(s)
 char *s;
 {
@@ -780,30 +852,41 @@ int *loginp, *restrictedp;
 	return(s);
 }
 
-static int NEAR ismeta(s, ptr, quote, len)
+static int NEAR ismeta(s, ptr, quote, len, flags)
 char *s;
-int ptr, quote, len;
+int ptr, quote, len, flags;
 {
 #ifdef	FAKEMETA
 	return(0);
 #else	/* !FAKEMETA */
-	if (s[ptr] != PMETA) return(0);
+	if (s[ptr] != PMETA || quote == '\'') return(0);
+
 	if (len >= 0) {
-		if (ptr + 1 >= len) return(0);
-		if (quote && s[ptr + 1] == quote && ptr + 2 >= len) return(0);
+		if (!(flags & EA_EOLMETA) && ptr + 1 >= len) return(0);
+# ifndef	BASHSTYLE
+	/* bash does not treat "\" as \ */
+		if (quote == '"' && s[ptr + 1] == quote && ptr + 2 >= len)
+			return(0);
+# endif
 	}
 	else {
-		if (!s[ptr + 1]) return(0);
-		if (quote) {
-			if (s[ptr + 1] == quote && !s[ptr + 2]) return(0);
-# ifdef	BSPATHDELIM
-			if (!strchr(DQ_METACHAR, s[ptr + 1])) return(0);
-# endif
-		}
-# ifdef	BSPATHDELIM
-		else if (!strchr(METACHAR, s[ptr + 1])) return(0);
+		if (!(flags & EA_EOLMETA) && !s[ptr + 1]) return(0);
+# ifndef	BASHSTYLE
+	/* bash does not treat "\" as \ */
+		if (quote == '"' && s[ptr + 1] == quote && !s[ptr + 2])
+			return(0);
 # endif
 	}
+
+# ifdef	BSPATHDELIM
+	if (quote == '"') {
+		if (!strchr(DQ_METACHAR, s[ptr + 1])) return(0);
+	}
+	else {	/* if (!quote || quote == '`') */
+		if (!strchr(METACHAR, s[ptr + 1])) return(0);
+	}
+# endif
+
 	return(1);
 #endif	/* !FAKEMETA */
 }
@@ -833,7 +916,7 @@ int len;
 			re[j++] = s[i];
 			continue;
 		}
-		else if (ismeta(s, i, '\0', len)) {
+		else if (ismeta(s, i, '\0', len, 0)) {
 			re[j++] = s[i++];
 			re[j++] = s[i];
 		}
@@ -989,7 +1072,7 @@ int len;
 	ALLOC_T size;
 	int i, j, n, pc, plen, metachar, quote;
 
-	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
+	skipdotfile = 0;
 	if (len < 0) len = strlen(s);
 	re = NULL;
 	paren = NULL;
@@ -1077,34 +1160,34 @@ int len;
 #endif
 				else paren[plen++] = s[i];
 			}
-			else {
-#ifdef	CODEEUC
-				if (isekana(s, i)) {
-					paren[plen++] = s[i++];
-					paren[plen++] = s[i];
-				}
-				else
-#endif
-				if (iskanji1(s, i)) {
-					paren[plen++] = s[i++];
-					paren[plen++] = s[i];
-				}
-#ifndef	PATHNOCASE
-				else if (!pathignorecase) paren[plen++] = s[i];
-#endif
-				else paren[plen++] = toupper2(s[i]);
+			else if (iskanji1(s, i)) {
+				paren[plen++] = s[i++];
+				paren[plen++] = s[i];
 			}
+#ifdef	CODEEUC
+			else if (isekana(s, i)) {
+				paren[plen++] = s[i++];
+				paren[plen++] = s[i];
+			}
+#endif
+#ifndef	PATHNOCASE
+			else if (!pathignorecase) paren[plen++] = s[i];
+#endif
+			else paren[plen++] = toupper2(s[i]);
 		}
 		else if (!quote && !metachar) switch (s[i]) {
 			case '?':
 				cp = wildsymbol1;
+				if (!n) skipdotfile++;
 				break;
 			case '*':
 				cp = wildsymbol2;
+				if (!n) skipdotfile++;
 				break;
 			case '[':
 				paren = c_realloc(NULL, 0, &size);
 				plen = 0;
+				if (!n) skipdotfile++;
 				break;
 			default:
 				break;
@@ -1114,17 +1197,16 @@ int len;
 		if (!cp) {
 			cp = malloc2(2 + 1);
 			j = 0;
-#ifdef	CODEEUC
-			if (isekana(s, i)) {
-				cp[j++] = s[i++];
-				cp[j++] = s[i];
-			}
-			else
-#endif
 			if (iskanji1(s, i)) {
 				cp[j++] = s[i++];
 				cp[j++] = s[i];
 			}
+#ifdef	CODEEUC
+			else if (isekana(s, i)) {
+				cp[j++] = s[i++];
+				cp[j++] = s[i];
+			}
+#endif
 #ifndef	PATHNOCASE
 			else if (!pathignorecase) cp[j++] = s[i];
 #endif
@@ -1157,11 +1239,10 @@ char *s;
 
 	for (n1 = n2 = 0; re[n1] && s[n2]; n1++, n2++) {
 		c1 = (u_char)(s[i = n2]);
-#ifdef	CODEEUC
-		if (isekana(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
-		else
-#endif
 		if (iskanji1(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
+#ifdef	CODEEUC
+		else if (isekana(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
+#endif
 #ifndef	PATHNOCASE
 		else if (!pathignorecase) /*EMPTY*/;
 #endif
@@ -1195,13 +1276,12 @@ char *s;
 				i++;
 			}
 			c2 = (u_char)(re[n1][i]);
-#ifdef	CODEEUC
-			if (isekana(re[n1], i))
-				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
-			else
-#endif
 			if (iskanji1(re[n1], i))
 				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
+#ifdef	CODEEUC
+			else if (isekana(re[n1], i))
+				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
+#endif
 			if (beg >= 0) {
 				if (beg && c1 >= beg && c1 <= c2) break;
 				beg = -1;
@@ -1245,245 +1325,226 @@ reg_t *re;
 # endif		/* !USEREGCOMP */
 #endif		/* !USERE_COMP */
 
-static char *NEAR catpath(path, file, plenp, flen, isoverwrite)
-char *path, *file;
-int *plenp, flen, isoverwrite;
+static void NEAR addstrbuf(sp, s, len)
+strbuf_t *sp;
+char *s;
+int len;
 {
-	char *new;
-	int i, sc;
-
-#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-	if (*plenp >= 2 && _dospath(path)) i = 2;
-	else
-#endif
-	i = 0;
-
-	if (*plenp <= i) sc = 0;
-	else {
-		sc = 1;
-		if (path[i] == _SC_) {
-			for (i++; i < *plenp; i++) if (path[i] != _SC_) break;
-			if (i >= *plenp) sc = 0;
-		}
-	}
-
-	if (isoverwrite) new = realloc2(path, *plenp + sc + flen + 1);
-	else {
-		new = malloc2(*plenp + sc + flen + 1);
-		strncpy(new, path, *plenp);
-	}
-	if (sc) new[(*plenp)++] = _SC_;
-	strncpy2(&(new[*plenp]), file, flen);
-	*plenp += flen;
-	return(new);
+	sp -> s = c_realloc(sp -> s, sp -> len + len, &(sp -> size));
+	memcpy(&(sp -> s[sp -> len]), s, len);
+	sp -> len += len;
+	sp -> s[sp -> len] = '\0';
 }
 
-#ifdef	NODIRLOOP
-static int NEAR _evalwild(argc, argvp, s, fixed, len)
-int argc;
-char ***argvp, *s, *fixed;
-int len;
-#else
-static int NEAR _evalwild(argc, argvp, s, fixed, len, nino, ino)
-int argc;
-char ***argvp, *s, *fixed;
-int len, nino;
-devino_t *ino;
-#endif
+static void NEAR duplwild(dst, src)
+wild_t *dst, *src;
 {
+	memcpy(dst, src, sizeof(wild_t));
+	dst -> fixed.s = (char *)malloc2(src -> fixed.size);
+	memcpy(dst -> fixed.s, src -> fixed.s, src -> fixed.len + 1);
+	dst -> path.s = (char *)malloc2(src -> path.size);
+	memcpy(dst -> path.s, src -> path.s, src -> path.len + 1);
+
 #ifndef	NODIRLOOP
-	devino_t *dupino;
+	if (src -> ino) {
+		dst -> ino = (devino_t *)malloc2(src -> nino
+			* sizeof(devino_t));
+		memcpy(dst -> ino, src -> ino, src -> nino * sizeof(devino_t));
+	}
+#endif	/* !NODIRLOOP */
+}
+
+static void NEAR freewild(wp)
+wild_t *wp;
+{
+	if (wp -> fixed.s) free(wp -> fixed.s);
+	if (wp -> path.s) free(wp -> path.s);
+#ifndef	NODIRLOOP
+	if (wp -> ino) free(wp -> ino);
 #endif
+}
+
+static int NEAR _evalwild(argc, argvp, wp)
+int argc;
+char ***argvp;
+wild_t *wp;
+{
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
 	reg_t *re;
+	wild_t dupl;
+	ALLOC_T flen, plen;
 	char *cp;
-	int i, n, l;
+	int i, n, w, pc, quote, isdir;
 
-	if (!*s) {
-		if (len) {
-			fixed = realloc2(fixed, len + 1 + 1);
-			fixed[len++] = _SC_;
-			fixed[len] = '\0';
-			*argvp = (char **)realloc2(*argvp,
-				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = fixed;
-		}
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
+	if (!*(wp -> s)) {
+		if (!(wp -> fixed.len)) return(argc);
+		*argvp = (char **)realloc2(*argvp,
+			(argc + 2) * sizeof(char *));
+		(*argvp)[argc++] = wp -> fixed.s;
+		wp -> fixed.s = NULL;
 		return(argc);
 	}
 
-#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-	if (!len && _dospath(s)) {
-		fixed = malloc2(2 + 1);
-		fixed[0] = *s;
-		fixed[1] = ':';
-		fixed[2] = '\0';
-# ifdef	NODIRLOOP
-		return(_evalwild(argc, argvp, &(s[2]), fixed, 2));
-# else
-		return(_evalwild(argc, argvp, &(s[2]), fixed, 2, nino, ino));
-# endif
-	}
-#endif
+	flen = wp -> fixed.len;
+	plen = wp -> path.len;
+	quote = wp -> quote;
 
-	n = 0;
-	for (i = 0; s[i]; i++) {
-		if (s[i] == _SC_) break;
-		if (s[i] == '*' || s[i] == '?'
-		|| s[i] == '[' || s[i] == ']') n = 1;
-#ifdef	BSPATHDELIM
-		if (iskanji1(s, i)) i++;
-#endif
-	}
+	if (wp -> fixed.len) addstrbuf(&(wp -> path), _SS_, 1);
 
-	if (!i) {
-		fixed = realloc2(fixed, len + 1 + 1);
-		fixed[len++] = _SC_;
-		fixed[len] = '\0';
-#ifdef	NODIRLOOP
-		return(_evalwild(argc, argvp, &(s[1]), fixed, len));
-#else
-		return(_evalwild(argc, argvp, &(s[1]), fixed, len, nino, ino));
-#endif
-	}
+	for (i = w = 0; wp -> s[i] && wp -> s[i] != _SC_; i++) {
+		pc = parsechar(&(wp -> s[i]), -1,
+			'\0', 0, &(wp -> quote), NULL);
+		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
+			if (!(wp -> flags & EA_STRIPQ))
+				addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			continue;
+		}
+		else if (pc == PC_WORD) {
+			addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
+			i++;
+		}
+		else if (pc == PC_META) {
+			if (wp -> flags & EA_KEEPMETA)
+				addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+			if (wp -> s[i + 1] == _SC_) continue;
 
-	if (!n) {
-		fixed = catpath(fixed, s, &len, i, 1);
-		if (Xstat(fixed, &st) < 0) free(fixed);
-		else if (s[i]) {
-			if ((st.st_mode & S_IFMT) != S_IFDIR) free(fixed);
-			else {
-#ifdef	NODIRLOOP
-				return(_evalwild(argc, argvp,
-					&(s[i + 1]), fixed, len));
-#else
-				ino = (devino_t *)realloc2(ino,
-					(nino + 1) * sizeof(devino_t));
-				ino[nino].dev = st.st_dev;
-				ino[nino++].ino = st.st_ino;
-				return(_evalwild(argc, argvp,
-					&(s[i + 1]), fixed, len, nino, ino));
-#endif
+			if (wp -> quote == '\''
+			|| (wp -> quote == '"'
+			&& !strchr(DQ_METACHAR, wp -> s[i + 1]))) {
+				if (!(wp -> flags & EA_KEEPMETA))
+					addstrbuf(&(wp -> fixed),
+						&(wp -> s[i]), 1);
+				addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
 			}
+			i++;
 		}
-		else {
-			*argvp = (char **)realloc2(*argvp,
-				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = fixed;
+		else if (pc == PC_NORMAL && strchr("?*[", wp -> s[i])) {
+			if (w >= 0 && wp -> s[i] == '*') w++;
+			else w = -1;
 		}
+
+		addstrbuf(&(wp -> fixed), &(wp -> s[i]), 1);
+		addstrbuf(&(wp -> path), &(wp -> s[i]), 1);
+	}
+
+	if (!(wp -> s[i])) isdir = 0;
+	else {
+		isdir = 1;
+		addstrbuf(&(wp -> fixed), _SS_, 1);
+	}
+
+	if (!w) {
+		if (wp -> path.len <= plen) w++;
+		else if (stat2(wp -> path.s, &st) < 0) return(argc);
+
+		wp -> s += i;
+		if (isdir) {
+			if (!w && (st.st_mode & S_IFMT) != S_IFDIR)
+				return(argc);
+			(wp -> s)++;
+		}
+
 #ifndef	NODIRLOOP
-		if (ino) free(ino);
+		if (!w) {
+			wp -> ino = (devino_t *)realloc2(wp -> ino,
+				(wp -> nino + 1) * sizeof(devino_t));
+			wp -> ino[wp -> nino].dev = st.st_dev;
+			wp -> ino[(wp -> nino)++].ino = st.st_ino;
+		}
 #endif
-		return(argc);
+		return(_evalwild(argc, argvp, wp));
 	}
 
-	if (i == 2 && s[i] && s[0] == '*' && s[1] == '*') {
-#ifdef	NODIRLOOP
-		argc = _evalwild(argc, argvp, &(s[3]), strdup2(fixed), len);
-#else
-		if (!ino) dupino = NULL;
-		else {
-			dupino = (devino_t *)malloc2(nino * sizeof(devino_t));
-			for (n = 0; n < nino; n++) {
-				dupino[n].dev = ino[n].dev;
-				dupino[n].ino = ino[n].ino;
-			}
-		}
-		argc = _evalwild(argc, argvp,
-			&(s[3]), strdup2(fixed), len, nino, dupino);
-#endif
+	if (w != 2 || !isdir || strcmp(&(wp -> path.s[plen]), "**")) w = -1;
+	wp -> fixed.len = flen;
+	wp -> path.len = plen;
+	wp -> fixed.s[flen] = wp -> path.s[plen] = '\0';
+
+	if (w > 0) {
+		duplwild(&dupl, wp);
+		dupl.s += i + 1;
+		argc = _evalwild(argc, argvp, &dupl);
+		freewild(&dupl);
 		re = NULL;
 	}
-	else if (!(re = regexp_init(s, i))) {
-		if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
-		return(argc);
+	else {
+		cp = malloc2(i + 2);
+		n = 0;
+		if (quote) cp[n++] = quote;
+		memcpy(&(cp[n]), wp -> s, i);
+		n += i;
+		if (wp -> quote) cp[n++] = wp -> quote;
+		re = regexp_init(cp, n);
+		free(cp);
+		if (!re) return(argc);
+		wp -> s += i + 1;
 	}
-	if (!(dirp = Xopendir((len) ? fixed : "."))) {
+
+	if (wp -> path.len) cp = wp -> path.s;
+	else if (wp -> fixed.len) cp = _SS_;
+	else cp = ".";
+
+	if (!(dirp = Xopendir(cp))) {
 		regexp_free(re);
-		if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-		if (ino) free(ino);
-#endif
 		return(argc);
 	}
+	if (wp -> path.len || wp -> fixed.len)
+		addstrbuf(&(wp -> path), _SS_, 1);
 
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
 
-		l = len;
-		cp = catpath(fixed, dp -> d_name, &l, strlen(dp -> d_name), 0);
-		if (s[i]) {
-			if (Xstat(cp, &st) < 0
+		duplwild(&dupl, wp);
+		n = strlen(dp -> d_name);
+		addstrbuf(&(dupl.fixed), dp -> d_name, n);
+		addstrbuf(&(dupl.path), dp -> d_name, n);
+
+		if (isdir) {
+			if (re) n = regexp_exec(re, dp -> d_name, 1);
+			else n = (*(dp -> d_name) == '.') ? 0 : 1;
+
+			if (!n || Xstat(dupl.path.s, &st) < 0
 			|| (st.st_mode & S_IFMT) != S_IFDIR) {
-				free(cp);
+				freewild(&dupl);
 				continue;
 			}
 
 #ifndef	NODIRLOOP
-			dupino = (devino_t *)malloc2((nino + 1)
-				* sizeof(devino_t));
-			for (n = 0; n < nino; n++) {
-				dupino[n].dev = ino[n].dev;
-				dupino[n].ino = ino[n].ino;
-			}
-			dupino[n].dev = st.st_dev;
-			dupino[n].ino = st.st_ino;
-#endif
 			if (!re) {
-				if (*(dp -> d_name) == '.') {
-					free(cp);
-#ifndef	NODIRLOOP
-					free(dupino);
-#endif
+				for (n = 0; n < dupl.nino; n++)
+					if (dupl.ino[n].dev == st.st_dev
+					&& dupl.ino[n].ino == st.st_ino)
+						break;
+				if (n < dupl.nino) {
+					freewild(&dupl);
 					continue;
 				}
-#ifdef	NODIRLOOP
-				argc = _evalwild(argc, argvp, s, cp, l);
-#else
-				for (n = 0; n < nino; n++)
-					if (ino[n].dev == st.st_dev
-					&& ino[n].ino == st.st_ino) break;
-				if (n < nino) {
-					free(cp);
-					free(dupino);
-				}
-				else argc = _evalwild(argc, argvp,
-						s, cp, l, nino + 1, dupino);
-#endif
 			}
-			else if (!regexp_exec(re, dp -> d_name, 1)) {
-				free(cp);
-#ifndef	NODIRLOOP
-				free(dupino);
-#endif
-			}
-#ifdef	NODIRLOOP
-			else argc = _evalwild(argc, argvp, &(s[i + 1]), cp, l);
-#else
-			else argc = _evalwild(argc, argvp,
-					&(s[i + 1]), cp, l, nino + 1, dupino);
+
+			dupl.ino = (devino_t *)realloc2(dupl.ino,
+				(dupl.nino + 1) * sizeof(devino_t));
+			dupl.ino[dupl.nino].dev = st.st_dev;
+			dupl.ino[(dupl.nino)++].ino = st.st_ino;
 #endif
+
+			addstrbuf(&(dupl.fixed), _SS_, 1);
+			argc = _evalwild(argc, argvp, &dupl);
 		}
-		else if (!re || !regexp_exec(re, dp -> d_name, 1)) free(cp);
-		else {
+		else if (regexp_exec(re, dp -> d_name, 1)) {
 			*argvp = (char **)realloc2(*argvp,
 				(argc + 2) * sizeof(char *));
-			(*argvp)[argc++] = cp;
+			(*argvp)[argc++] = dupl.fixed.s;
+			dupl.fixed.s = NULL;
 		}
+
+		freewild(&dupl);
 	}
 	Xclosedir(dirp);
 	regexp_free(re);
-	if (fixed) free(fixed);
-#ifndef	NODIRLOOP
-	if (ino) free(ino);
-#endif
+
 	return(argc);
 }
 
@@ -1494,22 +1555,34 @@ CONST VOID_P vp2;
 	return(strpathcmp2(*((char **)vp1), *((char **)vp2)));
 }
 
-char **evalwild(s)
+char **evalwild(s, flags)
 char *s;
+int flags;
 {
+	wild_t w;
 	char **argv;
 	int argc;
 
 	argv = (char **)malloc2(1 * sizeof(char *));
-#ifdef	NODIRLOOP
-	argc = _evalwild(0, &argv, s, NULL, 0);
-#else
-	argc = _evalwild(0, &argv, s, NULL, 0, 0, NULL);
+	w.s = s;
+	w.fixed.s = c_realloc(NULL, 0, &(w.fixed.size));
+	w.path.s = c_realloc(NULL, 0, &(w.path.size));
+	w.fixed.len = w.path.len = (ALLOC_T)0;
+	w.quote = '\0';
+#ifndef	NODIRLOOP
+	w.nino = 0;
+	w.ino = NULL;
 #endif
+	w.flags = flags;
+
+	argc = _evalwild(0, &argv, &w);
+	freewild(&w);
+
 	if (!argc) {
 		free(argv);
 		return(NULL);
 	}
+
 	argv[argc] = NULL;
 	if (argc > 1) qsort(argv, argc, sizeof(char *), cmppath);
 	return(argv);
@@ -1636,7 +1709,7 @@ int dirok, exe;
 	struct stat st;
 	int d;
 
-	if (Xstat(path, &st) < 0) return(-1);
+	if (stat2(path, &st) < 0) return(-1);
 	d = ((st.st_mode & S_IFMT) == S_IFDIR);
 	if (!exe) return(d);
 	if (d) return(dirok ? d : -1);
@@ -1672,21 +1745,23 @@ int searchhash(hpp, com, search)
 hashlist **hpp;
 char *com, *search;
 {
-	char *cp, *tmp, *next, *path;
-	int len, dlen, cost, size, ret, recalc;
 #if	MSDOS
 	char *ext;
 #endif
+	char *cp, *tmp, *next, *path;
+	int len, dlen, cost, size, ret;
 #ifndef	_NOUSEHASH
-	static char *searchcwd = NULL;
-	char buf[MAXPATHLEN];
-	int n, duperrno;
+	hashlist *hp;
+	int n, recalc, duperrno;
 
-	if (!com && !search) {
+	if (!hpp || (!com && !search)) {
 		duperrno = errno;
-		freehash(hpp);
-		if (searchcwd) free(searchcwd);
-		searchcwd = NULL;
+		if (!com && !search) freehash(hpp);
+		if (!hpp && hashtable) for (n = 0; n < MAXHASH; n++) {
+			for (hp = hashtable[n]; hp; hp = hp -> next)
+				if (hp -> type & CM_RECALC)
+					hp -> type |= CM_REHASH;
+		}
 		errno = duperrno;
 		return(CM_NOTFOUND);
 	}
@@ -1713,11 +1788,9 @@ char *com, *search;
 	if (!hashtable) inithash();
 	n = calchash(com);
 	if ((*hpp = findhash(com, n))) {
-		path = (*hpp) -> path;
-		if (((*hpp) -> type & CM_RECALC)
-		&& (!searchcwd || !Xgetwd(buf) || strcmp(searchcwd, buf)))
-			/*EMPTY*/;
-		else if (isexecute(path, 0, 1) >= 0) return((*hpp) -> type);
+		if (!((*hpp) -> type & CM_REHASH)
+		&& isexecute((*hpp) -> path, 0, 1) >= 0)
+			return((*hpp) -> type);
 		rmhash(com, n);
 	}
 #endif
@@ -1740,14 +1813,14 @@ char *com, *search;
 		*hpp = newhash(com, path, 0, hashtable[n]);
 		hashtable[n] = *hpp;
 		(*hpp) -> type = (ret |= (CM_HASH | CM_RECALC));
-		if (searchcwd) free(searchcwd);
-		searchcwd = strdup2(Xgetwd(buf));
 # endif
 		return(ret);
 	}
 #endif	/* CWDINPATH */
 
+#ifndef	_NOUSEHASH
 	recalc = 0;
+#endif
 	if ((next = (search) ? search : getconstvar("PATH"))) {
 		len = strlen(com);
 		size = ret = 0;
@@ -1758,12 +1831,14 @@ char *com, *search;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 			if (_dospath(cp)) next += 2;
 #endif
+#ifndef	_NOUSEHASH
 			if (*next != _SC_) recalc = CM_RECALC;
+#endif
 			next = strchr(next, PATHDELIM);
 			dlen = (next) ? (next++) - cp : strlen(cp);
 			if (!dlen) tmp = NULL;
 			else {
-				tmp = _evalpath(cp, cp + dlen, 1, 1);
+				tmp = _evalpath(cp, cp + dlen, 0);
 				dlen = strlen(tmp);
 			}
 			if (dlen + len + 1 + EXTWIDTH + 1 > size) {
@@ -1791,8 +1866,6 @@ char *com, *search;
 			*hpp = newhash(com, path, cost, hashtable[n]);
 			hashtable[n] = *hpp;
 			(*hpp) -> type = (ret |= (CM_HASH | recalc));
-			if (searchcwd) free(searchcwd);
-			searchcwd = strdup2(Xgetwd(buf));
 #endif
 			return(ret);
 		}
@@ -1963,7 +2036,7 @@ char ***argvp;
 # endif
 		next = strchr(cp, PATHDELIM);
 		dlen = (next) ? (next++) - cp : strlen(cp);
-		tmp = _evalpath(cp, cp + dlen, 1, 1);
+		tmp = _evalpath(cp, cp + dlen, 0);
 		dlen = strlen(tmp);
 		argc = completefile(file, len, argc, argvp, tmp, dlen, 1);
 		free(tmp);
@@ -2026,9 +2099,9 @@ char **argv;
 }
 #endif	/* !FDSH && !_NOCOMPLETE */
 
-static int NEAR addmeta(s1, s2, stripm, quoted)
+static int NEAR addmeta(s1, s2, quoted, flags)
 char *s1, *s2;
-int stripm, quoted;
+int quoted, flags;
 {
 	int i, j;
 
@@ -2041,12 +2114,17 @@ int stripm, quoted;
 		else if (s2[i] == PMETA) {
 			if (s1) s1[j] = PMETA;
 			j++;
-			if (stripm && s2[i + 1] == PMETA) i++;
+			if (!(flags & EA_KEEPMETA) && s2[i + 1] == PMETA) i++;
 		}
 		else if (!quoted && s2[i] == '\'') {
 			if (s1) s1[j] = PMETA;
 			j++;
 		}
+		else if (iskanji1(s2, i)) {
+			if (s1) s1[j] = s2[i];
+			j++;
+			i++;
+		}
 #ifdef	CODEEUC
 		else if (isekana(s2, i)) {
 			if (s1) s1[j] = s2[i];
@@ -2054,11 +2132,6 @@ int stripm, quoted;
 			i++;
 		}
 #endif
-		else if (iskanji1(s2, i)) {
-			if (s1) s1[j] = s2[i];
-			j++;
-			i++;
-		}
 		if (s1) s1[j] = s2[i];
 	}
 	return(j);
@@ -2125,9 +2198,9 @@ int margin;
 	return(dupl);
 }
 
-int parsechar(s, len, spc, qflag, qp, pqp)
+int parsechar(s, len, spc, flags, qp, pqp)
 char *s;
-int len, spc, qflag, *qp, *pqp;
+int len, spc, flags, *qp, *pqp;
 {
 	if (*s == *qp) {
 		if (!pqp) *qp = '\0';
@@ -2137,28 +2210,42 @@ int len, spc, qflag, *qp, *pqp;
 		}
 		return(PC_CLQUOTE);
 	}
+	else if (iskanji1(s, 0)) return(PC_WORD);
 #ifdef	CODEEUC
-	if (isekana(s, 0)) return(PC_WORD);
+	else if (isekana(s, 0)) return(PC_WORD);
 #endif
-	if (iskanji1(s, 0)) return(PC_WORD);
-	if (*qp == '`') return(PC_BQUOTE);
-	if (*qp == '\'') return(PC_SQUOTE);
-	if (spc && *s == spc) return(*s);
-	if (ismeta(s, 0, *qp, len)) return(PC_META);
-	if (qflag > 0 && *s == '`') {
+#ifdef	NESTINGQUOTE
+	else if (*qp == '`') return(PC_BQUOTE);
+#endif
+	else if (*qp == '\'') return(PC_SQUOTE);
+	else if (spc && *s == spc) return(*s);
+	else if (ismeta(s, 0, *qp, len, flags)) return(PC_META);
+#ifdef	NESTINGQUOTE
+	else if ((flags & EA_BACKQ) && *s == '`') {
 		if (pqp && *qp) *pqp = *qp;
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-	if (*qp) return(PC_DQUOTE);
-	if (qflag >= 0 && *s == '\'') {
+#else
+	else if (*qp == '`') return(PC_BQUOTE);
+#endif
+	else if (*qp) return(PC_DQUOTE);
+	else if (!(flags & EA_NOEVALQ) && *s == '\'') {
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
-	if (*s == '"') {
+	else if (*s == '"') {
 		*qp = *s;
 		return(PC_OPQUOTE);
 	}
+#ifndef	NESTINGQUOTE
+	else if ((flags & EA_BACKQ) && *s == '`') {
+		if (pqp && *qp) *pqp = *qp;
+		*qp = *s;
+		return(PC_OPQUOTE);
+	}
+#endif
+
 	return(PC_NORMAL);
 }
 
@@ -2174,8 +2261,7 @@ int *eolp, *ptrp, qed;
 
 	if (isidentchar((*bufp)[*ptrp])) {
 		while ((*bufp)[++(*ptrp)])
-			if (!isidentchar((*bufp)[*ptrp])
-			&& !isdigit((*bufp)[*ptrp])) break;
+			if (!isidentchar2((*bufp)[*ptrp])) break;
 		if (eolp) *eolp = *ptrp;
 		return(mode);
 	}
@@ -2211,8 +2297,7 @@ int *eolp, *ptrp, qed;
 #endif
 	if (isidentchar((*bufp)[*ptrp])) {
 		while ((*bufp)[++(*ptrp)])
-			if (!isidentchar((*bufp)[*ptrp])
-			&& !isdigit((*bufp)[*ptrp])) break;
+			if (!isidentchar2((*bufp)[*ptrp])) break;
 	}
 	else (*ptrp)++;
 
@@ -2268,23 +2353,22 @@ char *next;
 int qed, nonl, nest;
 #endif
 {
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp;
 	int pc, q, len;
 
-#ifdef	BASHSTYLE
+#ifdef	NESTINGQUOTE
 	pq = '\0';
 #endif
 	q = '\0';
 	len = strlen(next);
 	while (s[*ptrp]) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
-		pc = parsechar(&(s[*ptrp]), -1, '$', 1, &q, &pq);
+#ifdef	NESTINGQUOTE
+		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, &pq);
 #else
-		pc = parsechar(&(s[*ptrp]), -1, '$', 1, &q, NULL);
+		pc = parsechar(&(s[*ptrp]), -1, '$', EA_BACKQ, &q, NULL);
 #endif
 		if (pc == PC_WORD || pc == PC_META) (*ptrp)++;
 		else if (pc == '$') {
@@ -2329,7 +2413,7 @@ int plen, mode;
 
 	if (!s || !*s) return(NULL);
 	tmp = strndup2(pattern, plen);
-	pattern = evalarg(tmp, 0, 1, '\0');
+	pattern = evalarg(tmp, '\0', EA_BACKQ);
 	free(tmp);
 	re = regexp_init(pattern, -1);
 	free(pattern);
@@ -2441,18 +2525,19 @@ int plen, *modep;
 
 			for (i = j = 0; arglist[i + 1]; i++)
 				j += addmeta(NULL,
-					arglist[i + 1], 0, quoted);
+					arglist[i + 1], quoted, EA_KEEPMETA);
 			if (i <= 0) cp = strdup2("");
 			else {
 				j += (i - 1) * 3;
 				cp = malloc2(j + 1);
-				j = addmeta(cp, arglist[1], 0, quoted);
+				j = addmeta(cp, arglist[1],
+					quoted, EA_KEEPMETA);
 				for (i = 2; arglist[i]; i++) {
 					cp[j++] = quoted;
 					cp[j++] = sp;
 					cp[j++] = quoted;
-					j += addmeta(&(cp[j]),
-						arglist[i], 0, quoted);
+					j += addmeta(&(cp[j]), arglist[i],
+						quoted, EA_KEEPMETA);
 				}
 				cp[j] = '\0';
 			}
@@ -2527,7 +2612,9 @@ int s, len, vlen, mode, quoted;
 	else if (mode == '=' && !isidentchar(*arg)) return(-1);
 
 	val = strndup2(&(arg[s]), vlen);
-	*cpp = evalarg(val, (mode == '=' || mode == '?'), 1, quoted);
+	*cpp = evalarg(val, quoted,
+		(mode == '=' || mode == '?')
+		? EA_STRIPQ | EA_BACKQ : EA_BACKQ);
 	free(val);
 	if (!*cpp) return(-1);
 
@@ -2544,7 +2631,7 @@ int s, len, vlen, mode, quoted;
 	/* bash does not evaluates a quoted string in substitution itself */
 		free(*cpp);
 		val = strndup2(&(arg[s]), vlen);
-		*cpp = evalarg(val, 0, 1, quoted);
+		*cpp = evalarg(val, quoted, EA_BACKQ);
 		free(val);
 		if (!*cpp) return(-1);
 # ifdef	FD
@@ -2639,7 +2726,7 @@ int quoted;
 	c = '\0';
 	if ((cp = new)) /*EMPTY*/;
 	else if (isidentchar(*top)) cp = getenvvar(top, len);
-	else if (isdigit(*top)) {
+	else if (isdigit2(*top)) {
 		if (len > 1) return(-1);
 		if (!argvar) {
 			(*bufp)[ptr++] = '$';
@@ -2692,9 +2779,9 @@ int quoted;
 	}
 
 	if (!mode && (c != '@' || !quoted)) {
-		vlen = addmeta(NULL, cp, 0, quoted);
+		vlen = addmeta(NULL, cp, quoted, EA_KEEPMETA);
 		*bufp = insertarg(*bufp, ptr, arg, *argp - arg + 1, vlen);
-		addmeta(&((*bufp)[ptr]), cp, 0, quoted);
+		addmeta(&((*bufp)[ptr]), cp, quoted, EA_KEEPMETA);
 	}
 	else if (!cp) vlen = 0;
 	else if (c == '@' && !*cp && quoted
@@ -2871,10 +2958,10 @@ int rest;
 
 	stripquote(bbuf, 0);
 	if (!(tmp = (*backquotefunc)(bbuf))) return(buf);
-	len = addmeta(NULL, tmp, 1, '\0');
+	len = addmeta(NULL, tmp, '\0', 0);
 	size = *ptrp + len + rest + 1;
 	buf = realloc2(buf, size);
-	addmeta(&(buf[*ptrp]), tmp, 1, '\0');
+	addmeta(&(buf[*ptrp]), tmp, '\0', 0);
 	*ptrp += len;
 	free(tmp);
 
@@ -2979,24 +3066,28 @@ char **argp;
 }
 #endif	/* !MINIMUMSHELL */
 
-char *evalarg(arg, stripq, backq, qed)
+char *evalarg(arg, qed, flags)
 char *arg;
-int stripq, backq, qed;
+int qed, flags;
 {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char path[MAXPATHLEN], alias[MAXPATHLEN];
 #endif
-#ifdef	BASHSTYLE
+#ifndef	MINIMUMSHELL
+	int prev;
+#endif
+#ifdef	NESTINGQUOTE
 	int pq;
 #endif
 	char *cp, *buf, *bbuf;
-	int i, j, pc, prev, q;
+	int i, j, pc, q;
 
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	if (*arg == '"' && (i = strlen(arg)) > 2 && arg[i - 1] == '"') {
 		strncpy2(path, &(arg[1]), i - 2);
 		if (shortname(path, alias) == alias) {
-			if (stripq) return(strdup2(alias));
+			if (flags & (EA_STRIPQ | EA_STRIPQLATER))
+				return(strdup2(alias));
 			i = strlen(alias);
 			buf = malloc2(i + 2 + 1);
 			buf[0] = '"';
@@ -3010,20 +3101,23 @@ int stripq, backq, qed;
 
 	i = strlen(arg) + 1;
 	buf = malloc2(i);
-	if (!backquotefunc) backq = 0;
-	bbuf = (backq) ? malloc2(i) : NULL;
-#ifdef	BASHSTYLE
-	pq = '\0';
+	if (!backquotefunc) flags &= ~EA_BACKQ;
+	bbuf = (flags & EA_BACKQ) ? malloc2(i) : NULL;
+#ifndef	MINIMUMSHELL
+	prev =
 #endif
-	prev = q = '\0';
+#ifdef	NESTINGQUOTE
+	pq =
+#endif
+	q = '\0';
 	i = j = 0;
+	cp = arg;
 
-	for (cp = arg; *cp; prev = *(cp++)) {
-#ifdef	BASHSTYLE
-	/* bash can include `...` in "..." */
-		pc = parsechar(cp, -1, '$', backq, &q, &pq);
+	while (*cp) {
+#ifdef	NESTINGQUOTE
+		pc = parsechar(cp, -1, '$', flags, &q, &pq);
 #else
-		pc = parsechar(cp, -1, '$', backq, &q, NULL);
+		pc = parsechar(cp, -1, '$', flags, &q, NULL);
 #endif
 		if (pc == PC_CLQUOTE) {
 			if (*cp == '`') {
@@ -3032,7 +3126,7 @@ int stripq, backq, qed;
 					bbuf, strlen(cp + 1));
 				j = 0;
 			}
-			else if (!stripq) buf[i++] = *cp;
+			else if (!(flags & EA_STRIPQ)) buf[i++] = *cp;
 		}
 		else if (pc == PC_WORD) {
 			if (q == '`') {
@@ -3067,16 +3161,18 @@ int stripq, backq, qed;
 		}
 		else if (pc == PC_META) {
 			cp++;
-			if (*cp == '$') /*EMPTY*/;
-			else if (backq && *cp == '`') /*EMPTY*/;
-			else if (stripq && (*cp == '\'' || *cp == '"'))
+			if (flags & EA_KEEPMETA) buf[i++] = PMETA;
+			else if (*cp == '$') /*EMPTY*/;
+			else if ((flags & EA_BACKQ) && *cp == '`') /*EMPTY*/;
+			else if ((flags & EA_STRIPQ)
+			&& (*cp == '\'' || *cp == '"'))
 				/*EMPTY*/;
 			else buf[i++] = PMETA;
 			buf[i++] = *cp;
 		}
 		else if (pc == PC_OPQUOTE) {
 			if (*cp == '`') j = 0;
-			else if (!stripq) buf[i++] = *cp;
+			else if (!(flags & EA_STRIPQ)) buf[i++] = *cp;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
 #ifndef	MINIMUMSHELL
@@ -3084,16 +3180,25 @@ int stripq, backq, qed;
 			i = evalhome(&buf, i, &cp);
 #endif
 		else buf[i++] = *cp;
+
+#ifdef	MINIMUMSHELL
+		cp++;
+#else
+		prev = *(cp++);
+#endif
 	}
 #ifndef	BASHSTYLE
 	/* bash does not allow unclosed quote */
-	if (backq && q == '`') {
+	if ((flags & EA_BACKQ) && q == '`') {
 		bbuf[j] = '\0';
 		buf = replacebackquote(buf, &i, bbuf, 0);
 	}
 #endif
 	if (bbuf) free(bbuf);
 	buf[i] = '\0';
+
+	if (flags & EA_STRIPQLATER) stripquote(buf, EA_STRIPQ);
+
 	return(buf);
 }
 
@@ -3110,9 +3215,9 @@ char ***argvp, *ifs;
 				'\0', 0, &quote, NULL);
 			if (pc == PC_WORD || pc == PC_META) i++;
 			else if (pc != PC_NORMAL) /*EMPTY*/;
-			else if (strchr(ifs, (*argvp)[n][i])) {
+			else if (strchr2(ifs, (*argvp)[n][i])) {
 				for (j = i + 1; (*argvp)[n][j]; j++)
-					if (!strchr(ifs, (*argvp)[n][j]))
+					if (!strchr2(ifs, (*argvp)[n][j]))
 						break;
 				if (!i) {
 					for (i = 0; (*argvp)[n][i + j]; i++)
@@ -3138,78 +3243,47 @@ char ***argvp, *ifs;
 			free((*argvp)[n--]);
 			memmove((char *)(&((*argvp)[n + 1])),
 				(char *)(&((*argvp)[n + 2])),
-				(--argc - n) * sizeof(char *));
+				(argc-- - n) * sizeof(char *));
 		}
 	}
 	return(argc);
 }
 
-int evalglob(argc, argvp, stripq)
+int evalglob(argc, argvp, flags)
 int argc;
 char ***argvp;
-int stripq;
+int flags;
 {
-	char *cp, **wild;
-	ALLOC_T size;
-	int i, j, n, pc, w, quote;
+	char **wild;
+	int i, n;
 
 	for (n = 0; n < argc; n++) {
-		cp = c_realloc(NULL, 0, &size);
-		for (i = j = w = 0, quote = '\0'; (*argvp)[n][i]; i++) {
-			cp = c_realloc(cp, j + 1, &size);
-			pc = parsechar(&((*argvp)[n][i]), -1,
-				'\0', 0, &quote, NULL);
-			if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
-				if (stripq) continue;
-			}
-			else if (pc == PC_WORD) cp[j++] = (*argvp)[n][i++];
-			else if (pc == PC_META) {
-				i++;
-				if (quote
-				&& !strchr(DQ_METACHAR, (*argvp)[n][i]))
-					cp[j++] = PMETA;
-			}
-			else if (pc != PC_NORMAL) /*EMPTY*/;
-			else if (!strchr("?*[", (*argvp)[n][i])) /*EMPTY*/;
-			else if ((wild = evalwild((*argvp)[n]))) {
-				w = countvar(wild);
-				if (w > 1) {
-					*argvp = (char **)realloc2(*argvp,
-						(argc + w) * sizeof(char *));
-					if (!*argvp) {
-						free(cp);
-						freevar(wild);
-						return(argc);
-					}
-					memmove((char *)(&((*argvp)[n + w])),
-						(char *)(&((*argvp)[n + 1])),
-						(argc - n) * sizeof(char *));
-					argc += w - 1;
-				}
-				free((*argvp)[n]);
-				free(cp);
-				memmove((char *)(&((*argvp)[n])),
-					(char *)wild, w * sizeof(char *));
-				free(wild);
-				n += w - 1;
-				break;
-			}
-
-			cp[j++] = (*argvp)[n][i];
+		if (!(wild = evalwild((*argvp)[n], flags))) {
+			stripquote((*argvp)[n], flags);
+			continue;
 		}
 
-		if (!w) {
-			cp[j] = '\0';
-			free((*argvp)[n]);
-			(*argvp)[n] = cp;
+		i = countvar(wild);
+		if (i > 1) {
+			*argvp = (char **)realloc2(*argvp,
+				(argc + i) * sizeof(char *));
+			memmove((char *)(&((*argvp)[n + i])),
+				(char *)(&((*argvp)[n + 1])),
+				(argc - n) * sizeof(char *));
+			argc += i - 1;
 		}
+		free((*argvp)[n]);
+		memmove((char *)(&((*argvp)[n])),
+			(char *)wild, i * sizeof(char *));
+		free(wild);
+		n += i - 1;
 	}
 	return(argc);
 }
 
-int stripquote(arg, stripq)
+int stripquote(arg, flags)
 char *arg;
-int stripq;
+int flags;
 {
 	int i, j, pc, quote, stripped;
 
@@ -3219,14 +3293,16 @@ int stripq;
 		pc = parsechar(&(arg[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_OPQUOTE || pc == PC_CLQUOTE) {
 			stripped = 1;
-			if (stripq) continue;
+			if (flags & EA_STRIPQ) continue;
 		}
 		else if (pc == PC_WORD) arg[j++] = arg[i++];
 		else if (pc == PC_META) {
-			i++;
 			stripped = 1;
-			if (quote && arg[i] != quote && arg[i] != PMETA)
-				arg[j++] = PMETA;
+			if ((flags & EA_KEEPMETA)
+			|| quote == '\''
+			|| (quote == '"' && !strchr(DQ_METACHAR, arg[i + 1])))
+				arg[j++] = arg[i];
+			i++;
 		}
 
 		arg[j++] = arg[i];
@@ -3235,9 +3311,9 @@ int stripq;
 	return(stripped);
 }
 
-char *_evalpath(path, eol, uniqdelim, evalq)
+char *_evalpath(path, eol, flags)
 char *path, *eol;
-int uniqdelim, evalq;
+int flags;
 {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char alias[MAXPATHLEN];
@@ -3250,7 +3326,7 @@ int uniqdelim, evalq;
 	else i = strlen(path);
 	cp = strndup2(path, i);
 
-	if (!(tmp = evalarg(cp, 0, 0, '\''))) {
+	if (!(tmp = evalarg(cp, '\'', EA_KEEPMETA))) {
 		*cp = '\0';
 		return(cp);
 	}
@@ -3264,7 +3340,7 @@ int uniqdelim, evalq;
 		pc = parsechar(&(cp[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_CLQUOTE) {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
-			if (!evalq && top >= 0 && ++top < j) {
+			if ((flags & EA_NOEVALQ) && top >= 0 && ++top < j) {
 				tmp[j] = '\0';
 				if (shortname(&(tmp[top]), alias) == alias) {
 					int len;
@@ -3278,23 +3354,26 @@ int uniqdelim, evalq;
 			}
 			top = -1;
 #endif
-			if (evalq) continue;
+			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc == PC_WORD) tmp[j++] = cp[i++];
 		else if (pc == PC_META) {
 			i++;
-			if (!evalq
-			|| (quote && !strchr(DQ_METACHAR, cp[i])))
+			if ((flags & EA_KEEPMETA)
+			|| quote == '\''
+			|| (quote == '"' && !strchr(DQ_METACHAR, cp[i])))
 				tmp[j++] = PMETA;
 		}
 		else if (pc == PC_OPQUOTE) {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 			if (cp[i] == '"') top = j;
 #endif
-			if (evalq) continue;
+			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
-		else if (uniqdelim && cp[i] == _SC_ && c == _SC_) continue;
+		else if (!(flags & EA_NOUNIQDELIM)
+		&& cp[i] == _SC_ && c == _SC_)
+			continue;
 		tmp[j++] = cp[i];
 	}
 	tmp[j] = '\0';
@@ -3302,15 +3381,15 @@ int uniqdelim, evalq;
 	return(tmp);
 }
 
-char *evalpath(path, uniqdelim)
+char *evalpath(path, flags)
 char *path;
-int uniqdelim;
+int flags;
 {
 	char *cp;
 
 	if (!path || !*path) return(path);
 	for (cp = path; *cp == ' ' || *cp == '\t'; cp++);
-	cp = _evalpath(cp, NULL, uniqdelim, 1);
+	cp = _evalpath(cp, NULL, flags);
 	free(path);
 	return(cp);
 }
