@@ -4,7 +4,6 @@
  *	Tree Mode Module
  */
 
-#include <ctype.h>
 #include "fd.h"
 #include "func.h"
 #include "kanji.h"
@@ -97,9 +96,7 @@ int disp;
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
 		strcpy(&(path[len]), dp -> d_name);
-		if (limit-- <= 0
-		|| (stat2(path, &st) >= 0
-		&& (st.st_mode & S_IFMT) == S_IFDIR)) {
+		if (limit-- <= 0 || (stat2(path, &st) >= 0 && s_isdir(&st))) {
 			if (!disp) {
 				i++;
 				break;
@@ -185,7 +182,8 @@ int level, *maxp;
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)
 		|| Xstat(nodospath(tmp, dp -> d_name), &st) < 0
-		|| (st.st_mode & S_IFMT) != S_IFDIR) continue;
+		|| !s_isdir(&st))
+			continue;
 		list = b_realloc(list, *maxp, treelist);
 		if (!subdir) {
 			list[*maxp].name = strdup2(dp -> d_name);
@@ -666,7 +664,7 @@ static int NEAR _tree_input(VOID_A)
 			break;
 		case 'l':
 			if (!(cwd = inputstr(LOGD_K, 0, -1, NULL, 1))
-			|| !*(cwd = evalpath(cwd, 1))) break;
+			|| !*(cwd = evalpath(cwd, 0))) break;
 			if (chdir2(cwd) >= 0) {
 				free(cwd);
 				break;
@@ -681,7 +679,7 @@ static int NEAR _tree_input(VOID_A)
 		case K_ESC:
 			break;
 		default:
-			if (ch < 'A' || ch > 'Z') break;
+			if (!isupper2(ch)) break;
 			if (tr_line == tr_bottom - 1) {
 				tr_line = tr_top = LFILETOP + 1;
 				searchtree();
